@@ -1,6 +1,7 @@
 /***************************************************************************
  *   
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -36,6 +37,9 @@
 #include <zorp/stream.h>
 #include <zorp/source.h>
 
+/* FIXME: should be run-time configurable */
+#define MAX_READ_AT_A_TIME 30
+
 typedef struct _ZPlugIOBuffer
 {
   gchar *buf;
@@ -45,8 +49,7 @@ typedef struct _ZPlugIOBuffer
 
 struct _ZPlugSession
 {
-  ZProxy *owner;
-  ZSessionVars *vars;
+  ZRefCount ref_count;
   ZPlugSessionData *session_data;
   ZPoll *poll;
   ZStream *endpoints[EP_MAX];
@@ -58,6 +61,8 @@ struct _ZPlugSession
   GSource *stats_timeout;
   GTimeVal started_time;
   guint global_packet_count;
+  gpointer *user_data;
+  gboolean started;
 };
 
 /* possible eofmask values */
@@ -92,15 +97,23 @@ z_plug_update_eof_mask(ZPlugSession *self, guint add_mask)
   
   if ((self->eofmask & (EOF_DESTROYED | EOF_CLIENT_REMOVED | EOF_SERVER_REMOVED)) == (EOF_CLIENT_REMOVED | EOF_SERVER_REMOVED))
     {
-      self->session_data->destroy_queue = g_list_prepend(self->session_data->destroy_queue, self);
+      z_plug_session_cancel(self);
       self->eofmask |= EOF_DESTROYED;
     }
+
   /*LOG
     This message reports that the end-of-file status has been updated.
    */
-  z_proxy_log(self->owner, CORE_DEBUG, 7, "eofmask updated; old_mask='%04x', eof_mask='%04x'", old_mask, self->eofmask);
-}
+  z_log(NULL, CORE_DEBUG, 7, "eofmask updated; old_mask='%04x', eof_mask='%04x'", old_mask, self->eofmask);
 
+  if (!(old_mask & EOF_DESTROYED) && (self->eofmask & EOF_DESTROYED))
+    {
+      /* WARNING: calling ->finish() may have freed _this_ PlugSession
+         instance! It's forbidden to touch self after this call! */
+      if (self->session_data->finish)
+        self->session_data->finish(self, self->user_data);
+    }
+}
 
 
 static guint
@@ -108,7 +121,7 @@ z_plug_read_input(ZPlugSession *self, ZStream *input, ZPlugIOBuffer *buf)
 {
   GIOStatus rc;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   rc = z_stream_read(input, buf->buf, self->session_data->buffer_size, &buf->end, NULL);
   if (rc == G_IO_STATUS_NORMAL)
     {
@@ -118,16 +131,16 @@ z_plug_read_input(ZPlugSession *self, ZStream *input, ZPlugIOBuffer *buf)
       if (self->session_data->packet_stats_interval_packet &&
          (self->global_packet_count % self->session_data->packet_stats_interval_packet) == 0)
         {
-          if (!self->session_data->packet_stats(self->owner, self->vars, 
+          if (!self->session_data->packet_stats(self,
                                                 self->buffers[EP_CLIENT].packet_bytes, 
                                                 self->buffers[EP_CLIENT].packet_count,
                                                 self->buffers[EP_SERVER].packet_bytes,
-                                                self->buffers[EP_SERVER].packet_count))
+                                                self->buffers[EP_SERVER].packet_count,
+                                                self->user_data))
             z_plug_update_eof_mask(self, EOF_ALL);
         }
     }
-  z_proxy_leave(self->owner);
-  return rc;
+  z_return(rc);
 }
 
 static GIOStatus
@@ -136,7 +149,7 @@ z_plug_write_output(ZPlugSession *self G_GNUC_UNUSED, ZPlugIOBuffer *buf, ZStrea
   GIOStatus rc;
   gsize bytes_written;
   
-  z_proxy_enter(self->owner);
+  z_enter();
   if (buf->ofs != buf->end)
     {
       /* buffer not empty */
@@ -146,28 +159,22 @@ z_plug_write_output(ZPlugSession *self G_GNUC_UNUSED, ZPlugIOBuffer *buf, ZStrea
         case G_IO_STATUS_NORMAL:
           buf->ofs += bytes_written;
           break;
+
         case G_IO_STATUS_AGAIN:
           break;
+
         default:
-          z_proxy_leave(self->owner);
-          return rc;
+          z_return(rc);
         }
+
       if (buf->ofs != buf->end)
         {
-          z_stream_set_cond(output,
-                           Z_STREAM_FLAG_WRITE,
-                           TRUE);
-          z_proxy_leave(self->owner);
-          return G_IO_STATUS_AGAIN;
-        }
-      else
-        {
-          z_proxy_leave(self->owner);
-          return G_IO_STATUS_NORMAL;
+          z_stream_set_cond(output, G_IO_OUT, TRUE);
+          z_leave();
+          z_return(G_IO_STATUS_AGAIN);
         }
     }
-  z_proxy_leave(self->owner);
-  return G_IO_STATUS_NORMAL;
+  z_return(G_IO_STATUS_NORMAL);
 }
 
 static GIOStatus
@@ -176,29 +183,25 @@ z_plug_copy_data(ZPlugSession *self, ZStream *from, ZStream *to, ZPlugIOBuffer *
   GIOStatus rc = G_IO_STATUS_ERROR;
   int pkt_count = 0;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   
   if (self->timeout)
     z_timeout_source_set_timeout(self->timeout, self->session_data->timeout);
   
-  z_stream_set_cond(from,
-                   Z_STREAM_FLAG_READ,
-                   FALSE);
+  if (!from || !buf)
+    z_return(G_IO_STATUS_ERROR);
+  
+  z_stream_set_cond(from, G_IO_IN, FALSE);
 
   if (to)
     {
-      z_stream_set_cond(to,
-                       Z_STREAM_FLAG_WRITE,
-                       FALSE);
+      z_stream_set_cond(to, G_IO_OUT, FALSE);
       rc = z_plug_write_output(self, buf, to);
       if (rc != G_IO_STATUS_NORMAL)
-        {
-          z_proxy_leave(self->owner);
-          return rc;
-        }
+        z_return(rc);
     }
 
-  while (1) 
+  while (pkt_count < MAX_READ_AT_A_TIME)
     {  
       buf->ofs = buf->end = 0;
       rc = z_plug_read_input(self, from, buf);
@@ -210,48 +213,33 @@ z_plug_copy_data(ZPlugSession *self, ZStream *from, ZStream *to, ZPlugIOBuffer *
               if (rc == G_IO_STATUS_AGAIN)
                 break;
               else if (rc != G_IO_STATUS_NORMAL)
-                {
-                  z_proxy_leave(self->owner);
-                  return rc;
-                }
+                z_return(rc);
             }
         }
       else if (rc == G_IO_STATUS_AGAIN)
         break;
       else if (rc == G_IO_STATUS_EOF)
-        {
-          z_proxy_leave(self->owner);
-          return rc;
-        }
+        z_return(rc);
       else
-        {
-          z_proxy_leave(self->owner);
-          return G_IO_STATUS_ERROR;
-        }
+        z_return(G_IO_STATUS_ERROR);
       pkt_count++;
     }
 
   if (buf->ofs == buf->end)
-    {
-      z_stream_set_cond(from,
-                       Z_STREAM_FLAG_READ,
-                       TRUE);
-    }
-  z_proxy_leave(self->owner);
-  return rc;
+    z_stream_set_cond(from, G_IO_IN, TRUE);
+
+  z_return(rc);
 }
 
 
 /* callbacks when no stacking is made */
 static gboolean
-z_plug_copy_client_to_server(ZStream *stream G_GNUC_UNUSED,
-                      GIOCondition  cond G_GNUC_UNUSED,
-                          gpointer  user_data)
+z_plug_copy_client_to_server(ZStream *stream G_GNUC_UNUSED, GIOCondition cond G_GNUC_UNUSED, gpointer user_data)
 {
-  ZPlugSession *self = (ZPlugSession *)user_data;
+  ZPlugSession *self = (ZPlugSession *) user_data;
   gboolean ret;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   if (self->session_data->copy_to_server)
     ret = z_plug_copy_data(self,
                          self->endpoints[EP_CLIENT],
@@ -268,6 +256,7 @@ z_plug_copy_client_to_server(ZStream *stream G_GNUC_UNUSED,
     case G_IO_STATUS_NORMAL:
     case G_IO_STATUS_AGAIN:
       break;
+      
     case G_IO_STATUS_EOF:
       if (self->session_data->shutdown_soft)
         {
@@ -280,24 +269,21 @@ z_plug_copy_client_to_server(ZStream *stream G_GNUC_UNUSED,
           z_plug_update_eof_mask(self, EOF_ALL);
         }
       break;
+      
     default:
       z_plug_update_eof_mask(self, EOF_ALL);
-      z_proxy_leave(self->owner);
-      return FALSE;
+      z_return(FALSE);
     }
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 static gboolean
-z_plug_copy_server_to_client(ZStream *stream G_GNUC_UNUSED,
-                      GIOCondition  cond G_GNUC_UNUSED,
-                          gpointer  user_data)
+z_plug_copy_server_to_client(ZStream *stream G_GNUC_UNUSED, GIOCondition cond G_GNUC_UNUSED, gpointer user_data)
 {
   ZPlugSession *self = (ZPlugSession *) user_data;
   GIOStatus ret;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   if (self->session_data->copy_to_client)
     ret = z_plug_copy_data(self,
                          self->endpoints[EP_SERVER],
@@ -314,6 +300,7 @@ z_plug_copy_server_to_client(ZStream *stream G_GNUC_UNUSED,
     case G_IO_STATUS_NORMAL:
     case G_IO_STATUS_AGAIN:
       break;
+
     case G_IO_STATUS_EOF:
       if (self->session_data->shutdown_soft)
         {
@@ -326,26 +313,22 @@ z_plug_copy_server_to_client(ZStream *stream G_GNUC_UNUSED,
           z_plug_update_eof_mask(self, EOF_ALL);
         }
       break;
+      
     default:
       z_plug_update_eof_mask(self, EOF_ALL);
-      z_proxy_leave(self->owner);
-      return FALSE;
+      z_return(FALSE);
     }
-
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 /* callbacks when a stacked module exists */
 static gboolean
-z_plug_copy_client_to_down(ZStream       *stream G_GNUC_UNUSED,
-                           GIOCondition  cond G_GNUC_UNUSED,
-                           gpointer      user_data)
+z_plug_copy_client_to_down(ZStream *stream G_GNUC_UNUSED, GIOCondition cond G_GNUC_UNUSED, gpointer user_data)
 {
   ZPlugSession *self = (ZPlugSession *) user_data;
   GIOStatus ret;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   if (self->session_data->copy_to_server)
     ret = z_plug_copy_data(self,
                          self->endpoints[EP_CLIENT],
@@ -362,15 +345,12 @@ z_plug_copy_client_to_down(ZStream       *stream G_GNUC_UNUSED,
     case G_IO_STATUS_NORMAL:
     case G_IO_STATUS_AGAIN:
       break;
+      
     case G_IO_STATUS_EOF:
       if (self->session_data->shutdown_soft)
         {
-          z_stream_shutdown(self->endpoints[EP_CLIENT],
-                            SHUT_RD,
-                            NULL);
-          z_stream_shutdown(self->stacked->downstreams[EP_CLIENT],
-                            SHUT_WR,
-                            NULL);
+          z_stream_shutdown(self->endpoints[EP_CLIENT], SHUT_RD, NULL);
+          z_stream_shutdown(self->stacked->downstreams[EP_CLIENT], SHUT_WR, NULL);
           z_plug_update_eof_mask(self, EOF_CLIENT_R);
         }
       else
@@ -378,44 +358,36 @@ z_plug_copy_client_to_down(ZStream       *stream G_GNUC_UNUSED,
           z_plug_update_eof_mask(self, EOF_ALL);
         }
       break;
+
     default:
       z_plug_update_eof_mask(self, EOF_ALL);
-      z_proxy_leave(self->owner);
-      return FALSE;
+      z_return(FALSE);
     }
-
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 static gboolean
-z_plug_copy_down_to_client(ZStream *stream G_GNUC_UNUSED,
-                    GIOCondition  cond G_GNUC_UNUSED,
-                        gpointer  user_data)
+z_plug_copy_down_to_client(ZStream *stream G_GNUC_UNUSED, GIOCondition cond G_GNUC_UNUSED, gpointer user_data)
 {
   ZPlugSession *self = (ZPlugSession *) user_data;
   GIOStatus ret;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   ret = z_plug_copy_data(self,
                        self->stacked->downstreams[EP_CLIENT],
                        self->endpoints[EP_CLIENT],
                        &self->buffers[EP_CLIENT]);
-
   switch (ret)
     {
     case G_IO_STATUS_NORMAL:
     case G_IO_STATUS_AGAIN:
       break;
+      
     case G_IO_STATUS_EOF:
       if (self->session_data->shutdown_soft)
         {
-          z_stream_shutdown(self->stacked->downstreams[EP_CLIENT],
-                            SHUT_RD,
-                            NULL);
-          z_stream_shutdown(self->endpoints[EP_CLIENT],
-                            SHUT_WR,
-                            NULL);
+          z_stream_shutdown(self->stacked->downstreams[EP_CLIENT], SHUT_RD, NULL);
+          z_stream_shutdown(self->endpoints[EP_CLIENT], SHUT_WR, NULL);
           z_plug_update_eof_mask(self, EOF_CLIENT_W);
         }
       else
@@ -423,24 +395,21 @@ z_plug_copy_down_to_client(ZStream *stream G_GNUC_UNUSED,
           z_plug_update_eof_mask(self, EOF_ALL);
         }
       break;
+      
     default:
       z_plug_update_eof_mask(self, EOF_ALL);
-      z_proxy_leave(self->owner);
-      return FALSE;
+      z_return(FALSE);
     }
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 static gboolean
-z_plug_copy_server_to_down(ZStream *stream G_GNUC_UNUSED,
-                    GIOCondition  cond G_GNUC_UNUSED,
-                        gpointer  user_data)
+z_plug_copy_server_to_down(ZStream *stream G_GNUC_UNUSED, GIOCondition cond G_GNUC_UNUSED, gpointer user_data)
 {
   ZPlugSession *self = (ZPlugSession *) user_data;
   GIOStatus ret;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   if (self->session_data->copy_to_client)
     ret = z_plug_copy_data(self,
                          self->endpoints[EP_SERVER],
@@ -451,21 +420,17 @@ z_plug_copy_server_to_down(ZStream *stream G_GNUC_UNUSED,
                          self->endpoints[EP_SERVER],
                          NULL,
                          &self->downbufs[EP_SERVER]);
-
   switch (ret)
     {
     case G_IO_STATUS_NORMAL:
     case G_IO_STATUS_AGAIN:
       break;
+      
     case G_IO_STATUS_EOF:
       if (self->session_data->shutdown_soft)
         {
-          z_stream_shutdown(self->endpoints[EP_SERVER],
-                            SHUT_RD,
-                            NULL);
-          z_stream_shutdown(self->stacked->downstreams[EP_SERVER],
-                            SHUT_WR,
-                            NULL);
+          z_stream_shutdown(self->endpoints[EP_SERVER], SHUT_RD, NULL);
+          z_stream_shutdown(self->stacked->downstreams[EP_SERVER], SHUT_WR, NULL);
           z_plug_update_eof_mask(self, EOF_SERVER_R);
         }
       else
@@ -473,45 +438,36 @@ z_plug_copy_server_to_down(ZStream *stream G_GNUC_UNUSED,
           z_plug_update_eof_mask(self, EOF_ALL);
         }
       break;
+      
     default:
       z_plug_update_eof_mask(self, EOF_ALL);
-      z_proxy_leave(self->owner);
-      return FALSE;
+      z_return(FALSE);
     }
-
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 static gboolean
-z_plug_copy_down_to_server(ZStream *stream G_GNUC_UNUSED,
-                    GIOCondition  cond G_GNUC_UNUSED,
-                        gpointer  user_data)
+z_plug_copy_down_to_server(ZStream *stream G_GNUC_UNUSED, GIOCondition cond G_GNUC_UNUSED, gpointer user_data)
 {
   ZPlugSession *self = (ZPlugSession *) user_data;
   GIOStatus ret;
 
-  z_proxy_enter(self->owner);
+  z_enter();
   ret = z_plug_copy_data(self,
                        self->stacked->downstreams[EP_SERVER],
                        self->endpoints[EP_SERVER],
                        &self->buffers[EP_SERVER]);
-
   switch (ret)
     {
     case G_IO_STATUS_NORMAL:
     case G_IO_STATUS_AGAIN:
-      z_proxy_leave(self->owner);
-      return TRUE;
+      z_return(TRUE);
+
     case G_IO_STATUS_EOF:
       if (self->session_data->shutdown_soft)
         {
-          z_stream_shutdown(self->stacked->downstreams[EP_SERVER],
-                            SHUT_RD,
-                            NULL);
-          z_stream_shutdown(self->endpoints[EP_SERVER],
-                            SHUT_WR,
-                            NULL);
+          z_stream_shutdown(self->stacked->downstreams[EP_SERVER], SHUT_RD, NULL);
+          z_stream_shutdown(self->endpoints[EP_SERVER], SHUT_WR, NULL);
           z_plug_update_eof_mask(self, EOF_SERVER_W);
         }
       else
@@ -519,14 +475,12 @@ z_plug_copy_down_to_server(ZStream *stream G_GNUC_UNUSED,
           z_plug_update_eof_mask(self, EOF_ALL);
         }
       break;
+
     default:
       z_plug_update_eof_mask(self, EOF_ALL);
-      z_proxy_leave(self->owner);
-      return FALSE;
+      z_return(FALSE);
     }
-
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 gboolean
@@ -534,128 +488,64 @@ z_plug_timeout(gpointer user_data)
 {
   ZPlugSession *self = (ZPlugSession *) user_data;
   
-  z_proxy_enter(self->owner);
+  z_enter();
   z_plug_update_eof_mask(self, EOF_ALL);
-  z_proxy_leave(self->owner);
-  return FALSE;
+  z_return(FALSE);
 }
 
 /* FIXME: merge these two functions */
 gboolean
 z_plug_session_init_streams(ZPlugSession *self)
 {
-  z_proxy_enter(self->owner);
-  
+  z_enter();
   self->buffers[EP_CLIENT].buf = g_new0(char, self->session_data->buffer_size);
   self->buffers[EP_SERVER].buf = g_new0(char, self->session_data->buffer_size);
   
   z_stream_set_nonblock(self->endpoints[EP_CLIENT], TRUE);
-  
-  z_stream_set_callback(self->endpoints[EP_CLIENT],
-                        Z_STREAM_FLAG_READ, 
-                        z_plug_copy_client_to_server,
-                        self,
-                        NULL);
-  z_stream_set_callback(self->endpoints[EP_CLIENT],
-                        Z_STREAM_FLAG_WRITE,
-                        z_plug_copy_server_to_client,
-                        self,
-                        NULL);
-  z_stream_set_cond(self->endpoints[EP_CLIENT],
-                   Z_STREAM_FLAG_READ,
-                   TRUE);
-
+  z_stream_set_callback(self->endpoints[EP_CLIENT], G_IO_IN, z_plug_copy_client_to_server, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+  z_stream_set_callback(self->endpoints[EP_CLIENT], G_IO_OUT, z_plug_copy_server_to_client, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+  z_stream_set_cond(self->endpoints[EP_CLIENT], G_IO_IN, TRUE);
   z_stream_set_timeout(self->endpoints[EP_CLIENT], -2);
   
   z_stream_set_nonblock(self->endpoints[EP_SERVER], TRUE);
-  z_stream_set_callback(self->endpoints[EP_SERVER],
-                        Z_STREAM_FLAG_READ,
-                        z_plug_copy_server_to_client,
-                        self,
-                        NULL);
-  z_stream_set_callback(self->endpoints[EP_SERVER],
-                       Z_STREAM_FLAG_WRITE,
-                       z_plug_copy_client_to_server,
-                       self,
-                       NULL);
-  z_stream_set_cond(self->endpoints[EP_SERVER],
-                   Z_STREAM_FLAG_READ,
-                   TRUE);
+  z_stream_set_callback(self->endpoints[EP_SERVER], G_IO_IN, z_plug_copy_server_to_client, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+  z_stream_set_callback(self->endpoints[EP_SERVER], G_IO_OUT, z_plug_copy_client_to_server, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+  z_stream_set_cond(self->endpoints[EP_SERVER], G_IO_IN, TRUE);
   z_stream_set_timeout(self->endpoints[EP_SERVER], -2);
 
   z_poll_add_stream(self->poll, self->endpoints[EP_CLIENT]);
   z_poll_add_stream(self->poll, self->endpoints[EP_SERVER]);
-
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 static gboolean
 z_plug_session_init_stacked_streams(ZPlugSession *self)
 {
-  z_proxy_enter(self->owner);
+  z_enter();
   
   if (self->stacked)
     {
       self->downbufs[EP_CLIENT].buf = g_new0(char, self->session_data->buffer_size);
       self->downbufs[EP_SERVER].buf = g_new0(char, self->session_data->buffer_size);
 
-      z_stream_set_callback(self->endpoints[EP_CLIENT],
-                            Z_STREAM_FLAG_READ,
-                            z_plug_copy_client_to_down,
-                            self,
-                            NULL);
+      z_stream_set_callback(self->endpoints[EP_CLIENT], G_IO_IN, z_plug_copy_client_to_down, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+      z_stream_set_callback(self->endpoints[EP_CLIENT], G_IO_OUT, z_plug_copy_down_to_client, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
 
-      z_stream_set_callback(self->endpoints[EP_CLIENT],
-                            Z_STREAM_FLAG_WRITE,
-                            z_plug_copy_down_to_client,
-                            self,
-                            NULL);
+      z_stream_set_callback(self->endpoints[EP_SERVER], G_IO_IN, z_plug_copy_server_to_down, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+      z_stream_set_callback(self->endpoints[EP_SERVER], G_IO_OUT, z_plug_copy_down_to_server, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
 
-      z_stream_set_callback(self->endpoints[EP_SERVER],
-                            Z_STREAM_FLAG_READ,
-                            z_plug_copy_server_to_down,
-                            self,
-                            NULL);
-      z_stream_set_callback(self->endpoints[EP_SERVER],
-                            Z_STREAM_FLAG_WRITE,
-                            z_plug_copy_down_to_server,
-                            self,
-                            NULL);
+      z_stream_set_callback(self->stacked->downstreams[EP_CLIENT], G_IO_IN, z_plug_copy_down_to_client, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+      z_stream_set_callback(self->stacked->downstreams[EP_CLIENT], G_IO_OUT, z_plug_copy_client_to_down, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+      z_stream_set_cond(self->stacked->downstreams[EP_CLIENT], G_IO_IN, TRUE);
 
-      z_stream_set_callback(self->stacked->downstreams[EP_CLIENT],
-                            Z_STREAM_FLAG_READ,
-                            z_plug_copy_down_to_client,
-                            self, 
-                            NULL);
-      z_stream_set_callback(self->stacked->downstreams[EP_CLIENT],
-                            Z_STREAM_FLAG_WRITE,
-                            z_plug_copy_client_to_down,
-                            self, 
-                            NULL);
-      z_stream_set_cond(self->stacked->downstreams[EP_CLIENT],
-                       Z_STREAM_FLAG_READ,
-                       TRUE);
-
-      z_stream_set_callback(self->stacked->downstreams[EP_SERVER],
-                            Z_STREAM_FLAG_READ,
-                            z_plug_copy_down_to_server,
-                            self,
-                            NULL);
-      z_stream_set_callback(self->stacked->downstreams[EP_SERVER],
-                            Z_STREAM_FLAG_WRITE,
-                            z_plug_copy_server_to_down,
-                            self,
-                            NULL);
-      z_stream_set_cond(self->stacked->downstreams[EP_SERVER],
-                       Z_STREAM_FLAG_READ,
-                       TRUE);
+      z_stream_set_callback(self->stacked->downstreams[EP_SERVER], G_IO_IN, z_plug_copy_down_to_server, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+      z_stream_set_callback(self->stacked->downstreams[EP_SERVER], G_IO_OUT, z_plug_copy_server_to_down, z_plug_session_ref(self), (GDestroyNotify) z_plug_session_unref);
+      z_stream_set_cond(self->stacked->downstreams[EP_SERVER], G_IO_IN, TRUE);
 
       z_poll_add_stream(self->poll, self->stacked->downstreams[EP_CLIENT]);
       z_poll_add_stream(self->poll, self->stacked->downstreams[EP_SERVER]);
     }
-  z_proxy_leave(self->owner);
-  return TRUE;
+  z_return(TRUE);
 }
 
 static gboolean
@@ -665,30 +555,67 @@ z_plug_session_stats_timeout(gpointer user_data)
   
   if (self->session_data->packet_stats)
     {
-      if (!self->session_data->packet_stats(self->owner, self->vars,
-              self->buffers[EP_CLIENT].packet_bytes,
-              self->buffers[EP_CLIENT].packet_count,
-              self->buffers[EP_SERVER].packet_bytes,
-              self->buffers[EP_SERVER].packet_count))
+      if (!self->session_data->packet_stats(self,
+                                            self->buffers[EP_CLIENT].packet_bytes,
+                                            self->buffers[EP_CLIENT].packet_count,
+                                            self->buffers[EP_SERVER].packet_bytes,
+                                            self->buffers[EP_SERVER].packet_count,
+                                            self->user_data))
         {
           z_plug_update_eof_mask(self, EOF_ALL);
         }
+      z_return(TRUE);
     }
-  else
-    {
-      /*LOG
-	This message indicates that packet stats interval was specified, but no action was configured to handle the event.
-	Check your policy for packetStats event.
-       */
-      z_proxy_log(self->owner, CORE_ERROR, 3, "Packet stats timeout elapsed, and no timeout callback specified;");
-      return FALSE;
-    }
-  return TRUE;
+  /*LOG
+    This message indicates that packet stats interval was specified, but no action was configured to handle the event.
+    Check your policy for packetStats event.
+   */
+  z_log(NULL, CORE_ERROR, 3, "Packet stats timeout elapsed, and no timeout callback specified;");
+  z_return(FALSE);
 }
 
+static ZPolicyObj *
+z_plug_session_query_bandwidth(ZPlugSession *self, gchar *name, gpointer value G_GNUC_UNUSED)
+{
+  GTimeVal now, spent;
+  double bandwidth = 0.0;
+  
+  g_get_current_time(&now);
+  spent.tv_sec = now.tv_sec - self->started_time.tv_sec;
+  spent.tv_usec = now.tv_usec - self->started_time.tv_usec;
+  if (spent.tv_usec < -500000)
+    spent.tv_sec++;
+    
+  if (strcmp(name, "bandwidth_to_client") == 0)
+    {
+      bandwidth = (double) self->buffers[EP_CLIENT].packet_bytes / spent.tv_sec;
+    }
+  else if (strcmp(name, "bandwidth_to_server") == 0)
+    {
+      bandwidth = (double) self->buffers[EP_SERVER].packet_bytes / spent.tv_sec;
+    }
+  return z_policy_var_build("d", bandwidth);
+}
+
+void
+z_plug_session_register_vars(ZPlugSession *self, ZPolicyDict *dict)
+{
+  z_policy_dict_register(dict, Z_VT_CUSTOM, "bandwidth_to_client", Z_VF_READ,
+                                 /* value, get, set, free, userdata, destroy-notify */
+                                 NULL, z_plug_session_query_bandwidth, NULL, NULL, self, NULL);
+
+  z_policy_dict_register(dict, Z_VT_CUSTOM, "bandwidth_to_server", Z_VF_READ, 
+                                 NULL, z_plug_session_query_bandwidth, NULL, NULL, self, NULL);
+                                 
+}
+
+
 gboolean
 z_plug_session_start(ZPlugSession *self, ZPoll *poll)
 {
+  if (self->started)
+    g_assert_not_reached();
+    
   z_poll_ref(poll);
   self->poll = poll;
   
@@ -713,99 +640,35 @@ z_plug_session_start(ZPlugSession *self, ZPoll *poll)
           context = z_poll_get_context(self->poll);
           g_source_attach(self->timeout, context);
         }
-      self->session_data->num_sessions++;
+      self->started = TRUE;
       return TRUE;
     }
   return FALSE;
 }
 
-static ZPolicyObj *
-z_plug_session_query_bandwidth(ZProxy *proxy G_GNUC_UNUSED, gchar *name, gpointer value)
-{
-  ZPlugSession *self = (ZPlugSession *) value;
-  GTimeVal now, spent;
-  double bandwidth = 0.0;
-  
-  g_get_current_time(&now);
-  spent.tv_sec = now.tv_sec - self->started_time.tv_sec;
-  spent.tv_usec = now.tv_usec - self->started_time.tv_usec;
-  if (spent.tv_usec < -500000)
-    spent.tv_sec++;
-    
-  if (strcmp(name, "bandwidth_to_client") == 0)
-    {
-      bandwidth = (double) self->buffers[EP_CLIENT].packet_bytes / spent.tv_sec;
-    }
-  else if (strcmp(name, "bandwidth_to_server") == 0)
-    {
-      bandwidth = (double) self->buffers[EP_SERVER].packet_bytes / spent.tv_sec;
-    }
-  return z_policy_var_build("d", bandwidth);
-}
-
-static void
-z_plug_do_nothing(gpointer user_data G_GNUC_UNUSED)
-{
-}
-
 void
-z_plug_session_register_vars(ZPlugSession *self)
-{
-  self->vars = z_session_vars_new();
-  
-  z_session_var_new(self->vars, "bandwidth_to_client", 
-                    Z_VAR_TYPE_CUSTOM | Z_VAR_GET,
-                    self, z_plug_session_query_bandwidth, NULL, z_plug_do_nothing);
-  z_session_var_new(self->vars, "bandwidth_to_server",
-                    Z_VAR_TYPE_CUSTOM | Z_VAR_GET,
-                    self, z_plug_session_query_bandwidth, NULL, z_plug_do_nothing);
-
-}
-
-ZPlugSession *
-z_plug_session_new(ZProxy *owner, ZPlugSessionData *session_data, ZStream *client_stream, ZStream *server_stream, ZStackedProxy *stacked)
-{
-  ZPlugSession *self = g_new0(ZPlugSession, 1);
-  
-  self->owner = owner;
-  z_stream_ref(client_stream);
-  z_stream_ref(server_stream);
-  
-  if (!client_stream->name[0])
-    g_snprintf(client_stream->name, sizeof(client_stream->name), "%s/%s", owner->session_id, "client");
-  if (!server_stream->name[0])
-    g_snprintf(server_stream->name, sizeof(server_stream->name), "%s/%s", owner->session_id, "server");
-
-  self->endpoints[EP_CLIENT] = client_stream;
-  self->endpoints[EP_SERVER] = server_stream;
-  self->stacked = stacked;
-  self->session_data = session_data;
-  z_plug_session_register_vars(self);
-  return self;
-}
-
-void
-z_plug_session_free(ZPlugSession *self)
+z_plug_session_cancel(ZPlugSession *self)
 {
   gint i;
   
+  if (!self->started)
+    return;
+    
   for (i = EP_CLIENT; i < EP_MAX; i++)
     {
       if (self->stacked)
         {
-          g_free(self->downbufs[i].buf);
           z_poll_remove_stream(self->poll, self->stacked->downstreams[i]);
         }
-      g_free(self->buffers[i].buf);
       z_poll_remove_stream(self->poll, self->endpoints[i]);
-      z_stream_shutdown(self->endpoints[i], SHUT_RDWR, NULL);
-      z_stream_close(self->endpoints[i], NULL);
-      z_stream_unref(self->endpoints[i]);
     }
   
   if (self->stacked)
-    z_stacked_proxy_destroy(self->stacked);
-  z_session_vars_destroy(self->vars);
+    {
+      z_stacked_proxy_destroy(self->stacked);
+      self->stacked = NULL;
+    }
+    
   if (self->stats_timeout)
     {
       g_source_destroy(self->stats_timeout);
@@ -818,21 +681,84 @@ z_plug_session_free(ZPlugSession *self)
       g_source_unref(self->timeout);
       self->timeout = NULL;
     }
-  z_poll_unref(self->poll);
+  self->started = FALSE;
+}
+
+static void
+z_plug_session_free(ZPlugSession *self)
+{
   g_free(self);
 }
 
-void
-z_plug_sessions_purge(ZPlugSessionData *session_data)
+ZPlugSession *
+z_plug_session_ref(ZPlugSession *self)
 {
-  while (session_data->destroy_queue)
+  z_refcount_inc(&self->ref_count);
+  return self;
+}
+
+void
+z_plug_session_unref(ZPlugSession *self)
+{
+  if (z_refcount_dec(&self->ref_count))
+    z_plug_session_free(self);
+}
+
+ZPlugSession *
+z_plug_session_new(ZPlugSessionData *session_data, ZStream *client_stream, ZStream *server_stream, ZStackedProxy *stacked, gpointer user_data)
+{
+  ZPlugSession *self = g_new0(ZPlugSession, 1);
+  gchar buf[Z_STREAM_MAX_NAME];
+  
+  self->user_data = user_data;
+  z_stream_ref(client_stream);
+  z_stream_ref(server_stream);
+  
+  if (!client_stream->name[0])
     {
-      ZPlugSession *session;
+      g_snprintf(buf, sizeof(buf), "%s/%s", fake_session_id, "client");
+      z_stream_set_name(client_stream, buf);
+    }
+  if (!server_stream->name[0])
+    {
+      g_snprintf(buf, sizeof(buf), "%s/%s", fake_session_id, "server");
+      z_stream_set_name(server_stream, buf);
+    }
+
+  self->endpoints[EP_CLIENT] = client_stream;
+  self->endpoints[EP_SERVER] = server_stream;
+  self->stacked = stacked;
+  self->session_data = session_data;
+  z_refcount_set(&self->ref_count, 1);
+  return self;
+}
+
+void
+z_plug_session_destroy(ZPlugSession *self)
+{
+  gint i;
+  
+  if (self)
+    {
+      g_assert(!self->started);
+      
+      for (i = EP_CLIENT; i < EP_MAX; i++)
+        {
+          if (self->downbufs[i].buf)
+            {
+              g_free(self->downbufs[i].buf);
+              self->downbufs[i].buf = NULL;
+            }
+          g_free(self->buffers[i].buf);
+          self->buffers[i].buf = NULL;
           
-      session = (ZPlugSession *) session_data->destroy_queue->data;
-      session_data->num_sessions--;
-      z_plug_session_free(session);
-      session_data->destroy_queue = g_list_delete_link(session_data->destroy_queue, session_data->destroy_queue);
+          z_stream_unref(self->endpoints[i]);
+          self->endpoints[i] = NULL;
+        }
+      
+      z_poll_unref(self->poll);
+      self->poll = NULL;
+      z_plug_session_unref(self);
     }
 }
 
