@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -42,16 +43,44 @@
 
 /* for python initialization functions */
 #include <zorp/pycore.h>
-#include <zorp/pyszig.h>
 #include <zorp/pydispatch.h>
 #include <zorp/pyattach.h>
 #include <zorp/pystream.h>
 #include <zorp/pyproxy.h>
 #include <zorp/pysockaddr.h>
-#include <zorp/pyfastpath.h>
+#include <zorp/pyproxygroup.h>
+#include <zorp/pybalance.h>
+#include <zorp/notification.h>
+
+/* for capability management */
+#include <zorp/cap.h>
 
 ZPolicy *current_policy;
 
+/* Notification thread function and indicating policy termination */
+extern gpointer z_notification_thread_main_func(gpointer data);
+extern void _notify_event_queue_terminate_request(ZPolicy *policy);
+
+
+const gchar *
+z_verdict_str(ZVerdict verdict)
+{
+  static const gchar *verdict_str[] =
+  {
+    [ZV_UNSPEC] = "Unspecified",
+    [ZV_ACCEPT] = "Accept data",
+    [ZV_DENY]   = "Deny data",
+    [ZV_REJECT] = "Reject data",
+    [ZV_ABORT]  = "Abort connection",
+    [ZV_DROP]   = "Drop data",
+    [ZV_POLICY] = "Call policy",
+    [ZV_ERROR]  = "Error while processing",
+  };
+  if ((guint) verdict <= 7)
+    return verdict_str[(guint) verdict];
+  return "Unknown";
+}
+
 /* helper functions used by proxies */
 
 gboolean 
@@ -63,13 +92,9 @@ z_policy_tuple_get_verdict(ZPolicyObj *tuple, guint *verdict)
   if (!z_policy_seq_check(tuple))
     {
       if (z_policy_var_parse(tuple, "i", verdict))
-        {
-          z_leave();
-          return TRUE;
-        }
+        z_return(TRUE);
       /* not a sequence nor an int */
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
                                                                                 
   tmp = z_policy_seq_getitem(tuple, 0);
@@ -77,12 +102,10 @@ z_policy_tuple_get_verdict(ZPolicyObj *tuple, guint *verdict)
     {
       /* policy syntax error */
       z_policy_var_unref(tmp);
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
   z_policy_var_unref(tmp);
-  z_leave();
-  return TRUE;
+  z_return(TRUE);
 }
 
 
@@ -96,7 +119,7 @@ z_policy_tuple_get_verdict(ZPolicyObj *tuple, guint *verdict)
  * The resulting list
  */
 PyObject *
-z_policy_convert_strv_to_list(const gchar **strv)
+z_policy_convert_strv_to_list(gchar const **strv)
 {
   PyObject *list;
   gint i;
@@ -301,7 +324,7 @@ z_global_getattr(const gchar *name)
  * its reference.  This function can be used to quickly parse Python values
  * this way:
  * 
- * z_policy_var_parse_str(z_global_getattr("config.vbuster.temp_directory"), &temp_directory);
+ * z_policy_var_parse_str(z_global_getattr("config.module.string_variable"), &c_variable_string);
  **/
 gboolean
 z_policy_var_parse_str(PyObject *val, const gchar **result)
@@ -322,6 +345,29 @@ z_policy_var_parse_str(PyObject *val, const gchar **result)
 }
 
 
+/**
+ * z_policy_var_parse_boolean:
+ * @val: PyObject to parse
+ * @result: result
+ *
+ * This function tries to parse @val as a boolean and puts the parsed value
+ * into @result.  It always consumes @val, thus the caller loses
+ * its reference.  This function can be used to quickly parse Python values
+ * this way:
+ * 
+ * z_policy_var_parse_int(z_global_getattr("config.module.boolean_variable"), &c_variable_boolean);
+ **/
+gboolean
+z_policy_var_parse_boolean(PyObject *val, gboolean *result)
+{
+  gboolean success;
+  gint intval = 0;
+  
+  success = z_policy_var_parse_int(val, &intval);
+  *result = intval;
+  return success;
+}
+
 /**
  * z_policy_var_parse_int:
  * @val: PyObject to parse
@@ -332,10 +378,10 @@ z_policy_var_parse_str(PyObject *val, const gchar **result)
  * its reference.  This function can be used to quickly parse Python values
  * this way:
  * 
- * z_policy_var_parse_int(z_global_getattr("config.vbuster.vdb_error_soft_fail"), &vdb_error_soft_fail);
+ * z_policy_var_parse_int(z_global_getattr("config.module.integer_variable"), &vc_variable_integer);
  **/
 gboolean
-z_policy_var_parse_int(PyObject *val, guint *result)
+z_policy_var_parse_int(PyObject *val, gint *result)
 {
   gboolean res = TRUE;
   
@@ -350,6 +396,80 @@ z_policy_var_parse_int(PyObject *val, guint *result)
   return res;
 }
 
+/**
+ * z_policy_var_parse_size:
+ * @val: PyObject to parse
+ * @result: result
+ *
+ * This function tries to parse @val as a gsize and puts the parsed value
+ * into @result.  It always consumes @val, thus the caller loses
+ * its reference.  This function can be used to quickly parse Python values
+ * this way:
+ * 
+ * z_policy_var_parse_int(z_global_getattr("config.module.gsize_variable"), &c_variable_gsize);
+ **/
+gboolean
+z_policy_var_parse_size(PyObject *val, gsize *result)
+{
+  gboolean res = TRUE;
+  
+  if (val)
+    {
+      switch (sizeof(gsize))
+        {
+        case sizeof(gchar):
+          res = z_policy_var_parse(val, "b", result);
+          break;
+
+        case sizeof(gshort):
+          res = z_policy_var_parse(val, "h", result);
+          break;
+
+        case sizeof(gint):
+          res = z_policy_var_parse(val, "i", result);
+          break;
+
+        case sizeof(gint64):
+          res = z_policy_var_parse(val, "L", result);
+          break;
+
+        default:
+          g_assert(0); /* NOTE: crash out instead of letting the error escalate */
+          break;
+        }
+      Py_XDECREF(val);
+    }
+  return res;
+}
+
+/**
+ * z_policy_var_parse_int64:
+ * @val: PyObject to parse
+ * @result: result
+ *
+ * This function tries to parse @val as an integer and puts the parsed value
+ * into @result.  It always consumes @val, thus the caller loses
+ * its reference.  This function can be used to quickly parse Python values
+ * this way:
+ * 
+ * z_policy_var_parse_int(z_global_getattr("config.module.int64_variable"), &c_variable_gint64);
+ **/
+gboolean
+z_policy_var_parse_int64(PyObject *val, gint64 *result)
+{
+  gboolean res = TRUE;
+  
+  if (val)
+    {
+      if (!z_policy_var_parse(val, "L", result))
+        {
+          res = FALSE;
+        }
+      Py_XDECREF(val);
+    }
+  return res;
+}
+
 
 /**
  * z_policy_call_object:
@@ -458,8 +578,7 @@ z_policy_call(PyObject *handler, char *name, PyObject *args, gboolean *called, g
       z_trace(NULL, "Function called; name='%s'", name);
       Py_XDECREF(attr);
     }
-  z_leave();
-  return res;
+  z_return(res);
 }
 
 
@@ -562,25 +681,21 @@ z_policy_event(PyObject *handler, char *name, PyObject *args, gchar *session_id)
 
 GStaticMutex policy_ref_lock = G_STATIC_MUTEX_INIT;
 
-struct _ZPolicy
-{
-  gint ref_cnt;
-  gchar *policy_filename;
-  ZPolicyThread *main_thread;
-};
-
 struct _ZPolicyThread
 {
   ZPolicy *policy;
   PyThreadState *thread;
   /* thread startup synchronization */
-  gboolean startable;
+  gboolean startable:1, used:1;
   GMutex   *startable_lock;
   GCond    *startable_signal;
 };
 
 GStaticPrivate policy_thread = G_STATIC_PRIVATE_INIT;
 
+static gboolean z_policy_purge(ZPolicy *self);
+#include <zorp/coredump.h>
+
 
 
 /**
@@ -632,6 +747,20 @@ z_policy_thread_acquire(ZPolicyThread *self)
   
   g_static_private_set(&policy_thread, self, NULL);
   PyEval_AcquireThread(self->thread);
+
+  /* NOTE: this is currently a warning, but it'd probably make sense to
+   * actually exclude parallel execution in the same thread by using a mutex
+   * in ZPolicyThread. However as this is a risky change at 3.1.x, x >= 14 I
+   * only added a warning here.
+   */
+  if (self->used)
+    {
+#if 0
+      z_log(NULL, CORE_ERROR, 0, "Internal error, ZPolicyThread reused, dumping core & continuing;");
+      z_coredump_create();
+#endif
+    }
+  self->used = TRUE;
 }
 
 
@@ -644,6 +773,7 @@ z_policy_thread_acquire(ZPolicyThread *self)
 void
 z_policy_thread_release(ZPolicyThread *self)
 {
+  self->used = FALSE;
   PyEval_ReleaseThread(self->thread);
   g_static_private_set(&policy_thread, NULL, NULL);
 }
@@ -738,6 +868,17 @@ z_policy_thread_destroy(ZPolicyThread *self)
       PyThreadState_Clear(self->thread);
       PyThreadState_Delete(self->thread);
       z_python_unlock();
+      z_policy_unref(self->policy);
+    }
+  else
+    {
+      /* we must be freed at last, when the policy is being destroyed */
+      g_assert(self->policy->ref_cnt == 1);
+      /* we are the main thread, destroy the interpreter */
+      z_policy_purge(self->policy);
+      PyEval_AcquireThread(self->thread);
+      Py_EndInterpreter(self->thread);
+      z_python_unlock();
     }
   g_mutex_free(self->startable_lock);
   g_cond_free(self->startable_signal);
@@ -745,6 +886,22 @@ z_policy_thread_destroy(ZPolicyThread *self)
 }
 
 
+/**
+ * z_policy_free:
+ * @self: this
+ *
+ * Destroys a ZPolicy by freeing its filename and destroying its threads.
+ */
+static void
+z_policy_free(ZPolicy *self)
+{
+  /* FIXME: we should acquire a thread state to be able to free ourselves */
+ 
+  g_free(self->policy_filename);
+  /* NOTE: This must be the last because that destroyer assume that he is the last men standing. */
+  z_policy_thread_destroy(self->main_thread);
+  g_free(self);
+}
 
 
 
@@ -761,9 +918,61 @@ z_policy_thread_destroy(ZPolicyThread *self)
 ZPolicy * 
 z_policy_ref(ZPolicy *self)
 {
+  g_static_mutex_lock(&policy_ref_lock);
+  g_assert(self->ref_cnt > 0);
+  
+  self->ref_cnt++;
+  g_static_mutex_unlock(&policy_ref_lock);
   return self;
 }
 
+/**
+ * z_policy_unref:
+ * @self: this
+ *
+ * Decrements the reference counter of a ZProxy, freeing the instance if
+ * the counter drops to zero.
+ */
+void
+z_policy_unref(ZPolicy *self)
+{
+  g_static_mutex_lock(&policy_ref_lock);
+  g_assert(self->ref_cnt > 0);
+  
+  --self->ref_cnt;
+
+  /* NOTE: This reference counter counts two different reference types:
+   *   * the number of references to this ZPolicy instance by the rest of Zorp
+   *   * the number of internal references between ZPolicy and the
+   *     ZPolicyThread instances started in z_policy_new(): main thread 
+   *     and notification thread
+   *
+   * The main and notification threads form circular references, their
+   * destruction must be triggered somehow. Whenever the number of
+   * references go low enough we trigger destruction, which will eventually
+   * cause the refcounter to go to 0.
+   *
+   * The trigger to terminate the notification thread is also sent from
+   * here, when the refcount goes to 3.
+   *
+   */   
+
+
+/* main thread */
+#define ZPOLICY_BASE_REFCOUNT 1
+
+  if (self->ref_cnt == ZPOLICY_BASE_REFCOUNT)
+    {
+      /* ok, only the notification thread & main thread remains, start destructing them now */
+      g_static_mutex_unlock(&policy_ref_lock);
+      z_policy_free(self);
+    }
+  else
+    {
+      g_static_mutex_unlock(&policy_ref_lock);
+    }
+}
+
 
 /**
  * z_policy_boot:
@@ -803,13 +1012,14 @@ z_policy_boot(ZPolicy *self)
   /* PyImport_AddModule("Zorp.Zorp"); */
   
   z_py_zorp_core_init();
-  z_py_zorp_szig_init();
-  z_py_zorp_dispatch_init();
-  z_py_zorp_attach_init();
-  z_py_zorp_stream_init();
-  z_py_zorp_proxy_init();
-  z_py_zorp_sockaddr_init();
-  z_py_zorp_fastpath_init();
+  z_policy_struct_module_init();
+  z_policy_dispatch_module_init();
+  z_policy_attach_module_init();
+  z_policy_stream_module_init();
+  z_policy_proxy_module_init();
+  z_policy_sockaddr_module_init();
+  z_policy_proxy_group_module_init();
+
 
   z_policy_thread_release(self->main_thread);
   
@@ -876,17 +1086,24 @@ z_policy_load(ZPolicy *self)
  * TRUE on success
  */
 gboolean
-z_policy_init(ZPolicy *self, const gchar **instance_name)
+z_policy_init(ZPolicy *self, gchar const **instance_name)
 {
   PyObject *main_module, *init_func, *res;
   gboolean success = FALSE;
+  cap_t saved_caps;
   
   z_policy_thread_acquire(self->main_thread);
   
   main_module = PyImport_AddModule("__main__");
   init_func = PyObject_GetAttrString(main_module, "init");
 
+  saved_caps = cap_save();
+  cap_enable(CAP_NET_ADMIN);
+
   res = PyObject_CallFunction(init_func, "(O)", z_policy_convert_strv_to_list(instance_name));
+
+  cap_restore(saved_caps);
+
   Py_XDECREF(init_func);
   if (res && z_policy_var_parse(res, "i", &success))
     {
@@ -902,6 +1119,111 @@ z_policy_init(ZPolicy *self, const gchar **instance_name)
   return success;
 }
 
+/**
+ * z_policy_deinit:
+ * @self: this
+ * @instance_name: array of instance name and aliases
+ *
+ * Deinitialises the current policy by calling ?Zorp.deinit?.
+ *
+ * Returns:
+ * TRUE on success
+ */
+gboolean
+z_policy_deinit(ZPolicy *self, gchar const **instance_name)
+{
+  PyObject *main_module, *deinit_func, *res;
+  
+  z_policy_thread_acquire(self->main_thread);
+  
+  main_module = PyImport_AddModule("__main__");
+  deinit_func = PyObject_GetAttrString(main_module, "deinit");
+
+  res = PyObject_CallFunction(deinit_func, "(O)", z_policy_convert_strv_to_list(instance_name));
+  Py_XDECREF(deinit_func);
+  if (!res)
+    {
+      PyErr_Print();
+    }
+  Py_XDECREF(res);
+  z_policy_thread_release(self->main_thread);
+
+  return res != NULL;
+}
+
+
+/**
+ * z_policy_purge:
+ * @self: this
+ *
+ * Purge the current thread context by calling ?Zorp.purge?
+ * FIXME: z_policy_thread_acquire(self->main_thread) -?-> z_policy_acquire_main(self)
+ *
+ * Returns:
+ * TRUE on success
+ */
+static gboolean
+z_policy_purge(ZPolicy *self)
+{
+  PyObject *main_module, *purge_func, *res;
+  
+  z_policy_thread_acquire(self->main_thread);
+  
+  main_module = PyImport_AddModule("__main__");
+  purge_func = PyObject_GetAttrString(main_module, "purge");
+
+  res = PyObject_CallFunction(purge_func, "()");
+  Py_XDECREF(purge_func);
+  if (!res)
+    {
+      PyErr_Print();
+    }
+  Py_XDECREF(res);
+  z_policy_thread_release(self->main_thread);
+
+  return res != NULL;
+}
+
+/**
+ * z_policy_cleanup:
+ * @self: this
+ * @instance_name: array of instance name and aliases
+ *
+ * Cleans up the current policy by calling ?Zorp.cleanup?.
+ * Currently used by KZorp to flush kernel data structures
+ * when Zorp is exiting.
+ *
+ * Returns:
+ * TRUE on success
+ */
+gboolean
+z_policy_cleanup(ZPolicy *self, gchar const **instance_name)
+{
+  PyObject *main_module, *cleanup_func, *res;
+  cap_t saved_caps;
+  
+  z_policy_thread_acquire(self->main_thread);
+  
+  main_module = PyImport_AddModule("__main__");
+  cleanup_func = PyObject_GetAttrString(main_module, "cleanup");
+
+  saved_caps = cap_save();
+  cap_enable(CAP_NET_ADMIN);
+
+  res = PyObject_CallFunction(cleanup_func, "(O)", z_policy_convert_strv_to_list(instance_name));
+
+  cap_restore(saved_caps);
+
+  Py_XDECREF(cleanup_func);
+  if (!res)
+    {
+      PyErr_Print();
+    }
+  Py_XDECREF(res);
+  z_policy_thread_release(self->main_thread);
+
+  return res != NULL;
+}
 
 /**
  * z_policy_acquire_main:
@@ -933,7 +1255,9 @@ z_policy_release_main(ZPolicy *self)
  * z_policy_new:
  * @filename: Name of the policy file
  *
- * Constructor of ZPolicy, creates a new instance and starts its main thread.
+ * Constructor of ZPolicy, creates a new instance and starts its main thread
+ * and notification thread. The policy's refcount is 3: one per thread
+ * plus one for the caller.
  *
  * Returns:
  * The new instance
@@ -950,8 +1274,8 @@ z_policy_new(const gchar *filename)
   z_python_unlock();
   z_policy_thread_ready(self->main_thread);
   
-  /* the main thread always references us, and we should be deleted when
-   * that reference is dropped */
+  /* the main thread and the notification thread always references us,
+   * and we should be deleted when that reference is dropped */
   
   return self;
 }
