@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -36,46 +37,51 @@
 #include <zorp/log.h>
 #include <zorp/stream.h>
 #include <zorp/streamfd.h>
+#include <zorp/streamline.h>
 #include <zorp/policy.h>
 
-PyObject *z_py_stream_new(ZStream *Stream);
-static PyObject *z_py_zorp_stream_new(PyObject *self, PyObject *args);
-static void z_py_zorp_stream_destroy(PyObject *o);
-static PyObject *z_py_zorp_stream_getattr(PyObject *o, char *name);
-static gint z_py_zorp_stream_setattr(PyObject *o, char *name,
+PyObject *z_policy_stream_new(ZStream *Stream);
+static PyObject *z_policy_stream_new_instance(PyObject *self, PyObject *args);
+static void z_policy_stream_destroy(PyObject *o);
+static PyObject *z_policy_stream_getattr(PyObject *o, char *name);
+static gint z_policy_stream_setattr(PyObject *o, char *name,
 				     PyObject *value);
-static PyObject *z_py_zorp_stream_repr(PyObject *o);
-static PyObject *z_py_zorp_stream_read(PyObject *o, PyObject *args);
-static PyObject *z_py_zorp_stream_write(PyObject *o, PyObject *args);
-static PyObject *z_py_zorp_stream_close(PyObject *o, PyObject *args);
+static PyObject *z_policy_stream_repr(PyObject *o);
+static PyObject *z_policy_stream_read(PyObject *o, PyObject *args);
+static PyObject *z_policy_stream_write(PyObject *o, PyObject *args);
+static PyObject *z_policy_stream_close(PyObject *o, PyObject *args);
+static PyObject *z_policy_stream_readline(PyObject *o, PyObject *args);
 
-PyMethodDef z_py_zorp_stream_funcs[] =
+static PyObject *z_policy_stream_exception = NULL; 
+
+PyMethodDef z_policy_stream_funcs[] =
 {
-  { "Stream",  z_py_zorp_stream_new, METH_VARARGS, NULL },
+  { "Stream",  z_policy_stream_new_instance, METH_VARARGS, NULL },
   { NULL,          NULL, 0, NULL }   /* sentinel*/
 };
 
 static PyMethodDef py_zorp_stream_methods[] =
 {
-  { "read",        z_py_zorp_stream_read, METH_VARARGS, NULL },
-  { "write",       z_py_zorp_stream_write, METH_VARARGS, NULL },
-  { "close",       (PyCFunction) z_py_zorp_stream_close, 0, NULL },
+  { "read",        z_policy_stream_read, METH_VARARGS, NULL },
+  { "write",       z_policy_stream_write, METH_VARARGS, NULL },
+  { "close",       (PyCFunction) z_policy_stream_close, 0, NULL },
+  { "readline",        z_policy_stream_readline, METH_VARARGS, NULL },
   { NULL,          NULL, 0, NULL }   /* sentinel*/
 };
 
-PyTypeObject z_py_zorp_stream_type = 
+PyTypeObject z_policy_stream_type = 
 {
   PyObject_HEAD_INIT(&PyType_Type)
   0,
-  "ZorpStream",
-  sizeof(ZorpStream),
+  "ZPolicyStream",
+  sizeof(ZPolicyStream),
   0,
-  (destructor) z_py_zorp_stream_destroy,
+  (destructor) z_policy_stream_destroy,
   0,
-  (getattrfunc) z_py_zorp_stream_getattr,
-  (setattrfunc) z_py_zorp_stream_setattr,
+  (getattrfunc) z_policy_stream_getattr,
+  (setattrfunc) z_policy_stream_setattr,
   0,
-  (reprfunc) z_py_zorp_stream_repr,
+  (reprfunc) z_policy_stream_repr,
   0,
   0,
   0,
@@ -86,13 +92,13 @@ PyTypeObject z_py_zorp_stream_type =
   0,
   0,
   0,
-  "ZorpStream class for Zorp",
+  "ZPolicyStream class for Zorp",
   0, 0, 0, 0,
   Z_PYTYPE_TRAILER
 };
 
 /**
- * z_py_stream_new:
+ * z_policy_stream_new:
  * @str: ZStream instance
  *
  * This function allocates a Python object representing the Zorp stream
@@ -102,27 +108,27 @@ PyTypeObject z_py_zorp_stream_type =
  * Returns: the newly allocated Python object
  **/
 PyObject *
-z_py_stream_new(ZStream *str)
+z_policy_stream_new(ZStream *str)
 {
-  ZorpStream *self;
+  ZPolicyStream *self;
 
   if (str == NULL)
     {
       /*LOG
         This message indicates an internal error, please contact your Zorp support for assistance.
        */
-      z_log(NULL, CORE_ERROR, 3, "Internal error in z_py_zorp_stream_new: input ZStream is NULL;");
+      z_log(NULL, CORE_ERROR, 3, "Internal error in z_policy_stream_new: input ZStream is NULL;");
       return NULL;
     }
 
-  self = PyObject_NEW(ZorpStream, &z_py_zorp_stream_type);
+  self = PyObject_New(ZPolicyStream, &z_policy_stream_type);
   z_stream_ref(str);
   self->stream = str;
   return (PyObject *) self;
 }
 
 /**
- * z_py_zorp_stream_new:
+ * z_policy_stream_new:
  * @o: Python self argument
  * @args: Python args argument
  *
@@ -134,15 +140,15 @@ z_py_stream_new(ZStream *str)
  * Returns: the newly allocated Python object
  **/
 static PyObject *
-z_py_zorp_stream_new(PyObject *o G_GNUC_UNUSED, PyObject *args)
+z_policy_stream_new_instance(PyObject *o G_GNUC_UNUSED, PyObject *args)
 {
-  ZorpStream *self;
+  ZPolicyStream *self;
   char *name;
   int fd;
 
   if (!PyArg_ParseTuple(args, "is", &fd, &name))
     return NULL;
-  self = PyObject_NEW(ZorpStream, &z_py_zorp_stream_type);
+  self = PyObject_New(ZPolicyStream, &z_policy_stream_type);
   if (!self)
     return NULL;
 
@@ -151,36 +157,36 @@ z_py_zorp_stream_new(PyObject *o G_GNUC_UNUSED, PyObject *args)
 }
 
 /**
- * z_py_zorp_stream_destroy:
+ * z_policy_stream_destroy:
  * @o: Python self argument
  *
- * This function is registered as the free function for ZorpStream, thus it is
- * called when a ZorpStream is to be freed.
+ * This function is registered as the free function for ZPolicyStream, thus it is
+ * called when a ZPolicyStream is to be freed.
  **/
 static void
-z_py_zorp_stream_destroy(PyObject *o)
+z_policy_stream_destroy(PyObject *o)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
 
   z_stream_unref(self->stream);
   PyObject_Del(self);
 }
 
 /**
- * z_py_zorp_stream_getattr:
+ * z_policy_stream_getattr:
  * @o: Python self argument
  * @name: attribute to return
  *
- * This function is called when an attribute from the ZorpStream object is
+ * This function is called when an attribute from the ZPolicyStream object is
  * read, the support for various builtin attributes (fd, name etc.) is
  * implemented here.
  *
  * Returns: the Python object referring to the attribute
  **/
 static PyObject *
-z_py_zorp_stream_getattr(PyObject *o, char *name)
+z_policy_stream_getattr(PyObject *o, char *name)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
   if (strcmp(name, "fd") == 0)
     {
       return z_policy_var_build("i", z_stream_get_fd(self->stream));
@@ -197,27 +203,43 @@ z_py_zorp_stream_getattr(PyObject *o, char *name)
     {
       return PyLong_FromUnsignedLong(self->stream->bytes_sent);
     }
+  else if (strcmp(name, "nul_nonfatal") == 0)
+    {
+      gboolean value;
+      z_stream_ctrl(self->stream, ZST_LINE_GET_NUL_NONFATAL,  &value, sizeof(gboolean));
+      return Py_BuildValue("i", !!value);
+    }
+  else if (strcmp(name, "split") == 0)
+    {
+      gboolean value;
+      z_stream_ctrl(self->stream, ZST_LINE_GET_SPLIT,  &value, sizeof(gboolean));
+      return Py_BuildValue("i", value);
+    }
+  else if (strcmp(name, "keepalive") == 0)
+    {
+      return PyLong_FromLong(z_stream_get_keepalive(self->stream));
+    }
 
   return Py_FindMethod(py_zorp_stream_methods, o, name);
 }
 
 /**
- * z_py_zorp_stream_setattr:
+ * z_policy_stream_setattr:
  * @o: Python self argument
  * @name: attribute to set
  * @value: new value for attribute
  *
- * This function is called when an attribute from the ZorpStream object is
+ * This function is called when an attribute from the ZPolicyStream object is
  * written to, the support for various builtin attributes (fd, name etc.) is
  * implemented here.
  *
  * Returns: 1 to indicate failure, 0 for success
  **/
 static gint 
-z_py_zorp_stream_setattr(PyObject *o, char *name,
+z_policy_stream_setattr(PyObject *o, char *name,
 			 PyObject *value)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
   gchar *str;
   if (strcmp(name, "name") == 0)
     {
@@ -228,32 +250,102 @@ z_py_zorp_stream_setattr(PyObject *o, char *name,
 	}
       else
 	{
-	  g_strlcpy(self->stream->name, str, Z_STREAM_MAX_NAME);
+	  z_stream_set_name(self->stream, str);
 	  return 0;
 	}
     }
+  else if (strcmp(name, "nul_nonfatal") == 0)
+    {
+      int cval;
+      if (!PyArg_Parse(value, "i", &cval))
+	{
+	  PyErr_SetString(PyExc_TypeError, "nul_nonfatal is boolean");
+	  return 1;
+	}
+      z_stream_ctrl(self->stream, ZST_LINE_SET_NUL_NONFATAL,  &cval, sizeof(int));
+      return 0;
+    }
+  else if (strcmp(name, "split") == 0)
+    {
+      int cval;
+      if (!PyArg_Parse(value, "i", &cval))
+	{
+	  PyErr_SetString(PyExc_TypeError, "split is boolean");
+	  return 1;
+	}
+      z_stream_ctrl(self->stream, ZST_LINE_SET_SPLIT,  &cval, sizeof(int));
+      return 0;
+    }
+  else if (strcmp(name, "keepalive") == 0)
+    {
+      gint keepalive; 
+      if (!PyArg_Parse(value, "i", &keepalive))
+        {
+          PyErr_SetString(PyExc_TypeError, "Stream keepalive value is not an integer");
+          return 1;
+        }
+      else
+        {
+          z_stream_set_keepalive(self->stream, keepalive);
+          return 0;
+        }
+    }
 
   PyErr_SetString(PyExc_AttributeError, "No such attribute");
   return 1;
 }
 
 /**
- * z_py_zorp_stream_repr:
- * @o: ZorpStream object
+ * z_policy_stream_repr:
+ * @o: ZPolicyStream object
  *
- * __repr__ function for ZorpStream objects
+ * __repr__ function for ZPolicyStream objects
  **/
 static PyObject *
-z_py_zorp_stream_repr(PyObject *o)
+z_policy_stream_repr(PyObject *o)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
 
   return PyString_FromString(self->stream->name);
 }
 
+
 /**
- * z_py_zorp_stream_read:
- * @o: Python self, ZorpStream object
+ * z_policy_stream_readline
+ * @o: Python self, ZPolicyStream object
+ * @args: Python args argument
+ *
+ * readline method exported to Python with this declaration:
+ *   def readline(self)
+ *
+ * gets a line from the stream
+ */
+ 
+static PyObject *
+z_policy_stream_readline(PyObject *o, PyObject *args G_GNUC_UNUSED)
+{
+  ZPolicyStream *self = (ZPolicyStream *) o;
+  gchar *buf;
+  PyObject *pybuf;
+  gsize bytes_read;
+  gint res;
+  
+  Py_BEGIN_ALLOW_THREADS
+  res = z_stream_line_get(self->stream, &buf, &bytes_read, NULL);
+  Py_END_ALLOW_THREADS
+  if (res == G_IO_STATUS_NORMAL)
+    {
+      pybuf = Py_BuildValue("s#", buf, bytes_read);
+      return pybuf;
+    }
+  PyErr_SetObject(z_policy_stream_exception, Py_BuildValue("(i,O)", res, Py_None));
+  return NULL;
+}
+
+
+/**
+ * z_policy_stream_read:
+ * @o: Python self, ZPolicyStream object
  * @args: Python args argument
  *
  * read method exported to Python with this declaration: 
@@ -262,9 +354,10 @@ z_py_zorp_stream_repr(PyObject *o)
  * the length argument specifies how many bytes need to be read.
  **/
 static PyObject *
-z_py_zorp_stream_read(PyObject *o, PyObject *args)
+z_policy_stream_read(PyObject *o, PyObject *args)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
+  PyObject *pybuf;
   gchar *buf;
   guint length;
   gsize bytes_read;
@@ -273,22 +366,24 @@ z_py_zorp_stream_read(PyObject *o, PyObject *args)
   if (!PyArg_ParseTuple(args, "i", &length))
     return NULL;
 
-  buf = alloca(length);
+  buf = g_new0(char, length);
   Py_BEGIN_ALLOW_THREADS
   res = z_stream_read(self->stream, buf, length, &bytes_read, NULL);
   Py_END_ALLOW_THREADS
-  
-  if (res != G_IO_STATUS_NORMAL)
+  if (res == G_IO_STATUS_NORMAL)
     {
-      PyErr_SetString(PyExc_IOError, "I/O error reading stream.");
-      return NULL;      
+      pybuf = Py_BuildValue("s#", buf, bytes_read);
+      g_free(buf);
+      return pybuf;
     }
-  return Py_BuildValue("s#", buf, bytes_read);
+  g_free(buf);
+  PyErr_SetObject(z_policy_stream_exception, Py_BuildValue("(i,O)", res, Py_None));
+  return NULL;
 }
 
 /**
- * z_py_zorp_stream_write:
- * @o: Python self, ZorpStream object
+ * z_policy_stream_write:
+ * @o: Python self, ZPolicyStream object
  * @args: Python args argument
  *
  * read method exported to Python with this declaration: 
@@ -298,9 +393,9 @@ z_py_zorp_stream_read(PyObject *o, PyObject *args)
  * be written.
  **/
 static PyObject *
-z_py_zorp_stream_write(PyObject *o, PyObject *args)
+z_policy_stream_write(PyObject *o, PyObject *args)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
   gchar *buf; 
   guint length;
   gsize bytes_written;
@@ -325,16 +420,16 @@ z_py_zorp_stream_write(PyObject *o, PyObject *args)
 
 
 /**
- * z_py_zorp_stream_close:
- * @o: Python self argument, ZorpStream object
+ * z_policy_stream_close:
+ * @o: Python self argument, ZPolicyStream object
  * @args: Python args argument
  *
  * Close method exported to Python. 
  **/
 static PyObject *
-z_py_zorp_stream_close(PyObject *o, PyObject *args G_GNUC_UNUSED)
+z_policy_stream_close(PyObject *o, PyObject *args G_GNUC_UNUSED)
 {
-  ZorpStream *self = (ZorpStream *) o;
+  ZPolicyStream *self = (ZPolicyStream *) o;
 
   z_stream_close(self->stream, NULL);
   Py_XINCREF(Py_None);
@@ -342,14 +437,20 @@ z_py_zorp_stream_close(PyObject *o, PyObject *args G_GNUC_UNUSED)
 }
 
 /**
- * z_py_zorp_stream_init:
+ * z_policy_stream_init:
  * 
- * This function is called at Python initialization to export ZorpStream
+ * This function is called at Python initialization to export ZPolicyStream
  * related functions.
  **/
 void
-z_py_zorp_stream_init(void)
+z_policy_stream_module_init(void)
 {
+  PyObject* module;
+
   PyImport_AddModule("Zorp.Stream");
-  Py_InitModule("Zorp.Stream", z_py_zorp_stream_funcs);
+  module = Py_InitModule("Zorp.Stream", z_policy_stream_funcs);
+  
+  z_policy_stream_exception = PyErr_NewException("Zorp.Stream.StreamException", NULL, NULL);
+  Py_INCREF(z_policy_stream_exception);
+  PyModule_AddObject(module, "StreamException", z_policy_stream_exception);
 }
