@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -29,46 +30,40 @@
 
 #include <zorp/zorp.h>
 #include <zorp/poll.h>
-#include <zorp/proxy.h>
+#include <zorp/proxystack.h>
 
 typedef struct _ZPlugSession ZPlugSession;
-typedef gboolean (*ZPlugSessionStatsFunc)(ZProxy *proxy, ZSessionVars *vars, 
-					  guint64 client_bytes, guint64 client_pkts, 
-					  guint64 server_bytes, guint64 server_pkts);
 
 typedef struct _ZPlugSessionData
 {
-  guint num_sessions;
-  GList *destroy_queue;
-
   gint timeout;
   gboolean copy_to_server, copy_to_client;
   gboolean shutdown_soft;
   guint buffer_size;
   guint packet_stats_interval_time, packet_stats_interval_packet;
-  ZPlugSessionStatsFunc packet_stats;
+  
+  gboolean (*packet_stats)(ZPlugSession *self, 
+                           guint64 client_bytes, guint64 client_pkts, 
+                           guint64 server_bytes, guint64 server_pkts,
+                           gpointer user_data);
+  void (*finish)(ZPlugSession *self, gpointer user_data);
 } ZPlugSessionData;
 
-gboolean
-z_plug_session_start(ZPlugSession *self, ZPoll *poll);
+gboolean z_plug_session_start(ZPlugSession *self, ZPoll *poll);
+void z_plug_session_cancel(ZPlugSession *self);
+void z_plug_session_register_vars(ZPlugSession *self, ZPolicyDict *dict);
 
 ZPlugSession *
-z_plug_session_new(ZProxy *owner, 
-                   ZPlugSessionData *session_data, 
+z_plug_session_new(ZPlugSessionData *session_data, 
                    ZStream *client_stream, 
                    ZStream *server_stream, 
-                   ZStackedProxy *stacked);
+                   ZStackedProxy *stacked,
+                   gpointer user_data);
 
-void
-z_plug_session_free(ZPlugSession *self);
+void z_plug_session_destroy(ZPlugSession *self);
 
-static inline gboolean
-z_plug_sessions_running(ZPlugSessionData *session_data)
-{
-  return session_data->num_sessions > 0;
-}
+ZPlugSession *z_plug_session_ref(ZPlugSession *self);
+void z_plug_session_unref(ZPlugSession *self);
 
-void
-z_plug_sessions_purge(ZPlugSessionData *session_data);
 
 #endif
