@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -30,7 +31,7 @@
 
 #include <zorp/zpython.h>
 
-void z_py_zorp_attach_init(void);
+void z_policy_attach_module_init(void);
 
 #endif
 
