@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -29,15 +30,21 @@
 #define ZORP_PYDISPATCH_H_INCLUDED
 
 #include <zorp/zpython.h>
+#include <zorp/pystruct.h>
+#include <zorp/dispatch.h>
 
-/*+
+typedef struct _ZPolicyDispatchBind ZPolicyDispatchBind;
 
-  ZorpDispatch is the python interface to ZDispatch
+ZDispatchBind *
+z_policy_dispatch_bind_get_db(ZPolicyObj *self);
 
-  +*/
-typedef struct _ZorpDispatch ZorpDispatch;
+static inline gboolean
+z_policy_dispatch_bind_check(ZPolicyObj *self)
+{
+  return z_policy_struct_check(self, Z_PST_DB_SOCKADDR) || z_policy_struct_check(self, Z_PST_DB_IFACE) || z_policy_struct_check(self, Z_PST_DB_IFACE_GROUP);
+}
 
-void z_py_zorp_dispatch_init(void);
+void z_policy_dispatch_module_init(void);
 
 #endif
 
