@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -30,22 +31,22 @@
 
 #include <zorp/zpython.h>
 #include <zorp/thread.h>
+#include <zorp/proxy.h>
 
-struct _ZProxy;
+typedef struct _ZPolicyProxy ZPolicyProxy;
 
-/*+
-  ZorpProxy is the Python interface to ZProxy.
-  +*/
-typedef struct _ZorpProxy
+extern PyTypeObject z_policy_proxy_type;
+
+gboolean z_policy_proxy_bind_implementation(PyObject *self);
+ZProxy *z_policy_proxy_get_proxy(PyObject *obj);
+
+void z_policy_proxy_module_init(void);
+
+static inline gboolean 
+z_policy_proxy_check(ZPolicyObj *s)
 {
-  PyObject_HEAD
-//  PyObject *name;
-  struct _ZProxy *proxy;
-} ZorpProxy;
-
-gboolean z_py_zorp_proxy_check(PyObject *obj);
-struct _ZProxy *z_py_zorp_proxy_get_proxy(PyObject *obj);
-
+  return PyObject_TypeCheck(s, &z_policy_proxy_type);
+}
 
 #endif
 
