@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -28,26 +29,19 @@
 #ifndef ZORP_PYSOCKADDR_H_INCLUDED
 #define ZORP_PYSOCKADDR_H_INCLUDED
 
-#include <zorp/zpython.h>
+#include <zorp/pystruct.h>
 #include <zorp/sockaddr.h>
 
-/*+
+ZPolicyObj *z_policy_sockaddr_new(ZSockAddr *sa);
+ZSockAddr *z_policy_sockaddr_get_sa(ZPolicyObj *s);
 
-  ZorpSockAddr is the Python interface to ZSockAddr
-
-  +*/
-typedef struct _ZorpSockAddr
+static inline gboolean
+z_policy_sockaddr_check(ZPolicyObj *s)
 {
-  PyObject_HEAD
-  ZSockAddr *sa;
-} ZorpSockAddr;
+  return z_policy_struct_check(s, Z_PST_SOCKADDR_INET) || z_policy_struct_check(s, Z_PST_SOCKADDR_UNIX) || z_policy_struct_check(s, Z_PST_SOCKADDR_INET6);
+}
 
-extern PyTypeObject z_py_zorp_sockaddr_type;
-
-PyObject *z_py_zorp_sockaddr_new(ZSockAddr *sa);
-#define z_py_zorp_sockaddr_check(ob) ((ob)->ob_type == &z_py_zorp_sockaddr_type)
-
-void z_py_zorp_sockaddr_init(void);
+void z_policy_sockaddr_module_init(void);
 
 #endif
 
