@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -33,23 +34,31 @@
 
 /*+
 
-  ZorpStream is the Python interface to ZStream.
+  ZPolicyStream is the Python interface to ZStream.
 
   +*/
-typedef struct _ZorpStream
+typedef struct _ZPolicyStream
 {
   PyObject_HEAD
   ZStream *stream;
-} ZorpStream;
+} ZPolicyStream;
 
-extern PyTypeObject z_py_zorp_stream_type;
+extern PyTypeObject z_policy_stream_type;
 
-#define z_py_zorp_stream_check(ob) ((ob)->ob_type == &z_py_zorp_stream_type)
+#define z_policy_stream_check(ob) ((ob)->ob_type == &z_policy_stream_type)
 
-void z_py_zorp_stream_init(void);
-void z_py_zorp_proxy_init(void);
+void z_policy_stream_module_init(void);
 
-PyObject *z_py_stream_new(ZStream *Stream);
+PyObject *z_policy_stream_new(ZStream *Stream);
+
+static inline ZStream *
+z_policy_stream_get_stream(PyObject *s)
+{
+  ZPolicyStream *self = (ZPolicyStream *) s;
+  
+  g_assert(z_policy_stream_check(s));
+  return z_stream_ref(self->stream);
+}
 
 #endif
 
