@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -30,23 +31,64 @@
 
 #include <zorp/zorp.h>
 
-#define Z_SZIG_THREAD_START 0
-#define Z_SZIG_THREAD_STOP  1
-#define Z_SZIG_MIN1_TICK    2
-#define Z_SZIG_MIN5_TICK    3
-#define Z_SZIG_MIN15_TICK   4
-#define Z_SZIG_COUNTED_IP   5
+/* these values are copied to Python, change carefully */
+enum
+{
+  Z_SZIG_THREAD_START = 0,
+  Z_SZIG_THREAD_STOP  = 1,
+  Z_SZIG_MIN1_TICK    = 2,
+  Z_SZIG_MIN5_TICK    = 3,
+  Z_SZIG_MIN15_TICK   = 4,
+  Z_SZIG_COUNTED_IP   = 5,
+  Z_SZIG_SERVICE_PROPS= 6,
+  Z_SZIG_SERVICE_STOP = 7,
+  Z_SZIG_AUDIT_START  = 8,
+  Z_SZIG_AUDIT_STOP   = 9,
+  Z_SZIG_RELOAD       = 10,
+  Z_SZIG_MAX
+};
 
-#define Z_SZIG_MAX          5
+/* these values are copied to Python, change carefully */
+enum
+{
+  Z_SZIG_TYPE_NOTINIT = 0,
+  Z_SZIG_TYPE_LONG,
+  Z_SZIG_TYPE_TIME,
+  Z_SZIG_TYPE_STRING,
+  Z_SZIG_TYPE_PROPS,
+  Z_SZIG_TYPE_SERVICE_PROPS
+};
 
-#define Z_SZIG_TYPE_NOTINIT 0
-#define Z_SZIG_TYPE_LONG    1
-#define Z_SZIG_TYPE_TIME    2
-#define Z_SZIG_TYPE_STRING  3
+#define Z_SZIG_MAX_PROPS 16
 
 typedef gint ZSzigEvent;
 typedef struct _ZSzigValue ZSzigValue;
 typedef struct _ZSzigNode ZSzigNode;
+typedef struct _ZSzigProps ZSzigProps;
+typedef struct _ZSzigServiceProps ZSzigServiceProps;
+
+/*
+ * NOTE: this could be represented as a nested structure of ZSzigProps,
+ * however it is special cased for speed.
+ */
+struct _ZSzigServiceProps
+{
+  gchar *name;
+  gint instance_id;
+  gushort sec_conn_id;
+  gushort related_id; 
+  gint string_count;
+  gchar *string_list[Z_SZIG_MAX_PROPS * 2];
+};
+
+struct _ZSzigProps
+{
+  gchar *name;
+  gint value_count;
+  gchar *name_list[Z_SZIG_MAX_PROPS];
+  ZSzigValue *value_list[Z_SZIG_MAX_PROPS];
+};
+
 
 struct _ZSzigValue
 {
@@ -56,9 +98,13 @@ struct _ZSzigValue
     glong long_value;
     GTimeVal time_value;
     GString *string_value;
+    ZSzigProps props_value;
+    ZSzigServiceProps service_props;
   } u;
 };
 
+
+
 typedef void (*ZSzigEventHandler)(ZSzigNode *node, ZSzigEvent ev, ZSzigValue *param, gpointer user_data);
 
 void z_szig_register_handler(ZSzigEvent ev, ZSzigEventHandler func, const gchar *node_name, gpointer user_data);
@@ -71,6 +117,10 @@ void z_szig_init(const gchar *instance_name);
 ZSzigValue *z_szig_value_new_long(glong val);
 ZSzigValue *z_szig_value_new_time(GTimeVal *val);
 ZSzigValue *z_szig_value_new_string(const gchar *val);
-void z_szig_value_free(ZSzigValue *v);
+void z_szig_value_add_service_prop(ZSzigValue *v, const gchar *name, const gchar *value);
+ZSzigValue *z_szig_value_new_service_props(const gchar *service, gint instance_id, gushort sec_conn_id, gushort related_id, const gchar *name, ...);
+void z_szig_value_add_prop(ZSzigValue *v, const gchar *name, ZSzigValue *value);
+ZSzigValue *z_szig_value_new_props(const gchar *name, const gchar *first_prop, ...);
+void z_szig_value_free(ZSzigValue *v, gboolean free_inst);
 
 #endif
