@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -36,37 +37,55 @@
 #include <zorp/misc.h>
 #include <zorp/memtrace.h>
 
-/* return values of callbacks are composed as follows:
- *
- * 0. byte: Error condition
- * 1. byte: action to take
- * 2. byte: undefined
- * 3. byte: undefined
- */
-
 
 #define ZORP_POLICY_FILE	ZORP_SYSCONFDIR "/policy.py"
 #define ZORP_POLICY_BOOT_FILE	ZORP_DATADIR "/policy.boot"
 #define ZORP_AUTH_CERT_FILE	ZORP_SYSCONFDIR "/zorp.crt"
 #define ZORP_AUTH_KEY_FILE	ZORP_SYSCONFDIR "/zorp.key"
-#define ZORP_PID_FILE_DIR       "/var/run/zorp"
+#define ZORP_STATE_DIR          ZORP_STATEDIR
+#define ZORP_PID_FILE_DIR       ZORP_PIDFILEDIR
+#define ZORP_SZIG_SOCKET_NAME      ZORP_PID_FILE_DIR "zorpctl"
 
 #define MAX_SESSION_ID		128
+#define DEADLOCK_CHECKER_DEFAULT_TIMEOUT  60
 
 #define CORE_POLICY  "core.policy"
+#define CORE_AUDIT   "core.audit"
+
+#if SIZEOF_VOID_P == 4
+#   define G_GPOINTER_FORMAT "08" G_GSIZE_MODIFIER "x"
+#elif SIZEOF_VOID_P == 8
+#   define G_GPOINTER_FORMAT "016" G_GSIZE_MODIFIER "x"
+#else
+#   error "Can't find suitable printf format for pointers"
+#endif
 
 extern GMainLoop *main_loop;
 extern gint exit_code;
 extern gboolean usr1_received;
 extern gboolean usr2_received;
+extern guint32 startup_id;
+extern const gchar *instance_name;
+
 
 void z_main_loop_initiate_reload(gboolean called_from_sighandler);
 gboolean z_main_loop_get_last_reload_result(void);
 
 void z_main_loop_initiate_termination(gboolean called_from_sighandler);
 
-void z_main_loop(const gchar *policy_file, const gchar *instance_name, const gchar **instance_policy_list);
+void z_main_loop(const gchar *policy_file, const gchar *instance_name, gchar const **instance_policy_list);
 void z_main_loop_quit(int exit_code);
+void z_main_loop_init(void);
+void z_main_loop_destroy(void);
+
 void z_log_set_fake_session_id(const gchar *instance_name G_GNUC_UNUSED);
 
+#if GLIB_MINOR_VERSION < 8
+# define G_GNUC_NULL_TERMINATED
+#endif
+
+#if GLIB_MINOR_VERSION < 10
+# define G_GNUC_WARN_UNUSED_RESULT
+#endif
+
 #endif
