@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -40,11 +41,6 @@
 
 /* ZDotTransfer implementation */
 
-#define DOT_DW_PREAMBLE  0
-#define DOT_DW_DATA      1
-#define DOT_DW_DATA_LF   2
-#define DOT_DW_DATA_DOT  3
-
 extern ZClass ZDotTransfer__class;
 
 /**
@@ -299,7 +295,6 @@ z_dot_transfer_dst_shutdown(ZTransfer2 *s, ZStream *stream, GError **err)
  * @poll: ZPoll instance
  * @client: client stream
  * @server: server stream
- * @stacked: stacked proxy 
  * @buffer_size: buffer size
  * @timeout: inactivity timeout
  * @flags: bit combination of ZT2F_*
@@ -314,7 +309,6 @@ z_dot_transfer_new(ZClass *class,
                    ZProxy *owner,
                    ZPoll *poll,
                    ZStream *client, ZStream *server,
-                   ZStackedProxy *stacked,
                    gsize buffer_size,
                    glong timeout,
                    gulong flags,
@@ -333,8 +327,7 @@ z_dot_transfer_new(ZClass *class,
                 ZDotTransfer);
 
   self->preamble = preamble;
-  self->super.stacked = stacked;
-
+  z_proxy_leave(owner);
   return self;
 }
 
