@@ -4,6 +4,141 @@
 
 #define MAX_READ_AT_A_TIME 30
 
+typedef struct _ZTransfer2PSIface
+{
+  ZProxyStackIface super;
+  ZTransfer2 *transfer;
+} ZTransfer2PSIface;
+
+extern ZClass ZTransfer2PSIface__class;
+
+static inline void
+z_transfer2_update_status(ZTransfer2 *self, guint32 status_bit, gint enable)
+{
+  guint32 old_mask = self->status & ZT2S_EOF_BITS;
+  if (enable)
+    self->status |= status_bit;
+  else
+    self->status &= ~status_bit;
+  
+  /*LOG
+    This message reports that the data-transfer to or from some endpoint is closed.
+   */
+  z_proxy_log(self->owner, CORE_DEBUG, 7, "Eofmask is updated; old_mask='%04x', eof_mask='%04x'", old_mask, self->status & ZT2S_EOF_BITS);
+}
+
+/**
+ * z_transfer2_ps_iface_set_stacked_verdict:
+ * @self: ZProxyStackIface instance
+ * @verdict: verdict sent by the child proxy
+ * @description: additional information about @decision
+ *
+ * This function is the virtual set_result method in the ZProxyStackIface
+ * interface callable by child proxies. It simply stores the verdict sent
+ * by the child proxy in self->stack_decision.
+ *
+ * NOTE: this function runs in the thread of the child proxy and there is no
+ * synchronization between this and the main proxy thread except for the fact
+ * that the child proxy should call this function before he sends anything
+ * and HTTP will make use of this value when it first receives something,
+ * thus there must be at least one context switch in between, and the HTTP
+ * is possibly sleeping when this function is called.
+ **/
+static void
+z_transfer2_ps_iface_set_stacked_verdict(ZProxyStackIface *s, ZVerdict verdict, const gchar *description)
+{
+  ZTransfer2PSIface *self = Z_CAST(s, ZTransfer2PSIface);
+
+  g_string_assign(self->transfer->stack_info, description ? description : "");
+  self->transfer->stack_decision = verdict;
+}
+
+/**
+ * http_set_content_length:
+ * @self: ZProxyStackIface instance
+ * @content_length: the length of the content to be transferred
+ *
+ * This function is the virtual set_content_length method in the
+ * ZProxyStackIface interface callable by child proxies. It simply stores
+ * the future size of the transferred data blob. This information will then
+ * be used by the transfer code to decide whether to include a
+ * content-length header. Calling this function is not mandatory by the
+ * child proxy, if it is not called chunked mode transfer-encoding will be
+ * used.
+ */
+static void
+z_transfer2_ps_iface_set_content_hint(ZProxyStackIface *s, gint64 content_length)
+{
+  ZTransfer2PSIface *self = Z_CAST(s, ZTransfer2PSIface);
+
+  self->transfer->child_content_length_hint = content_length;
+  self->transfer->child_content_length_hint_set = TRUE;
+}
+
+/**
+ * http_get_content_hint:
+ * @self: ZProxyStackIface instance
+ * @content_length: the length of the content to be transferred
+ *
+ * This function is the virtual get_content_length method in the
+ * ZProxyStackIface interface callable by child proxies.
+ */
+static gboolean
+z_transfer2_ps_iface_get_content_hint(ZProxyStackIface *s, gint64 *content_length, const gchar **content_format)
+{
+  ZTransfer2PSIface *self = Z_CAST(s, ZTransfer2PSIface);
+
+  /* Note: Save it with a mutex because it's only be set in z_transfer2_setup in
+   * other thread than this called.
+   */
+  g_mutex_lock(self->transfer->startup_lock);
+  *content_format = self->transfer->content_format;
+  if (self->transfer->our_content_length_hint_set)
+    *content_length = self->transfer->our_content_length_hint;
+  else
+    *content_length = -1;
+  g_mutex_unlock(self->transfer->startup_lock);
+
+  return TRUE;
+}
+
+
+static ZProxyIface *
+z_transfer2_ps_iface_new(ZTransfer2 *transfer)
+{
+  ZTransfer2PSIface *self;
+  
+  self = Z_CAST(z_proxy_stack_iface_new(Z_CLASS(ZTransfer2PSIface), transfer->owner), ZTransfer2PSIface);
+  self->transfer = transfer;
+  return &self->super;
+}
+
+static void
+z_transfer2_ps_iface_free(ZObject *s)
+{
+  z_proxy_stack_iface_free_method(s);
+}
+
+ZProxyStackIfaceFuncs z_transfer2_ps_iface_funcs =
+{
+  {
+    Z_FUNCS_COUNT(ZProxyStackIface),
+    .free_fn = z_transfer2_ps_iface_free,
+  },
+  .set_verdict = z_transfer2_ps_iface_set_stacked_verdict,
+  .set_content_hint = z_transfer2_ps_iface_set_content_hint,
+  .get_content_hint = z_transfer2_ps_iface_get_content_hint
+};
+
+ZClass ZTransfer2PSIface__class =
+{
+  Z_CLASS_HEADER,
+  Z_CLASS(ZProxyStackIface),
+  "ZTransfer2PSIface",
+  sizeof(ZTransfer2PSIface),
+  &z_transfer2_ps_iface_funcs.super,
+};
+
 /**
  * z_transfer2_buffer_empty:
  * @self: ZTransfer2Buffer instance
@@ -65,23 +200,6 @@ z_transfer2_buffer_destroy(ZTransfer2Buffer *self)
   g_free(self->buf);
 }
 
-/**
- * z_transfer2_get_stream:
- * @self: ZTransfer2 instance
- * @endpoint: endpoint index
- *
- * This function returns the stream associated to the endpoint specified
- * by @endpoint. @endpoint should be one of the ZT2E_* values.
- **/
-static inline ZStream *
-z_transfer2_get_stream(ZTransfer2 *self, gint endpoint)
-{
-  if (endpoint & ZT2E_STACKED)
-    return self->stacked ? self->stacked->downstreams[endpoint & ~ZT2E_STACKED] : NULL;
-  else
-    return self->endpoints[endpoint];
-}
-
 /**
  * z_transfer2_timeout:
  * @user_data: ZTransfer2 instance passed as a generic pointer
@@ -152,25 +270,25 @@ z_transfer2_update_cond(ZTransfer2 *self)
     {
       if ((i & ZT2E_STACKED) == 0 || self->stacked)
         {
-          z_stream_set_cond(z_transfer2_get_stream(self, i), Z_STREAM_FLAG_READ, FALSE);
-          z_stream_set_cond(z_transfer2_get_stream(self, i), Z_STREAM_FLAG_WRITE, FALSE);
+          z_stream_set_cond(z_transfer2_get_stream(self, i), G_IO_IN, FALSE);
+          z_stream_set_cond(z_transfer2_get_stream(self, i), G_IO_OUT, FALSE);
         }
     }
   if (self->stacked)
     {
       if (!z_transfer2_get_status(self, ZT2S_EOF_SOURCE))
         {
-          if (z_transfer2_buffer_empty(&self->buffers[0]))
-            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), Z_STREAM_FLAG_READ, TRUE);
+          if (z_transfer2_buffer_empty(&self->buffers[0]) && !z_transfer2_get_status(self, ZT2S_PROXY_OUT))
+            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), G_IO_IN, TRUE);
           else
-            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE), Z_STREAM_FLAG_WRITE, TRUE);
+            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE), G_IO_OUT, TRUE);
         }
       if (!z_transfer2_get_status(self, ZT2S_EOF_DEST))
         {
           if (z_transfer2_buffer_empty(&self->buffers[1]))
-            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DOWN_DEST), Z_STREAM_FLAG_READ, TRUE);
+            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DOWN_DEST), G_IO_IN, TRUE);
           else
-            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DEST), Z_STREAM_FLAG_WRITE, TRUE);
+            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DEST), G_IO_OUT, TRUE);
         }
     }
   else
@@ -178,10 +296,10 @@ z_transfer2_update_cond(ZTransfer2 *self)
       /* no stacking */
       if (!z_transfer2_get_status(self, ZT2S_EOF_SOURCE))
         {
-          if (z_transfer2_buffer_empty(&self->buffers[0]) || z_transfer2_get_status(self, ZT2S_EOF_DEST) != 0)
-            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), Z_STREAM_FLAG_READ, TRUE);
+          if ((z_transfer2_buffer_empty(&self->buffers[0]) || z_transfer2_get_status(self, ZT2S_EOF_DEST) != 0) && !z_transfer2_get_status(self, ZT2S_PROXY_OUT))
+            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), G_IO_IN, TRUE);
           else
-            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DEST), Z_STREAM_FLAG_WRITE, TRUE);
+            z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DEST), G_IO_OUT, TRUE);
         }
     }
   z_proxy_leave(self->owner);
@@ -430,6 +548,7 @@ z_transfer2_copy_data(ZTransfer2 *self, gint ep_from, gint ep_to, GError **error
               else 
                 {
                   z_transfer2_update_status(self, ZT2S_FINISHED + ZT2S_ABORTED, TRUE);
+                  z_transfer2_eof(self, ep_from);
                   leave_while = TRUE;
                   break;
                 }
@@ -562,11 +681,25 @@ gboolean
 z_transfer2_start(ZTransfer2 *self)
 {
   gboolean res;
+  ZProxyIface *iface;
   
   z_proxy_enter(self->owner);
   
+  iface = z_transfer2_ps_iface_new(self);
+  z_proxy_add_iface(self->owner, iface);
+  z_object_unref(&iface->super);
+        
+  g_mutex_lock(self->startup_lock);
+  if (!z_transfer2_stack_proxy(self, &self->stacked))
+    {
+      g_mutex_unlock(self->startup_lock);
+      z_proxy_log(self->owner, CORE_ERROR, 3, "Could not start stacked proxy, rejecting transfer;");
+      z_proxy_leave(self->owner);
+      return FALSE;
+    }
+ 
   z_transfer2_switch_to_transfer_context(self);
-  
+
   /* NOTE: shutdown goes back to blocking mode in which case timeout is significant */
   z_stream_set_timeout(z_transfer2_get_stream(self, ZT2E_SOURCE), self->timeout);
   z_stream_set_timeout(z_transfer2_get_stream(self, ZT2E_DEST), self->timeout);
@@ -578,45 +711,46 @@ z_transfer2_start(ZTransfer2 *self)
       z_poll_add_stream(self->poll, z_transfer2_get_stream(self, ZT2E_SOURCE));
       z_poll_add_stream(self->poll, z_transfer2_get_stream(self, ZT2E_DEST));
     }
-
-  z_transfer2_stack_proxy(self);
   
   /* initialize stacked streams */
   if (self->stacked)
     {
-      /* shutdown the reverse direction */
-      z_stream_shutdown(z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE), SHUT_RD, NULL);
-      z_stream_shutdown(z_transfer2_get_stream(self, ZT2E_DOWN_DEST), SHUT_WR, NULL);
+      if ((self->stacked->flags & Z_SPF_HALF_DUPLEX) == 0)
+        {
+          /* shutdown the reverse direction */
+          z_stream_shutdown(z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE), SHUT_RD, NULL);
+          z_stream_shutdown(z_transfer2_get_stream(self, ZT2E_DOWN_DEST), SHUT_WR, NULL);
+        }
 
       z_transfer2_buffer_init(&self->buffers[1], self->buffer_size);
       z_poll_add_stream(self->poll, z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE));
       z_poll_add_stream(self->poll, z_transfer2_get_stream(self, ZT2E_DOWN_DEST));
 
       z_stream_set_callback(z_transfer2_get_stream(self, ZT2E_SOURCE),
-                            Z_STREAM_FLAG_READ,
+                            G_IO_IN,
                             z_transfer2_copy_src_to_down,
                             self,
                             NULL);
       z_stream_set_callback(z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE),
-                            Z_STREAM_FLAG_WRITE,
+                            G_IO_OUT,
                             z_transfer2_copy_src_to_down,
                             self,
                             NULL);
                                                                                                                              
-      z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), Z_STREAM_FLAG_READ, TRUE);
+      z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), G_IO_IN, TRUE);
 
       z_stream_set_callback(z_transfer2_get_stream(self, ZT2E_DOWN_DEST),
-                            Z_STREAM_FLAG_READ,
+                            G_IO_IN,
                             z_transfer2_copy_down_to_dst,
                             self,
                             NULL);
       z_stream_set_callback(z_transfer2_get_stream(self, ZT2E_DEST),
-                            Z_STREAM_FLAG_WRITE,
+                            G_IO_OUT,
                             z_transfer2_copy_down_to_dst,
                             self,
                             NULL);
                                                                                                                              
-      z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DEST | ZT2E_STACKED), Z_STREAM_FLAG_READ, TRUE);
+      z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_DEST | ZT2E_STACKED), G_IO_IN, TRUE);
 
       z_stream_set_nonblock(z_transfer2_get_stream(self, ZT2E_DOWN_SOURCE), TRUE);
       z_stream_set_nonblock(z_transfer2_get_stream(self, ZT2E_DOWN_DEST), TRUE);
@@ -625,17 +759,17 @@ z_transfer2_start(ZTransfer2 *self)
   else
     {
       z_stream_set_callback(z_transfer2_get_stream(self, ZT2E_SOURCE),
-                            Z_STREAM_FLAG_READ,
+                            G_IO_IN,
                             z_transfer2_copy_src_to_dst,
                             self,
                             NULL);
       z_stream_set_callback(z_transfer2_get_stream(self, ZT2E_DEST),
-                            Z_STREAM_FLAG_WRITE,
+                            G_IO_OUT,
                             z_transfer2_copy_src_to_dst,
                             self,
                             NULL);
 
-      z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), Z_STREAM_FLAG_READ, TRUE);
+      z_stream_set_cond(z_transfer2_get_stream(self, ZT2E_SOURCE), G_IO_IN, TRUE);
     }
   z_stream_set_nonblock(z_transfer2_get_stream(self, ZT2E_SOURCE), TRUE);
   z_stream_set_nonblock(z_transfer2_get_stream(self, ZT2E_DEST), TRUE);
@@ -643,6 +777,7 @@ z_transfer2_start(ZTransfer2 *self)
   res = z_transfer2_setup(self);
 
   z_transfer2_switch_to_proxy_context(self);
+  g_mutex_unlock(self->startup_lock);
 
   if (self->timeout > 0)
     {
@@ -663,6 +798,8 @@ z_transfer2_start(ZTransfer2 *self)
       g_source_attach(self->progress_source, context);
     }
     
+  
+  z_proxy_leave(self->owner);
   return res;
 }
 
@@ -688,28 +825,44 @@ z_transfer2_suspend(ZTransfer2 *self, gint suspend_reason)
 }
 
 /**
- * z_transfer2_cancel:
+ * z_transfer2_rollback:
  * @self: ZTransfer2 instance
  *
  * This function can be called when transfer was suspended and the data 
  * stream is to be dropped. It basically consumes the incoming data stream
  * without actually writing it to the server side.
+ *
+ * This function can only be called when the transfer is suspended.
  **/
 gboolean 
-z_transfer2_cancel(ZTransfer2 *self)
+z_transfer2_rollback(ZTransfer2 *self)
 {
   z_enter();
   if (z_transfer2_get_status(self, ZT2S_STARTED) && !z_transfer2_get_status(self, ZT2S_FINISHED))
     {
       /* roll back transfer state */
-      
       z_transfer2_update_status(self, ZT2S_COPYING_TAIL, TRUE);
       while (z_transfer2_run(self) == ZT2_RESULT_SUSPENDED)
         ;
     }
-    
-  z_leave();
-  return TRUE;
+  z_return(TRUE);
+}
+
+/**
+ * z_transfer2_cancel:
+ * @self: ZTransfer2 instance
+ *
+ * This function abort the transfer. This function can be called from anywhere but
+ * it's only set the state of the transfer. You have to continue running the
+ * transfer untill it's quit.
+ **/
+gboolean 
+z_transfer2_cancel(ZTransfer2 *self)
+{
+  z_enter();
+  if (!z_transfer2_get_status(self, ZT2S_FINISHED))
+    z_transfer2_update_status(self, ZT2S_FINISHED + ZT2S_ABORTED, TRUE);
+  z_return(TRUE);
 }
 
 /**
@@ -723,12 +876,17 @@ z_transfer2_run_method(ZTransfer2 *self)
 {
   z_enter();
   z_transfer2_switch_to_transfer_context(self);
+  z_transfer2_update_cond(self);  
   
   z_transfer2_update_status(self, ZT2S_STARTED, TRUE);
   z_transfer2_update_status(self, ZT2S_SUSPENDED, FALSE);
   while (((self->status & (ZT2S_FINISHED+ZT2S_SUSPENDED)) == 0) && z_poll_iter_timeout(self->poll, -1))
     {
-      ;
+      if (!z_proxy_loop_iteration(self->owner))
+        {
+          z_transfer2_update_status(self, ZT2S_ABORTED + ZT2S_FINISHED, TRUE);
+          break;
+        }
     }
   z_transfer2_switch_to_proxy_context(self);
   if (self->status & ZT2S_SUSPENDED)
@@ -785,7 +943,7 @@ z_transfer2_simple_run(ZTransfer2 *self)
   success = (tr != ZT2_RESULT_FAILED) && (tr != ZT2_RESULT_ABORTED);
   
   if (tr == ZT2_RESULT_FAILED)
-    z_transfer2_cancel(self);
+    z_transfer2_rollback(self);
     
   return success;
 }
@@ -814,6 +972,7 @@ z_transfer2_new(ZClass *class,
   ZTransfer2 *self;
   
   z_proxy_enter(owner);
+  
   self = Z_NEW_COMPAT(class, ZTransfer2);
   z_proxy_ref(owner);
   self->owner = owner;
@@ -824,6 +983,13 @@ z_transfer2_new(ZClass *class,
   self->buffer_size = buffer_size;
   self->timeout = timeout;
   self->flags = flags;
+  self->content_format = "file";
+  
+  self->startup_lock = g_mutex_new();
+  
+  self->stack_info = g_string_sized_new(32);
+  self->stack_decision = Z_ACCEPT;
+  
   z_proxy_leave(owner);
   return self;
 }
@@ -839,12 +1005,24 @@ void
 z_transfer2_free_method(ZObject *s)
 {
   ZTransfer2 *self = Z_CAST(s, ZTransfer2);
+  ZProxyIface *iface;
   guint i;
 
   z_enter();
+  iface = z_proxy_find_iface(self->owner, Z_CLASS(ZTransfer2PSIface));
+  if (iface)
+    {
+      z_proxy_del_iface(self->owner, iface);
+      z_object_unref(&iface->super);
+    }
   
   z_proxy_unref(self->owner);
-  z_poll_unref(self->poll);
+  if ((self->flags & ZT2F_PROXY_STREAMS_POLLED) == 0)
+    {
+      z_poll_remove_stream(self->poll, z_transfer2_get_stream(self, ZT2E_SOURCE));
+      z_poll_remove_stream(self->poll, z_transfer2_get_stream(self, ZT2E_DEST));
+    }
+
   z_stream_unref(self->endpoints[0]);
   z_stream_unref(self->endpoints[1]);
   z_transfer2_buffer_destroy(&self->buffers[0]);
@@ -870,13 +1048,16 @@ z_transfer2_free_method(ZObject *s)
   for (i = 0; i < EP_MAX; i++)
     {
       if (self->transfer_contexts[i].stream_extra)
-        {
-          z_stream_context_destroy(&self->transfer_contexts[i]);
-        }
+        z_stream_context_destroy(&self->transfer_contexts[i]);
     }
-  z_object_free_method(s);
+  z_poll_unref(self->poll);
+  g_string_free(self->stack_info, TRUE);
   
-  z_leave();
+  if (self->startup_lock)
+    g_mutex_free(self->startup_lock);
+  
+  z_object_free_method(s);
+  z_return();
 }
 
 ZTransfer2Funcs z_transfer2_funcs = 
