@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -86,7 +87,6 @@ static void
 finger_config_set_defaults(FingerProxy *self)
 {
   z_proxy_enter(self);
-  
   self->max_line_length = 132;
   self->max_username_length = 8;
   self->max_hostname_length = 30;
@@ -96,10 +96,8 @@ finger_config_set_defaults(FingerProxy *self)
   self->hostnames = g_string_sized_new(0);
   self->response_header = g_string_sized_new(0);
   self->response_footer = g_string_sized_new(0);
-
   self->timeout = 30000;
-
-  z_proxy_leave(self);
+  z_proxy_return(self);
 }
 
 /**
@@ -112,7 +110,15 @@ static void
 finger_register_vars(FingerProxy *self)
 {
   z_proxy_enter(self);
-  
+#if 0
+  z_polict_dict_register(self->super.dict, 
+                         Z_VT_INT, "timeout", Z_VF_READ | Z_VF_CFG_RW, &self->timeout, NULL,
+                         Z_VT_INT, "max_line_length", Z_VF_READ | Z_VF_CFG_RW, &self->max_line_length, NULL,
+                         Z_VT_INT, "max_username_length", Z_VF_READ | Z_VF_CFG_RW, &self->max_username_length, NULL,
+                         Z_VT_INT, "max_hostname_length", Z_VF_READ | Z_VF_CFG_RW, &self->max_hostname_length, NULL,
+                         Z_VT_INT, "max_hop_count", Z_VF_READ | Z_VF_CFG_RW, &self->max_hop_count, NULL,
+#endif
+
   z_proxy_var_new(&self->super, "timeout",
                   Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                   &self->timeout);
@@ -165,8 +171,7 @@ finger_register_vars(FingerProxy *self)
   z_proxy_var_new(&self->super, "strict_username_check",
                   Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                   &self->strict_username_check);
-
-  z_proxy_leave(self);
+  z_proxy_return(self);
 }
 
 /**
@@ -182,15 +187,11 @@ finger_init_client_stream(FingerProxy *self)
   ZStream *tmpstream;
   
   z_proxy_enter(self);
-  
   self->super.endpoints[EP_CLIENT]->timeout = self->timeout;
   tmpstream = self->super.endpoints[EP_CLIENT];
-  
   self->super.endpoints[EP_CLIENT] = z_stream_line_new(tmpstream, self->max_line_length, ZRL_EOL_CRLF);
-  
   z_stream_unref(tmpstream);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -207,20 +208,12 @@ finger_init_server_stream(FingerProxy *self)
   
   z_proxy_enter(self);
   if (!self->super.endpoints[EP_SERVER])
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
+    z_proxy_return(self, FALSE);
   self->super.endpoints[EP_SERVER]->timeout = self->timeout;
-  
   tmpstream = self->super.endpoints[EP_SERVER];
-  
   self->super.endpoints[EP_SERVER] = z_stream_line_new(tmpstream, self->max_line_length, ZRL_EOL_CRLF);
-  
   z_stream_unref(tmpstream);
-  
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -247,20 +240,16 @@ finger_fetch_request(FingerProxy *self)
         This message is appears when zorp cannot read finger request
        */
       z_proxy_log(self, FINGER_ERROR, 1, "Error reading request;");
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
   
   /*LOG
     This message is say about read finger request
    */
   z_proxy_log(self, FINGER_REQUEST, 6, "Request details; req='%.*s'", (gint) line_length, line);
-  
   p = line;
   left = line_length;
-  
   self->long_req = FALSE;
-  
   while (*p == ' ' && left)
     {
       p++;
@@ -282,8 +271,7 @@ finger_fetch_request(FingerProxy *self)
             This message appear when zorp cannot parse request
            */
           z_proxy_log(self, FINGER_VIOLATION, 1, "Parse error, dropping request; req='%.*s'", (gint) line_length, line);
-          z_proxy_leave(self);
-          return FALSE;
+          z_proxy_return(self, FALSE);
         }
     }
   while (*p == ' ' && left)
@@ -317,8 +305,7 @@ finger_fetch_request(FingerProxy *self)
             in finger request.
            */
           z_proxy_log(self, FINGER_VIOLATION, 1, "Invalid character, dropping request; line='%.*s'", (gint) line_length, line);
-          z_proxy_leave(self);
-          return FALSE;
+          z_proxy_return(self, FALSE);
         }
       if (fetch_user)
         {
@@ -330,8 +317,7 @@ finger_fetch_request(FingerProxy *self)
                 in the request.
                */
               z_proxy_log(self, FINGER_VIOLATION, 1, "Username too long, dropping request; line='%.*s'", (gint) line_length, line);
-              z_proxy_leave(self);
-              return FALSE;
+              z_proxy_return(self, FALSE);
             }
         }
       else
@@ -344,17 +330,14 @@ finger_fetch_request(FingerProxy *self)
                 in the request horname chain.
                */
               z_proxy_log(self, FINGER_VIOLATION, 1, "One hostname is too long in hostname chain, dropping request; req='%.*s'", (gint) line_length, line);
-              z_proxy_leave(self);
-              return FALSE;
+              z_proxy_return(self, FALSE);
             }
           hostlen++;
         }
       p++;
       left--;
     }
-  
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -371,10 +354,9 @@ finger_send_request(FingerProxy *self)
   gsize bytes_written;
   
   z_proxy_enter(self);
-  
   if (self->long_req)
     {
-      if (self->username->len >0)
+      if (self->username->len > 0)
         {
           if (self->hostnames->len > 0)
             g_snprintf(request, sizeof(request), "/W %s%s\r\n",
@@ -395,7 +377,7 @@ finger_send_request(FingerProxy *self)
     }
   else
     {
-      if (self->username->len >0)
+      if (self->username->len > 0)
         {
           if (self->hostnames->len > 0)
             g_snprintf(request, sizeof(request), "%s%s\r\n",
@@ -426,11 +408,9 @@ finger_send_request(FingerProxy *self)
         found in server side.
        */
       z_proxy_log(self, FINGER_ERROR, 1, "Error write request;");
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -459,8 +439,7 @@ finger_copy_response(FingerProxy *self)
         found in client side when writting the header.
        */
       z_proxy_log(self, FINGER_ERROR, 1, "Error write request;");
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
 
   while (1)
@@ -469,6 +448,9 @@ finger_copy_response(FingerProxy *self)
       gsize line_len;
       gchar *response;
       
+      if (!z_proxy_loop_iteration(&self->super))
+        break;
+      
       res = z_stream_line_get(self->super.endpoints[EP_SERVER], &line, &line_len, NULL);
       if (res != G_IO_STATUS_NORMAL)
         /* EOF or read error */
@@ -477,7 +459,6 @@ finger_copy_response(FingerProxy *self)
       response = alloca(line_len + 3);
       memcpy(response, line, line_len);
       strcpy(response + line_len, "\r\n");
-
       if (z_stream_write(self->super.endpoints[EP_CLIENT],
                          response,
                          line_len + 2,
@@ -489,8 +470,7 @@ finger_copy_response(FingerProxy *self)
             found in client side when writting the response.
            */
           z_proxy_log(self, FINGER_ERROR, 1, "Error write request;");
-          z_proxy_leave(self);
-          return FALSE;
+          z_proxy_return(self, FALSE);
         }
 
     }
@@ -507,12 +487,9 @@ finger_copy_response(FingerProxy *self)
         found in client side when writting the footer.
        */
       z_proxy_log(self, FINGER_ERROR, 1, "Error write request;");
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
-
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -529,7 +506,6 @@ finger_query_policy(FingerProxy *self)
   gint res;
   
   z_proxy_enter(self);
-  
   z_policy_lock(self->super.thread);
   res = z_policy_event(self->super.handler, "fingerRequest", z_policy_var_build("(ss)", self->username->str, self->hostnames->str), self->super.session_id);
   switch (res)
@@ -559,15 +535,14 @@ finger_query_policy(FingerProxy *self)
           z_proxy_log(self, FINGER_POLICY, 2, "Policy violation, drop session;");
         }
       z_policy_unlock(self->super.thread);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
+
     case FINGER_REQ_ACCEPT:
     default:
       break;
     }
   z_policy_unlock(self->super.thread);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -579,12 +554,7 @@ finger_config(ZProxy *s)
   FingerProxy *self = Z_CAST(s, FingerProxy);
   finger_config_set_defaults(self);  
   finger_register_vars(self);
-  
-  if (Z_SUPER(s, ZProxy)->config(s))
-    {
-      return TRUE;
-    }
-  return FALSE;
+  return Z_SUPER(s, ZProxy)->config(s);
 }
 
 
@@ -598,13 +568,10 @@ static void
 finger_main(ZProxy *s)
 {
   FingerProxy *self = Z_CAST(s, FingerProxy);
+ 
   z_proxy_enter(self);
-    
   if (!finger_init_client_stream(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
 
   /*LOG
     This debug message is about proxy state when start to
@@ -621,8 +588,7 @@ finger_main(ZProxy *s)
                      strlen(errmsg),
                      &bytes_written,
                      NULL);
-      z_proxy_leave(self);
-      return;
+      z_proxy_return(self);
     }
   
   /*LOG
@@ -630,31 +596,20 @@ finger_main(ZProxy *s)
     fetched request and asking policy about it
    */
   z_proxy_log(self, FINGER_DEBUG, 6, "asking policy;");
-  
   if (!finger_query_policy(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
   
   /*LOG
     This debug message is about proxy state when finger
     start connect to server.
    */
   z_proxy_log(self, FINGER_DEBUG, 6, "connecting server;");
-  
   /* this sets the server side endpoint if successful */
   if (!z_proxy_connect_server(&self->super, NULL, 0))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
 
   if (!finger_init_server_stream(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
   
   /*LOG
     This debug message is about proxy state when finger
@@ -662,10 +617,7 @@ finger_main(ZProxy *s)
    */
   z_proxy_log(self, FINGER_DEBUG, 6, "sending request;");
   if (!finger_send_request(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
   
   /*LOG
     This debug message is about proxy state when finger
@@ -673,17 +625,14 @@ finger_main(ZProxy *s)
    */
   z_proxy_log(self, FINGER_DEBUG, 6, "copying response;");
   if (!finger_copy_response(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
 
   /*LOG
     This debug message is about proxy state when finger
     stop it's work.
    */
   z_proxy_log(self, FINGER_DEBUG, 6, "everything is done;");
-  z_proxy_leave(self);
+  z_proxy_return(self);
 
 }
 
@@ -701,10 +650,7 @@ finger_proxy_new(ZProxyParams *params)
   
   z_enter();
   self = Z_CAST(z_proxy_new(Z_CLASS(FingerProxy), params), FingerProxy);
-  
-  z_proxy_start(&self->super);
-  z_leave();
-  return (ZProxy *) self;
+  z_return((ZProxy *) self);
 }
 
 ZProxyFuncs finger_proxy_funcs =
@@ -713,11 +659,8 @@ ZProxyFuncs finger_proxy_funcs =
     Z_FUNCS_COUNT(ZProxy),
     NULL
   },
-  finger_config,
-  NULL,
-  finger_main,
-  NULL,
-  NULL,
+  .config = finger_config,
+  .main = finger_main,
   NULL
 };
 
