@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -29,6 +30,7 @@
 #define ZORP_MODULES_FTP_H_INCLUDED
 
 #include <zorp/proxy.h>
+#include <zorp/proxystack.h>
 #include <zorp/streamline.h>
 #include <zorp/sockaddr.h>
 #include <zorp/poll.h>
@@ -37,6 +39,7 @@
 #include <zorp/dispatch.h>
 #include <zorp/connect.h>
 #include <zorp/misc.h>
+#include <zorp/proxy/transfer2.h>
 
 #include "ftphash.h"
 #include "ftpolicy.h"
@@ -47,7 +50,7 @@
 #define FTP_DEBUG     "ftp.debug"
 #define FTP_ERROR     "ftp.error"
 #define FTP_REQUEST   "ftp.request"
-#define FTP_REPLY     "ftp.reply"
+#define FTP_RESPONSE  "ftp.response"
 #define FTP_POLICY    "ftp.policy"
 #define FTP_VIOLATION "ftp.violation"
 #define FTP_SESSION   "ftp.session"
@@ -72,6 +75,9 @@
 #define FTP_RSP_ABORT  4   /* Doesn't allow, connection will dropped      */
 #define FTP_RSP_POLICY 6   /* Put the answer up to policy                 */
 
+#define FTP_STK_NONE           1
+#define FTP_STK_DATA           2
+
 //
 // FTP statemachine states
 //
@@ -202,10 +208,22 @@ typedef struct _FtpProxy
   guint timeout;
 
   GString *target_port_range;
+
+  guint max_continuous_line;
   
   /* Data connection protect */
   GMutex *lock;
 
+  gboolean ftp_data_hangup;
+
+  ZTransfer2 *transfer;
+
+  gsize buffer_size;
+
+  gboolean drop_answer;
+  
+  gchar *preamble;
+
 } FtpProxy;  
 
 extern ZClass FtpProxy__class;
@@ -238,6 +256,7 @@ extern ZClass FtpProxy__class;
 #define MSG_TIMED_OUT                  25
 #define MSG_ERROR_PARAMETER_EPSV       26
 #define MSG_ERROR_PARAMETER_EPRT       27
+#define MSG_DATA_TRANSFER_FAILED       28
 extern Ftp_message ftp_know_messages[];
 
 gboolean ftp_data_prepare(FtpProxy *self, gint side, gchar mode);
@@ -246,6 +265,10 @@ void ftp_data_reset(FtpProxy *self);
 
 void ftp_state_both(FtpProxy *self);
 
+gboolean ftp_data_transfer(FtpProxy *self, ZStream *from_stream, ZStream *to_stream);
+
+gchar *ftp_answer_setup(FtpProxy *self, gchar *answer_c, gchar *answer_p);
+gboolean ftp_answer_write(FtpProxy *self, gchar *msg);
 
 #define SET_ANSWER(answer) {\
 	g_string_assign(self->answer_cmd, ftp_know_messages[answer].code);\
