@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -63,5 +64,6 @@ Ftp_message ftp_know_messages[] =
   {"500", "Error parsing command"},
   {"500", "Connection timed out"},
   {"500", "Error parsing EPSV parameters"},
-  {"500", "Error parsing EPRT parameters"}
+  {"500", "Error parsing EPRT parameters"},
+  {"550", "Data transfer failed"},
 };
