@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -57,31 +58,21 @@ ftp_policy_bounce_check(FtpProxy *self, guint  side, ZSockAddr *remote, gboolean
   
   z_proxy_enter(self);
   z_policy_lock(self->super.thread);
-  zsock = z_py_zorp_sockaddr_new(remote);
+  zsock = z_policy_sockaddr_new(remote);
   res = z_policy_call(self->super.handler, "bounceCheck", z_policy_var_build("(Oii)", zsock, side, connect), &called, self->super.session_id);
   if (!called)
     {
       z_policy_unlock(self->super.thread);
-      z_proxy_leave(self);
-      return TRUE;
+      z_proxy_return(self, TRUE);
     }
   
-  if (res == NULL)
-    {
-      ret = FALSE;
-    }
-  else
-    {
-      if(!z_policy_var_parse(res, "i", &ret))
-        {
-          ret = FALSE;
-        }
-    }
+  if ((res == NULL) || !z_policy_var_parse(res, "i", &ret))
+    ret = FALSE;
+
   z_policy_var_unref(res);
   z_policy_var_unref(zsock);
   z_policy_unlock(self->super.thread);
-  z_proxy_leave(self);
-  return ret;
+  z_proxy_return(self, ret);
 }
 
 GHashTable *
@@ -144,6 +135,7 @@ ftp_policy_command_hash_do(FtpProxy *self)
   tmp = g_hash_table_lookup(self->policy_command_hash, self->request_cmd->str);
   if (!tmp)
     {
+      z_proxy_log(self, FTP_POLICY, 6, "Policy does not contain this request, trying the default; request='%s'", self->request_cmd->str);
       tmp = g_hash_table_lookup(self->policy_command_hash, "*");
     }
   if (!tmp)
@@ -152,10 +144,10 @@ ftp_policy_command_hash_do(FtpProxy *self)
         This message indicates that the policy does not contain any setting for the given
         request and Zorp rejects the request. Check the 'request' attribute.
        */
-      z_proxy_log(self, FTP_POLICY, 5, "Policy doesn't contain default. Use hard-coded; req='%s'", self->request_cmd->str);
-      z_proxy_leave(self);
-      return FTP_REQ_REJECT;
+      z_proxy_log(self, FTP_POLICY, 5, "Policy does not contain this request, using hard-coded default; request='%s'", self->request_cmd->str);
+      z_proxy_return(self, FTP_REQ_REJECT);
     }
+  
   z_policy_lock(self->super.thread);
   if (!ftp_hash_get_type(tmp,&command_do))
     {
@@ -164,21 +156,23 @@ ftp_policy_command_hash_do(FtpProxy *self)
         rejects the request. Check the 'request' attribute.
        */
       z_proxy_log(self, FTP_POLICY, 1, "Policy type invalid; req='%s", self->request_cmd->str);
-      z_proxy_leave(self);
-      return FTP_REQ_REJECT;
+      z_policy_unlock(self->super.thread);
+      z_proxy_return(self, FTP_REQ_REJECT);
     }
   z_policy_unlock(self->super.thread);
+  
   switch(command_do)
     {
     case FTP_REQ_ACCEPT:
     case FTP_REQ_ABORT:
       ret = command_do;
       break;
+      
     case FTP_REQ_REJECT:
       z_policy_lock(self->super.thread);
       if (z_policy_var_parse(tmp, "(is)", &command_do, &msg))
         {
-          for(i = 0; i < 3; i++)
+          for (i = 0; i < 3; i++)
             work[i]=msg[i];
 
           work[i]=0;
@@ -188,9 +182,10 @@ ftp_policy_command_hash_do(FtpProxy *self)
       ret = command_do;
       z_policy_unlock(self->super.thread);
       break;
+      
     case FTP_REQ_POLICY:
       z_policy_lock(self->super.thread);
-      if(!z_policy_var_parse(tmp,"(iO)",&command_do,&command_where))
+      if (!z_policy_var_parse(tmp,"(iO)",&command_do,&command_where))
         {
           /*LOG
             This message indicates that the policy for the given request is invalid
@@ -204,7 +199,6 @@ ftp_policy_command_hash_do(FtpProxy *self)
         {
           g_string_assign(self->answer_cmd, "500");
           g_string_assign(self->answer_param, "Error parsing command");
-          
           res = z_policy_call_object(command_where, z_policy_var_build("(s)",self->request_cmd->str), self->super.session_id);
           if (res == NULL)
             {
@@ -216,43 +210,42 @@ ftp_policy_command_hash_do(FtpProxy *self)
               z_proxy_log(self, FTP_POLICY, 1, "Error in policy calling; req='%s'", self->request_cmd->str);
               ret = FTP_REQ_ABORT;
             }
+          else if (!z_policy_var_parse(res,"i",&ret))
+            {
+              /*LOG
+                This message indicates that the returned value of the callback for the given request policy 
+                is invalid and Zorp rejects the request. Check the callback function.
+               */
+              z_proxy_log(self, FTP_POLICY, 1, "Can't parsing return code; command='%s'", self->request_cmd->str);
+              ret = FTP_REQ_ABORT;
+            }
           else
             {
-              if (!z_policy_var_parse(res,"i",&ret))
+              switch(ret)
                 {
-                  /*LOG
-                    This message indicates that the returned value of the callback for the given request policy 
-                    is invalid and Zorp rejects the request. Check the callback function.
-                   */
-                  z_proxy_log(self, FTP_POLICY, 1, "Can't parsing return code; command='%s'", self->request_cmd->str);
-                  ret = FTP_REQ_ABORT;
-                }
-              else
-                {
-                  switch(ret)
-                    {
-                    case FTP_REQ_ACCEPT:
-                    case FTP_REQ_ABORT:
-                    case FTP_REQ_REJECT:
-                      break;
-                    case Z_UNSPEC:
-                    case Z_DROP:
-                      ret = FTP_REQ_REJECT;
-                      break;
-                    default:
-                      break;
-                    }
+                case FTP_REQ_ACCEPT:
+                case FTP_REQ_ABORT:
+                case FTP_REQ_REJECT:
+                  break;
+                  
+                case Z_UNSPEC:
+                case Z_DROP:
+                  ret = FTP_REQ_REJECT;
+                  break;
+
+                default:
+                  break;
                 }
             }
         }
       z_policy_unlock(self->super.thread);
       break;
+      
     default:
       ret = FTP_REQ_ABORT;
       break;
     }
-  z_proxy_leave(self);
-  return ret;
+  z_proxy_return(self, ret);
 }
 
 ZDimHashTable *
@@ -261,7 +254,6 @@ ftp_policy_answer_hash_create(void)
   ZDimHashTable *tmp;
 
   tmp = z_dim_hash_table_new(2, 2, DIMHASH_WILDCARD, DIMHASH_CONSUME);
-
   return tmp;
 }
 
@@ -281,28 +273,23 @@ ftp_policy_answer_hash_do(FtpProxy *self)
   gchar work[10];
 
   z_proxy_enter(self);
-  
   if (self->request_cmd->len > 0)
     g_snprintf(key1, sizeof(key1), "%s", self->request_cmd->str);
   else
     g_snprintf(key1, sizeof(key1), "Null");
 
   g_snprintf(key2, sizeof(key2), "%s", self->answer_cmd->str);
-  
   key[0] = key1;
   key[1] = key2;
-  
   tmp = z_dim_hash_table_search(self->policy_answer_hash, 2, key);
-  
   if (!tmp)
     {
       /*LOG
         This message indicates that the policy does not contain any setting for the given
         response and Zorp rejects the response. Check the 'response' attribute.
        */
-      z_proxy_log(self, FTP_POLICY, 5, "Policy doesn't contain default. Use hard-coded default; req='%s', rsp='%s", self->request_cmd->str, self->answer_cmd->str);
-      z_proxy_leave(self);
-      return FTP_RSP_REJECT;
+      z_proxy_log(self, FTP_POLICY, 5, "Policy does not contain this response, using hard-coded default; request='%s', response='%s", self->request_cmd->str, self->answer_cmd->str);
+      z_proxy_return(self, FTP_RSP_REJECT);
     }
     
   z_policy_lock(self->super.thread);
@@ -313,8 +300,7 @@ ftp_policy_answer_hash_do(FtpProxy *self)
         rejects the request. Check the 'request' attribute.
        */
       z_proxy_log(self, FTP_POLICY, 1, "Answer type invalid; req='%s', rsp='%s'", self->request_cmd->str, self->answer_cmd->str);
-      z_proxy_leave(self);
-      return FTP_RSP_REJECT;
+      z_proxy_return(self, FTP_RSP_REJECT);
     }
   z_policy_unlock(self->super.thread);
   
@@ -322,7 +308,6 @@ ftp_policy_answer_hash_do(FtpProxy *self)
     {
     case FTP_RSP_REJECT:
       ret = FTP_RSP_REJECT;
-      
       z_policy_lock(self->super.thread);
       if (!z_policy_var_parse(tmp, "(is)", &answer_do, &msg))
         {
@@ -339,12 +324,13 @@ ftp_policy_answer_hash_do(FtpProxy *self)
         }
       z_policy_unlock(self->super.thread);
       break;
+      
     case FTP_RSP_ACCEPT:
       ret = FTP_RSP_ACCEPT;
       break;
+      
     case FTP_RSP_ABORT:
       ret = FTP_RSP_ABORT;
-      
       z_policy_lock(self->super.thread);
       if (!z_policy_var_parse(tmp, "(is)", &answer_do, &msg))
         {
@@ -361,9 +347,10 @@ ftp_policy_answer_hash_do(FtpProxy *self)
         }
       z_policy_unlock(self->super.thread);
       break;
+      
     case FTP_RSP_POLICY:
       z_policy_lock(self->super.thread);
-      if(!z_policy_var_parse(tmp,"(iO)", &answer_do, &answer_where))
+      if (!z_policy_var_parse(tmp,"(iO)", &answer_do, &answer_where))
         {
           /*LOG
             This message indicates that the policy for the given response is invalid
@@ -390,47 +377,47 @@ ftp_policy_answer_hash_do(FtpProxy *self)
               g_string_assign(self->answer_param, "Error parsing answer (bad policy)");
               ret = FTP_RSP_ABORT;
             }
+          else if (!z_policy_var_parse(res, "i", &ret))
+            {
+              /*LOG
+                This message indicates that the returned value of the callback for the given response policy 
+                is invalid and Zorp rejects the response. Check the callback function.
+               */
+              z_proxy_log(self, FTP_POLICY, 1, "Return code invalid from policy function; command='%s', answer='%s'", self->request_cmd->str, self->answer_cmd->str);
+              g_string_assign(self->answer_cmd, "500");
+              g_string_assign(self->answer_param, "Error parsing answer (bad policy)");
+              ret = FTP_RSP_ABORT;
+            }
           else
             {
-              if(!z_policy_var_parse(res, "i", &ret))
+              switch(ret)
                 {
-                  /*LOG
-                    This message indicates that the returned value of the callback for the given response policy 
-                    is invalid and Zorp rejects the response. Check the callback function.
-                   */
-                  z_proxy_log(self, FTP_POLICY, 1, "Return code invalid from policy function; command='%s', answer='%s'", self->request_cmd->str, self->answer_cmd->str);
+                case FTP_RSP_ACCEPT:
+                case FTP_RSP_REJECT:
+                case FTP_RSP_ABORT:
+                  break;
+                  
+                case Z_DROP:
+                case Z_UNSPEC:
+                  ret = FTP_RSP_REJECT;
+                  break;
+                  
+                default:
                   g_string_assign(self->answer_cmd, "500");
-                  g_string_assign(self->answer_param, "Error parsing answer (bad policy)");
+                  g_string_assign(self->answer_param, "Error parsing answer, connection dropped.");
                   ret = FTP_RSP_ABORT;
-                }
-              else
-                {
-                  switch(ret)
-                    {
-                    case FTP_RSP_ACCEPT:
-                    case FTP_RSP_REJECT:
-                    case FTP_RSP_ABORT:
-                      break;
-                    case Z_DROP:
-                    case Z_UNSPEC:
-                      ret = FTP_RSP_REJECT;
-                      break;
-                    default:
-                      g_string_assign(self->answer_cmd, "500");
-                      g_string_assign(self->answer_param, "Error parsing answer, connection dropped.");
-                      ret = FTP_RSP_ABORT;
-                    }
+                  break;
                 }
             }
         }
       z_policy_unlock(self->super.thread);
       break;
+      
     default:
       g_string_assign(self->answer_cmd, "500");
       g_string_assign(self->answer_param, "Error parsing answer, connection dropped.");
       ret = FTP_RSP_ABORT;
       break;
     }
-  z_proxy_leave(self);
-  return ret;
+  z_proxy_return(self, ret);
 }
