@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -30,11 +31,13 @@
 
 #include <zorp/zorp.h>
 #include <zorp/proxy.h>
+#include <zorp/proxystack.h>
 #include <zorp/streamline.h>
 #include <zorp/authprovider.h>
 #include <zorp/dimhash.h>
 #include <zorp/poll.h>
 #include <zorp/attach.h>
+#include <zorp/blob.h>
 
 /* general limits applied to headers, etc. */
 #define HTTP_MAX_LINE           32768
@@ -84,6 +87,7 @@
 /* connection mode */
 #define HTTP_CONNECTION_CLOSE		0
 #define HTTP_CONNECTION_KEEPALIVE	1
+#define HTTP_CONNECTION_UNKNOWN		2
 
 /* request type */
 #define HTTP_REQTYPE_SERVER        0     /* simple server request */
@@ -95,9 +99,44 @@
 #define HTTP_REQ_FLG_CONNECT    8
 
 /* response protocol flags */
-#define HTTP_RESP_FLG_CONTINUE  1
-#define HTTP_RESP_FLG_NODATA    2
-#define HTTP_RESP_FLG_STOP      4
+#define HTTP_RESP_FLG_CONTINUE     1
+#define HTTP_RESP_FLG_SUPPRESS     2
+#define HTTP_RESP_FLG_DONTEXPECT   4
+#define HTTP_RESP_FLG_STOP         8
+
+/* HTTP header value flags */
+
+/* format flags */
+#define HTTP_HDR_FF_URL             (1 << 1)
+#define HTTP_HDR_FF_ANY             (1 << 2)
+
+/* character flags */
+#define HTTP_HDR_CF_ALPHA           (1 << 3)
+#define HTTP_HDR_CF_NUMERIC         (1 << 4)
+#define HTTP_HDR_CF_SPACE           (1 << 5)
+#define HTTP_HDR_CF_COMMA           (1 << 6)
+#define HTTP_HDR_CF_DOT             (1 << 7)
+#define HTTP_HDR_CF_BRACKET         (1 << 8)
+#define HTTP_HDR_CF_EQUAL           (1 << 9)
+#define HTTP_HDR_CF_DASH            (1 << 10)
+#define HTTP_HDR_CF_SLASH           (1 << 11)
+#define HTTP_HDR_CF_COLON           (1 << 12)
+#define HTTP_HDR_CF_SEMICOLON       (1 << 13)
+#define HTTP_HDR_CF_AT              (1 << 14)
+#define HTTP_HDR_CF_UNDERLINE       (1 << 15)
+#define HTTP_HDR_CF_AND             (1 << 16)
+#define HTTP_HDR_CF_BACKSLASH       (1 << 17)
+// *
+#define HTTP_HDR_CF_ASTERIX         (1 << 19)
+#define HTTP_HDR_CF_DOLLAR          (1 << 20)
+// #
+#define HTTP_HDR_CF_HASHMARK        (1 << 21)
+#define HTTP_HDR_CF_PLUS            (1 << 22)
+#define HTTP_HDR_CF_QUOTE           (1 << 23)
+#define HTTP_HDR_CF_QUESTIONMARK    (1 << 24)
+#define HTTP_HDR_CF_PERCENT         (1 << 25)
+#define HTTP_HDR_CF_TILDE           (1 << 26)
+#define HTTP_HDR_CF_EXCLAM          (1 << 27)
 
 /* error codes */
 #define HTTP_MSG_OK             0
@@ -123,6 +162,10 @@
 #define HTTP_LENGTH_NONE		-2
 #define HTTP_LENGTH_UNKNOWN		-1
 
+#define HTTP_TRANSFER_NORMAL             0
+#define HTTP_TRANSFER_TO_BLOB            1
+#define HTTP_TRANSFER_FROM_BLOB          2
+
 typedef struct _HttpProxy HttpProxy;
 typedef struct _HttpTransfer HttpTransfer;
 typedef struct _HttpHeader HttpHeader;
@@ -163,6 +206,8 @@ struct _HttpHeaders
 
 struct _HttpURL
 {
+  /* this is the original form of the URL local part in case canonicalization was disabled */
+  GString *original_local;
   /* all gstrings below might contain NUL characters as they store the URL-decoded form */
   GString *scheme;
   GString *user;
@@ -174,6 +219,13 @@ struct _HttpURL
   GString *fragment;
 };
 
+typedef struct _HttpElementInfo
+{
+  gchar *name;
+  guint32 flags;
+  gssize max_len; /* only used for headers */
+} HttpElementInfo;
+
 /* This structure represents an HTTP proxy */
 struct _HttpProxy
 {
@@ -191,6 +243,13 @@ struct _HttpProxy
   /* timeout we wait for a request */
   guint timeout_request;
 
+  /* timeout we wait for a response */
+  guint timeout_response;
+
+  guint rerequest_attempts;
+  gboolean request_data_stored;
+  ZBlob *request_data;
+
   /* request/response header-sets */
   HttpHeaders headers[EP_MAX];
   
@@ -262,6 +321,9 @@ struct _HttpProxy
   guint remote_port;
   
   gboolean use_default_port_in_transparent_mode;
+  
+  /* whether to use canonicalized URLs by default */
+  gboolean use_canonicalized_urls;
 
   /* specifies the default HTTP port, which is used when the port is not
    * specified in the URL */
@@ -326,6 +388,7 @@ struct _HttpProxy
 
   /* rewrite host header when redirection is done */
   gboolean rewrite_host_header;   
+  gboolean reset_on_close;
   
   /* require the existance of the Host: header */
   gboolean require_host_header;   
@@ -335,6 +398,7 @@ struct _HttpProxy
   
   /* 0: accept rfc incompliant headers, 1: require rfc compliance */
   gboolean strict_header_checking; 
+  guint strict_header_checking_action;
 
   /* parsed protocol version: 0x major minor (0x0009 0x0100 0x0101) */
   guint proto_version[EP_MAX];          
@@ -396,12 +460,20 @@ struct _HttpProxy
   
   /* the directory where error file templates are stored */
   GString *error_files_directory;
-  
-  /* information supplied by child proxies */
-  gint stack_decision;
-  GString *stack_info;
-  gsize content_length_hint;
-  gboolean content_length_hint_set;
+
+  /* Maximum allowed time between two forced authentication request */
+  gint max_auth_time;
+
+  /* Enable authentication cache based on Cookies. */
+  gboolean auth_by_cookie;
+
+  /* Do not authenticate this amount of sec. */
+  gint auth_cache_time;
+
+  /* Update the auth cache stamp at each request */
+  gboolean auth_cache_update;
+
+  GString *append_cookie;
 };
 
 extern ZClass HttpProxy__class;
@@ -412,14 +484,14 @@ typedef guint (*HttpHeaderFilter)(HttpProxy *self, GString *header_name, GString
 guint http_write(HttpProxy *self, guint side, gchar *buf, size_t buflen);
 gboolean http_connect_server(HttpProxy *self);
 
-gboolean http_data_transfer(HttpProxy *self, guint from, ZStream *from_stream, guint to, ZStream *to_stream, gboolean expect_data, gboolean suppress_data, HttpTransferPreambleFunc format_preamble);
+gboolean http_data_transfer(HttpProxy *self, gint transfer_type, guint from, ZStream *from_stream, guint to, ZStream *to_stream, gboolean expect_data, gboolean suppress_data, HttpTransferPreambleFunc format_preamble);
 
-int http_proto_request_lookup(char *req);
-int http_proto_response_lookup(char *resp);
 
 gboolean
 http_lookup_header(HttpHeaders *headers, gchar *what, HttpHeader **p);
 
+GHashTable *http_parse_header_cookie(HttpHeaders *hdrs);
+
 gboolean
 http_fetch_headers(HttpProxy *self, int side);
 
@@ -447,10 +519,16 @@ http_add_header(HttpHeaders *hdrs, gchar *name, gint name_len, gchar *value, gin
 void
 http_log_headers(HttpProxy *self, gint side, gchar *tag);
 
+
+gint http_filter_hash_compare(gconstpointer a, gconstpointer b);
+gint http_filter_hash_bucket(gconstpointer a);
+
 /* URL processing */
 
-gboolean http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invalid_hex_escape, gchar *url_str, const gchar **reason);
-gboolean http_format_url(HttpURL *url, GString *encode_buf, gboolean format_absolute, gboolean permit_unicode_url, const gchar **reason);
+gboolean http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invalid_hex_escape,
+                        gboolean permit_relative_url, gchar *url_str, const gchar **reason);
+gboolean http_format_url(HttpURL *url, GString *encode_buf, gboolean format_absolute, gboolean permit_unicode_url,
+                         gboolean canonicalized, const gchar **reason);
 void http_init_url(HttpURL *url);
 void http_destroy_url(HttpURL *url);
 
@@ -466,4 +544,50 @@ http_parse_version(HttpProxy *self, gint side, gchar *version_str);
 gboolean 
 http_handle_ftp_request(HttpProxy *self);
 
+void http_proto_init(void);
+
+/* inline functions */
+
+extern GHashTable *request_proto_hash;
+extern GHashTable *response_proto_hash;
+extern GHashTable *request_hdr_proto_hash;
+extern GHashTable *response_hdr_proto_hash;
+
+static inline guint
+http_proto_lookup_hash(GHashTable *hash, const gchar *index)
+{
+  HttpElementInfo *e;
+
+  e = g_hash_table_lookup(hash, index);
+  if (e)
+    return e->flags;
+  return 0;
+}
+
+static inline guint
+http_proto_request_lookup(const gchar *req)
+{
+  return http_proto_lookup_hash(request_proto_hash, req);
+}
+
+static inline guint
+http_proto_response_lookup(const gchar *resp)
+{
+  return http_proto_lookup_hash(response_proto_hash, resp);
+}
+
+static inline HttpElementInfo *
+http_proto_request_hdr_lookup(const gchar *req)
+{
+  return g_hash_table_lookup(request_hdr_proto_hash, req);
+}
+
+static inline HttpElementInfo *
+http_proto_response_hdr_lookup(const gchar *resp)
+{
+  return g_hash_table_lookup(response_hdr_proto_hash, resp);
+}
+
+
+
 #endif
