@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -80,7 +81,7 @@ http_ftp_fetch_response(HttpProxy *self, gint *status, gchar *msg, gsize msglen)
             This message indicates that the response given by the FTP server
             was too short, not even the mandatory status code was included.
            */
-          z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid FTP response, line too short; line='%.*s'", length, line);
+          z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid FTP response, line too short; line='%.*s'", (gint)length, line);
           return FALSE;
         }
       value = 0;
@@ -94,7 +95,7 @@ http_ftp_fetch_response(HttpProxy *self, gint *status, gchar *msg, gsize msglen)
                 response, the status code returned by the server was not
                 numeric.
                */
-              z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid FTP response, response code not numeric; line='%.*s'", length, line);
+              z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid FTP response, response code not numeric; line='%.*s'", (gint)length, line);
               return FALSE;
             }
           value = value + mul * (line[i] - '0');
@@ -102,8 +103,11 @@ http_ftp_fetch_response(HttpProxy *self, gint *status, gchar *msg, gsize msglen)
         }
       if (first)
         {
+          gint copy = MIN(msglen-1, length - 4 + 1);
+          
           *status = value;
-          g_strlcpy(msg, &line[4], MIN(msglen, length - 4 + 1));
+          memcpy(msg, &line[4], copy);
+          msg[copy] = 0;
         }
       else if (*status != value)
         {
@@ -111,7 +115,7 @@ http_ftp_fetch_response(HttpProxy *self, gint *status, gchar *msg, gsize msglen)
             This message indicates that the FTP server gave an invalid response as the
             status code changed from the one which was present on the first line.
            */
-          z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid FTP response, continuation line contains different status code; ftp_status='%d', line='%.*s'", *status, length, line);
+          z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid FTP response, continuation line contains different status code; ftp_status='%d', line='%.*s'", *status, (gint)length, line);
           return FALSE;
         }
       continuation = line[3] == '-';
@@ -312,9 +316,10 @@ http_ftp_initiate_passive_data(HttpProxy *self)
   peer = z_sockaddr_inet_new(ip, 256 * ftp_pasv_endpoint[4] + ftp_pasv_endpoint[5]);
   
   memset(&params, 0, sizeof(params));
-  params.tcp.timeout = 30000;
+  params.timeout = 30000;
   
-  self->ftp_data_attach = z_attach_new(self->super.session_id, ZD_PROTO_TCP, NULL, peer, &params, NULL, NULL, NULL);
+  self->ftp_data_attach = z_attach_new(&self->super, ZD_PROTO_TCP, NULL, peer, &params, NULL, NULL, NULL);
+  z_sockaddr_unref(peer);
   /* attach not started yet */  
   
   return TRUE;
@@ -336,6 +341,20 @@ http_ftp_initiate_active_data(HttpProxy *self G_GNUC_UNUSED)
   return FALSE;
 }
 
+/**
+ * http_ftp_cleanup_data:
+ * @self: HttpProxy instance
+ *
+ * This function is called to clean up the FTP data state, e.g. dispose
+ * self->ftp_data_attach.
+ **/
+static inline void
+http_ftp_cleanup_data(HttpProxy *self)
+{
+  z_attach_free(self->ftp_data_attach);
+  self->ftp_data_attach = NULL;
+}
+
 /**
  * http_ftp_complete_data:
  * @self: HttpProxy
@@ -349,44 +368,25 @@ static gboolean
 http_ftp_complete_data(HttpProxy *self, ZStream **data_stream)
 {
   ZConnection *conn;
+  gboolean success;
   
   /* FIXME: supports PASV only */
-  if (!z_attach_start(self->ftp_data_attach))
+  success = z_attach_start_block(self->ftp_data_attach, &conn);
+  http_ftp_cleanup_data(self);
+  
+  if (!success)
     {
-      z_attach_unref(self->ftp_data_attach);
-      self->ftp_data_attach = NULL;
       *data_stream = NULL;
       return FALSE;
     }
-  conn = z_attach_block(self->ftp_data_attach);
-  z_attach_unref(self->ftp_data_attach);
-  self->ftp_data_attach = NULL;
-  if (conn)
+  else
     {
       *data_stream = z_stream_ref(conn->stream);
       z_connection_destroy(conn, FALSE);
       return TRUE;
     }
-  else
-    {
-      /* could not connect */
-      return FALSE;
-    }
 }
 
-/**
- * http_ftp_cleanup_data:
- * @self: HttpProxy instance
- *
- * This function is called to clean up the FTP data state, e.g. dispose
- * self->ftp_data_attach.
- **/
-static inline void
-http_ftp_cleanup_data(HttpProxy *self)
-{
-  z_attach_unref(self->ftp_data_attach);
-  self->ftp_data_attach = NULL;
-}
 
 /**
  * http_ftp_format_response: 
@@ -676,7 +676,7 @@ http_handle_ftp_request(HttpProxy *self)
           GString *url = g_string_sized_new(0);
           
           g_string_append_c(self->request_url_parts.file, '/');
-          if (!http_format_url(&self->request_url_parts, url, TRUE, FALSE, &reason))
+          if (!http_format_url(&self->request_url_parts, url, TRUE, FALSE, TRUE, &reason))
             {
               g_string_assign(url, "Error formatting URL");
             }
@@ -762,23 +762,24 @@ http_handle_ftp_request(HttpProxy *self)
       z_proxy_log(self, HTTP_ERROR, 4, "Error establishing data connection to FTP server;");
       return FALSE;
     }
-  if (!http_ftp_fetch_response(self, &status, response_msg, sizeof(response_msg)) || status != 150)
+  if (!http_ftp_fetch_response(self, &status, response_msg, sizeof(response_msg)) || (status != 150 && status != 125))
     {
       /* file was successfully downloaded, finished */
       z_stream_shutdown(data_stream, SHUT_RDWR, NULL);
+      z_stream_close(data_stream, NULL);
       z_stream_unref(data_stream);
       g_string_sprintf(self->error_info, "FTP error: %d %s", status, response_msg);
       /*LOG
         This message indicates that the FTP server did not react with a
         150 response after the data connection was established.
        */
-      z_proxy_log(self, HTTP_ERROR, 4, "Error reading 150 from FTP server; status='%d', response='%s'", status, response_msg);
+      z_proxy_log(self, HTTP_ERROR, 4, "Error reading data channel confirmation from FTP server; status='%d', response='%s'", status, response_msg);
       return FALSE;
     }
   
   if (fetch == HTTP_FTP_FETCH_FILE)
     {
-      if (!http_data_transfer(self, EP_SERVER, data_stream, EP_CLIENT, self->super.endpoints[EP_CLIENT], TRUE, FALSE, http_ftp_format_response))
+      if (!http_data_transfer(self, HTTP_TRANSFER_NORMAL, EP_SERVER, data_stream, EP_CLIENT, self->super.endpoints[EP_CLIENT], TRUE, FALSE, http_ftp_format_response))
         {
           /* file was not successfully downloaded, finished */
           res = FALSE;
@@ -805,6 +806,7 @@ http_handle_ftp_request(HttpProxy *self)
 
   /* file was successfully downloaded, finished */
   z_stream_shutdown(data_stream, SHUT_RDWR, NULL);
+  z_stream_close(data_stream, NULL);
   z_stream_unref(data_stream);
 
   if (res)
