@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -117,6 +118,7 @@ HttpHeader *
 http_add_header(HttpHeaders *hdrs, gchar *name, gint name_len, gchar *value, gint value_len)
 {
   HttpHeader *h;
+  HttpHeader *orig;
 
   h = g_new0(HttpHeader, 1);
   h->name = g_string_sized_new(name_len + 1);
@@ -127,7 +129,7 @@ http_add_header(HttpHeaders *hdrs, gchar *name, gint name_len, gchar *value, gin
   g_string_assign_len(h->value, value, value_len);
   h->present = TRUE;
 
-  if (!g_hash_table_lookup(hdrs->hash, h->name->str))
+  if (!http_lookup_header(hdrs, h->name->str, &orig))
     {
       hdrs->list = g_list_prepend(hdrs->list, h);
       g_hash_table_insert(hdrs->hash, h->name->str, hdrs->list);
@@ -152,7 +154,7 @@ http_add_header(HttpHeaders *hdrs, gchar *name, gint name_len, gchar *value, gin
         }
       else
         {
-          z_log(NULL, HTTP_VIOLATION, 3, "Possible smuggle attack, removing header duplication; header='%.*s', value='%.*s'", name_len, name, value_len, value);
+          z_log(NULL, HTTP_VIOLATION, 3, "Possible smuggle attack, removing header duplication; header='%.*s', value='%.*s', prev_value='%.*s'", name_len, name, value_len, value, orig->value->len, orig->value->str);
         }
     }
   return h;
@@ -222,6 +224,113 @@ http_insert_headers(gchar *key, ZPolicyObj *f, HttpHeaders *hdrs)
     }
 }
 
+static gboolean
+http_check_header_charset(HttpProxy *self, gchar *header, guint flags, const gchar **reason)
+{
+  gint i;
+
+  *reason = FALSE;
+  if (flags & HTTP_HDR_FF_ANY)
+    return TRUE;
+
+  if (flags & HTTP_HDR_FF_URL)
+    {
+      HttpURL url;
+      gboolean success;
+
+      http_init_url(&url);
+      success = http_parse_url(&url, self->permit_unicode_url, self->permit_invalid_hex_escape, TRUE, header, reason);
+      http_destroy_url(&url);
+
+      return success;
+    }
+  for (i = 0; header[i]; i++)
+    {
+      gboolean ok;
+      guchar c = header[i];
+
+      ok = FALSE;
+      if ((flags & HTTP_HDR_CF_ALPHA) &&
+          ((c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_NUMERIC) &&
+          (c >= '0' && c <= '9'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_SPACE) &&
+          (c == ' '))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_COMMA) &&
+          (c == ','))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_DOT) &&
+          (c == '.'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_BRACKET) &&
+          (c == '[' || c == ']' || c == '{' || c == '}' || c == '(' || c == ')'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_EQUAL) &&
+          (c == '='))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_DASH) &&
+          (c == '-'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_SLASH) &&
+          (c == '/'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_COLON) &&
+          (c == ':'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_SEMICOLON) &&
+          (c == ';'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_AT) &&
+          (c == '@'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_UNDERLINE) &&
+          (c == '_'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_AND) &&
+          (c == '&'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_BACKSLASH) &&
+          (c == '\\'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_ASTERIX) &&
+          (c == '*'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_DOLLAR) &&
+          (c == '$'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_HASHMARK) &&
+          (c == '#'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_PLUS) &&
+          (c == '+'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_QUOTE) &&
+          (c == '"' || c == '\''))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_QUESTIONMARK) &&
+          (c == '?'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_PERCENT) &&
+          (c == '%'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_TILDE) &&
+          (c == '~'))
+        ok = TRUE;
+      else if ((flags & HTTP_HDR_CF_EXCLAM) &&
+          (c == '!'))
+        ok = TRUE;
+      if (!ok)
+        {
+          *reason = "Invalid character found";
+          return FALSE;
+        }
+    }
+  return TRUE;
+}
+
 gboolean
 http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 {
@@ -237,7 +346,6 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
       HttpHeader *h = (HttpHeader *) l->data;
       ZPolicyObj *f;
 
-
       if (filter)
 	action = filter(self, h->name, h->value);
       else
@@ -260,8 +368,7 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 	    {
 	      /* filter has no type field */
 	      z_policy_unlock(self->super.thread);
-	      z_proxy_leave(self);
-	      return FALSE;
+	      z_proxy_return(self, FALSE);
 	    }
 	  z_policy_unlock(self->super.thread);
 	  switch (filter_type)
@@ -272,8 +379,7 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 		{
 		  /* error parsing HTTP_POLICY_CALL rule */
                   z_policy_unlock(self->super.thread);
-		  z_proxy_leave(self);
-		  return FALSE;
+		  z_proxy_return(self, FALSE);
 		}
 	      res = z_policy_call_object(handler, 
 					 z_policy_var_build("(s#s#)", 
@@ -290,24 +396,27 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 		  z_proxy_log(self, HTTP_POLICY, 1, "Policy call-back for header returned invalid value; header='%s'", self->current_header_name->str);
 		  z_policy_var_unref(res);
 		  z_policy_unlock(self->super.thread);
-		  z_proxy_leave(self);
-		  return FALSE;
+		  z_proxy_return(self, FALSE);
 		}
 	      z_policy_var_unref(res);
 	      z_policy_unlock(self->super.thread);
               g_string_assign_len(h->name, self->current_header_name->str, self->current_header_name->len);
               g_string_assign_len(h->value, self->current_header_value->str, self->current_header_value->len);
 	      break;
+              
 	    case HTTP_HDR_INSERT:
 	      /* insert header that already exists */
 	      action = HTTP_HDR_ACCEPT;
 	      break;
+              
 	    case HTTP_HDR_ACCEPT:
 	      break;
+              
 	    case HTTP_HDR_REPLACE:
 	    case HTTP_HDR_DROP:
 	      action = HTTP_HDR_DROP;
 	      break;
+              
 	    case HTTP_HDR_CHANGE_NAME:
 	      z_policy_lock(self->super.thread);
 	      if (!z_policy_var_parse(f, "(is)", &filter_type, &name))
@@ -320,13 +429,13 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 		   */
 		  z_proxy_log(self, HTTP_POLICY, 1, "Invalid HTTP_HDR_CHANGE_NAME rule in header processing; header='%s'", self->current_header_name->str);
 		  z_policy_unlock(self->super.thread);
-		  z_proxy_leave(self);
-		  return FALSE;
+		  z_proxy_return(self, FALSE);
 		}
 	      g_string_assign(h->name, name);
 	      z_policy_unlock(self->super.thread);
 	      action = HTTP_HDR_ACCEPT;
 	      break;
+              
 	    case HTTP_HDR_CHANGE_VALUE:
 	      z_policy_lock(self->super.thread);
 	      if (!z_policy_var_parse(f, "(is)", &filter_type, &value))
@@ -339,13 +448,13 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 		   */
 		  z_proxy_log(self, HTTP_POLICY, 1, "Invalid HTTP_HDR_CHANGE_VALUE rule in header processing; header='%s'", self->current_header_name->str);
 		  z_policy_unlock(self->super.thread);
-		  z_proxy_leave(self);
-		  return FALSE;
+		  z_proxy_return(self, FALSE);
 		}
 	      g_string_assign(h->value, value);
 	      z_policy_unlock(self->super.thread);
 	      action = HTTP_HDR_ACCEPT;
 	      break;
+
 	    case HTTP_HDR_CHANGE_BOTH:
 	      z_policy_lock(self->super.thread);
 	      if (!z_policy_var_parse(f, "(iss)", &filter_type, &name, &value))
@@ -358,17 +467,18 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 		   */
 		  z_proxy_log(self, HTTP_POLICY, 1, "Invalid HTTP_HDR_CHANGE_BOTH rule in header processing; header='%s'", self->current_header_name->str);
 		  z_policy_unlock(self->super.thread);
-		  z_proxy_leave(self);
-		  return FALSE;
+		  z_proxy_return(self, FALSE);
 		}
 	      g_string_assign(h->name, name);
 	      g_string_assign(h->value, value);
 	      z_policy_unlock(self->super.thread);
 	      action = HTTP_HDR_ACCEPT;
 	      break;
+
 	    case HTTP_HDR_ABORT:
 	      action = HTTP_HDR_ABORT;
 	      break;
+              
 	    default:
 	      action = HTTP_HDR_ABORT;
 	      /*LOG
@@ -381,28 +491,68 @@ http_filter_headers(HttpProxy *self, guint side, HttpHeaderFilter filter)
 	      break;
 	    }
 	}
+
+      if (action == HTTP_HDR_ACCEPT && self->strict_header_checking)
+        {
+          HttpElementInfo *info;
+          const gchar *reason;
+
+          if (side == EP_CLIENT)
+            info = http_proto_request_hdr_lookup(h->name->str);
+          else
+            info = http_proto_response_hdr_lookup(h->name->str);
+
+          if (info)
+            {
+              if (info->max_len >= 0 && h->value->len > (guint) info->max_len)
+                {
+                  z_proxy_log(self, HTTP_VIOLATION, 3,
+                              "Header failed strict checking, value too long; header='%s', value='%s', length='%d', max_length='%d'",
+                              h->name->str, h->value->str, h->value->len, info->max_len);
+                  action = self->strict_header_checking_action;
+                  goto exit_check;
+                }
+              if (!http_check_header_charset(self, h->value->str, info->flags, &reason))
+                {
+                  z_proxy_log(self, HTTP_VIOLATION, 3,
+                              "Header failed strict checking, it contains invalid characters; header='%s', value='%s', reason='%s'",
+                              h->name->str, h->value->str, reason);
+                  action = self->strict_header_checking_action;
+                  goto exit_check;
+                }
+            exit_check:
+              ;
+            }
+          else
+            {
+              z_proxy_log(self, HTTP_VIOLATION, 3,
+                          "Header failed strict checking, it is unknown; header='%s', value='%s'",
+                          h->name->str, h->value->str);
+              action = self->strict_header_checking_action;
+            }
+        }
+
       switch (action)
 	{
 	case HTTP_HDR_ACCEPT:
 	  break;
+          
 	case HTTP_HDR_DROP:
 	  h->present = FALSE;
 	  break;
+
 	default:
 	  self->error_code = HTTP_MSG_POLICY_VIOLATION;
-	  z_proxy_leave(self);
-	  return FALSE;
+	  z_proxy_return(self, FALSE);
 	}
-
       l = g_list_next(l);
+
     }
 
   z_policy_lock(self->super.thread);
   g_hash_table_foreach(hash, (GHFunc) http_insert_headers, headers);
   z_policy_unlock(self->super.thread);
-  
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 gboolean
@@ -417,13 +567,9 @@ http_fetch_headers(HttpProxy *self, int side)
 
   z_proxy_enter(self);
   http_clear_headers(headers);
-
   /* check if we have to fetch request headers */
   if (self->proto_version[side] < 0x0100)
-    {
-      z_proxy_leave(self);
-      return TRUE;
-    }
+    z_proxy_return(self, TRUE);
   
   while (1)
     {
@@ -440,17 +586,14 @@ http_fetch_headers(HttpProxy *self, int side)
 	    from the server.  Check the permit_null_response attribute.
 	   */
           z_proxy_log(self, HTTP_ERROR, 3, "Error reading from peer while fetching headers;");
-          z_proxy_leave(self);
-	  return FALSE;
+	  z_proxy_return(self, FALSE);
 	}
       if (line_length == 0)
-        {
-          break;
-        }
+        break;
+
       if (*line == ' ' || *line == '\t')
         {
           /* continuation line */
-          
           /* skip whitespace */
           while (line_length && (*line == ' ' || *line == '\t'))
             {
@@ -469,8 +612,7 @@ http_fetch_headers(HttpProxy *self, int side)
 		It is likely caused by a buggy server.
 	       */
               z_proxy_log(self, HTTP_VIOLATION, 2, "First header starts with white space; line='%.*s", (gint) line_length, line);
-              z_proxy_leave(self);
-              return FALSE;
+              z_proxy_return(self, FALSE);
             }
           goto next_header;
         }
@@ -497,12 +639,9 @@ http_fetch_headers(HttpProxy *self, int side)
 	   */
           z_proxy_log(self, HTTP_VIOLATION, 2, "Invalid HTTP header; line='%.*s'", (gint) line_length, line);
           if (self->strict_header_checking)
-            {
-              z_proxy_leave(self);
-	      return FALSE;
-	    }
-	  else
-	    goto next_header;
+            z_proxy_return(self, FALSE);
+
+          goto next_header;
         }
       /* strip trailing white space */
       while (line_length > 0 && line[line_length - 1] == ' ') 
@@ -512,8 +651,8 @@ http_fetch_headers(HttpProxy *self, int side)
         ;
 
       last_hdr = http_add_header(headers, line, name_len, value, &line[line_length] - value);
-    next_header:
 
+    next_header:
       count++;
       if (count > self->max_header_lines)
         {
@@ -524,17 +663,12 @@ http_fetch_headers(HttpProxy *self, int side)
 	    attribute.
 	   */
           z_proxy_log(self, HTTP_POLICY, 2, "Too many header lines; max_header_lines='%d'", self->max_header_lines);
-          z_proxy_leave(self);
-          return FALSE;
+          z_proxy_return(self, FALSE);
         }
-
     }
   /*  g_string_append(headers, "\r\n"); */
-  
   http_log_headers(self, side, "prefilter");
-  
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 gboolean
@@ -546,7 +680,12 @@ http_flat_headers_into(HttpHeaders *hdrs, GString *into)
   while (l)
     {
       if (((HttpHeader *) l->data)->present)
-        g_string_sprintfa(into, "%s: %s\r\n", ((HttpHeader *) l->data)->name->str, ((HttpHeader *) l->data)->value->str);
+        {
+          g_string_append_len(into, ((HttpHeader *) l->data)->name->str, ((HttpHeader *) l->data)->name->len);
+          g_string_append_len(into, ": ", 2);
+          g_string_append_len(into, ((HttpHeader *) l->data)->value->str, ((HttpHeader *) l->data)->value->len);
+          g_string_append_len(into, "\r\n", 2);
+        }
       l = g_list_previous(l);
     }
 
@@ -573,3 +712,101 @@ http_destroy_headers(HttpHeaders *hdrs)
   g_hash_table_destroy(hdrs->hash);
   g_string_free(hdrs->flat, TRUE);
 }
+
+enum
+{
+  HTTP_COOKIE_NAME,
+  HTTP_COOKIE_VALUE,
+  HTTP_COOKIE_DOTCOMA
+} HttpCookieState;
+
+GHashTable *
+http_parse_header_cookie(HttpHeaders *hdrs)
+{
+  GHashTable *cookie_hash = g_hash_table_new_full(g_str_hash, g_str_equal, g_free, g_free);
+  HttpHeader *hdr;
+  if (http_lookup_header(hdrs, "Cookie", &hdr))
+    {
+      gint state = HTTP_COOKIE_NAME;
+      gchar key[256];
+      guint key_pos = 0;
+      gchar value[4096];
+      guint value_pos = 0;
+      gint i = 0;
+      gchar *hdr_str = hdr->value->str;
+
+      while (hdr_str[i])
+        {
+          switch(state)
+            {
+            case HTTP_COOKIE_NAME:
+              if (hdr_str[i] == '=')
+                {
+                  key[key_pos] = 0;
+                  state = HTTP_COOKIE_VALUE;
+                }
+              else
+                {
+                  key[key_pos++] = hdr_str[i];
+                }
+              if (key_pos > sizeof(key))
+                {
+                  goto exit_error;
+                }
+              break;
+            case HTTP_COOKIE_VALUE:
+              if (hdr_str[i] == ';')
+                {
+                  state = HTTP_COOKIE_DOTCOMA;
+                }
+              else
+                {
+                  value[value_pos++] = hdr_str[i];
+                }
+              if (value_pos > sizeof(value))
+                {
+                  goto exit_error;
+                }
+              break;
+            case HTTP_COOKIE_DOTCOMA:
+              if (hdr_str[i] != ' ' &&
+                  hdr_str[i] != '\r' &&
+                  hdr_str[i] != '\n' &&
+                  hdr_str[i] != '\t')
+                {
+                  if (hdr_str[i] == '$' && FALSE)
+                    {
+                      value[value_pos++] = hdr_str[i];
+                      if (value_pos > sizeof(value))
+                        {
+                          goto exit_error;
+                        }
+                      state = HTTP_COOKIE_VALUE;
+                    }
+                  else
+                    {
+                      value[value_pos] = 0;
+                      g_hash_table_insert(cookie_hash, g_strdup(key), g_strdup(value));
+                      key_pos = value_pos = 0;
+                      key[key_pos++] = hdr_str[i];
+                      state = HTTP_COOKIE_NAME;
+                    }
+                }
+            }
+          i++;
+        }
+      if (key_pos && value_pos)
+        {
+          value[value_pos] = 0;
+          g_hash_table_insert(cookie_hash, g_strdup(key), g_strdup(value));
+          key_pos = value_pos = 0;
+        }
+      goto exit;
+    }
+exit_error:
+  g_hash_table_destroy(cookie_hash);
+  cookie_hash = NULL;
+exit:
+
+  return cookie_hash;
+}
