@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -593,14 +594,14 @@ http_string_assign_url_canonicalize_unicode(GString *result, gboolean permit_inv
  * Returns: TRUE for success, FALSE otherwise setting @reason to the explanation
  **/
 gboolean
-http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invalid_hex_escape, gchar *url_str, const gchar **reason)
+http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invalid_hex_escape,
+               gboolean permit_relative_url, gchar *url_str, const gchar **reason)
 {
   gchar *p, *end, *part[4], *sep[4], *query_start, *fragment_start, *file_start;
   gsize file_len, query_len = 0, fragment_len = 0;
   int i;
 
   z_enter();
-  
   g_string_truncate(url->scheme, 0);
   g_string_truncate(url->user, 0);
   g_string_truncate(url->passwd, 0);
@@ -613,20 +614,21 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
   p = url_str;
   while (*p && *p != ':')
     p++;
-    
   if (!*p)
     {
-      *reason = "URL has no scheme, colon missing";
-      z_leave();
-      return FALSE;
+      if (!permit_relative_url)
+        {
+          *reason = "URL has no scheme, colon missing";
+          z_return(FALSE);
+        }
+      else
+        goto relative_url;
     }
-
   if (*(p + 1) != '/' || *(p + 2) != '/')
     {
       /* protocol not terminated by '//' */
       *reason = "Scheme not followed by '//'";
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
   g_string_assign_len(url->scheme, url_str, p - url_str);
   p += 3;
@@ -647,27 +649,22 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
     case 0:
       /* hostname only */
       if (!http_string_assign_url_decode(url->host, permit_invalid_hex_escape, part[0], sep[0] - part[0], reason))
-        {
-          z_leave();
-          return FALSE;
-        }
+        z_return(FALSE);
+
       break;
+
     case 1:
       /* username && host || hostname && port number */
       if (*sep[0] == ':')
         {
           if (!http_string_assign_url_decode(url->host, permit_invalid_hex_escape, part[0], sep[0] - part[0], reason))
-            {
-              z_leave();
-              return FALSE;
-            }
+            z_return(FALSE);
           /* port number */
           url->port = strtoul(part[1], &end, 10);
           if (end != sep[1])
             {
               *reason = "Error parsing port number";
-              z_leave();
-              return FALSE;
+              z_return(FALSE);
             }
         }
       else if (*sep[0] == '@')
@@ -675,17 +672,14 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
           /* username */
           if (!http_string_assign_url_decode(url->user, permit_invalid_hex_escape, part[0], sep[0] - part[0], reason) ||
               !http_string_assign_url_decode(url->host, permit_invalid_hex_escape, part[1], sep[1] - part[1], reason))
-            {
-              z_leave();
-              return FALSE;
-            }
+            z_return(FALSE);
         }
       else
         {
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
       break;
+
     case 2:
       /* username && host && port || username && password && host */
       if (*sep[0] == '@' && *sep[1] == ':')
@@ -693,16 +687,12 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
           /* username, host, port */
           if (!http_string_assign_url_decode(url->user, permit_invalid_hex_escape, part[0], sep[0] - part[0], reason) ||
               !http_string_assign_url_decode(url->host, permit_invalid_hex_escape, part[1], sep[1] - part[1], reason))
-            {
-              z_leave();
-              return FALSE;
-            }
+            z_return(FALSE);
           url->port = strtoul(part[2], &end, 10);
           if (end != sep[2])
             {
               *reason = "Error parsing port number";
-              z_leave();
-              return FALSE;
+              z_return(FALSE);
             }
         }
       else if (*sep[0] == ':' && *sep[1] == '@')
@@ -711,17 +701,14 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
           if (!http_string_assign_url_decode(url->user, permit_invalid_hex_escape, part[0], sep[0] - part[0], reason) ||
               !http_string_assign_url_decode(url->passwd, permit_invalid_hex_escape, part[1], sep[1] - part[1], reason) ||
               !http_string_assign_url_decode(url->host, permit_invalid_hex_escape, part[2], sep[2] - part[2], reason))
-            {
-              z_leave();
-              return FALSE;
-            }
+            z_return(FALSE);
         }
       else
         {
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
       break;
+
     case 3:
       /* username && password && hostname && port */
       if (*sep[0] == ':' && *sep[1] == '@' && *sep[2] == ':')
@@ -729,42 +716,40 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
           if (!http_string_assign_url_decode(url->user, permit_invalid_hex_escape, part[0], sep[0] - part[0], reason) ||
               !http_string_assign_url_decode(url->passwd, permit_invalid_hex_escape, part[1], sep[1] - part[1], reason) ||
               !http_string_assign_url_decode(url->host, permit_invalid_hex_escape, part[2], sep[2] - part[2], reason))
-            {
-              z_leave();
-              return FALSE;
-            }
+            z_return(FALSE);
           url->port = strtoul(part[3], &end, 10);
           if (end != sep[3])
             {
               *reason = "Error parsing port number";
-              z_leave();
-              return FALSE;
+              z_return(FALSE);
             }
         }
       else
         {
-          z_leave();
-          return FALSE;
+          z_return(FALSE);
         }
       break;
+
     default:
       /* not reached */
-      z_leave();
-      return FALSE;
+      z_return(FALSE);
     }
 
+ relative_url:
+
   file_start = p;
   if (*file_start != '/')
     {
       if (*file_start == '\0')
         {
           g_string_assign(url->file, "/");
-          return TRUE;
+          z_return(TRUE);
         }
       *reason = "Invalid path component in URL";
-      return FALSE;
+      z_return(FALSE);
     }
-    
+  g_string_assign(url->original_local, file_start);
+
   query_start = strchr(p, '?');
   fragment_start = strchr(p, '#');
   if (query_start && fragment_start)
@@ -772,7 +757,7 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
       if (query_start > fragment_start)
         {
           *reason = "The fragment part starts earlier than the query";
-          return FALSE;
+          z_return(FALSE);
         }
       file_len = query_start - file_start;
       query_start++;
@@ -799,20 +784,18 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
   
   if (!(permit_unicode_url ? http_string_assign_url_decode_unicode : http_string_assign_url_decode)
              (url->file, permit_invalid_hex_escape, file_start, file_len, reason))
-    return FALSE;
+    z_return(FALSE);
     
   /* query and fragment is not url-decoded as it is impossible to get the original back */
-
   if (query_start && !(permit_unicode_url ? http_string_assign_url_canonicalize_unicode : http_string_assign_url_canonicalize)
                (url->query, permit_invalid_hex_escape, HTTP_URL_QUERY_ESCAPE_CHARS, query_start, query_len, reason))
-    return FALSE;
+    z_return(FALSE);
 
   if (fragment_start && !(permit_unicode_url ? http_string_assign_url_canonicalize_unicode : http_string_assign_url_canonicalize)
                  (url->fragment, permit_invalid_hex_escape, HTTP_URL_FRAGMENT_ESCAPE_CHARS, fragment_start, fragment_len, reason))
-    return FALSE;
+    z_return(FALSE);
     
-  z_leave();
-  return TRUE;
+  z_return(TRUE);
 }
 
 /**
@@ -826,7 +809,7 @@ http_parse_url(HttpURL *url, gboolean permit_unicode_url, gboolean permit_invali
  * Reformat the already parsed URL from its internal representation.
  **/
 gboolean
-http_format_url(HttpURL *url, GString *encode_buf, gboolean format_absolute, gboolean permit_unicode_url, const gchar **reason)
+http_format_url(HttpURL *url, GString *encode_buf, gboolean format_absolute, gboolean permit_unicode_url, gboolean canonicalized, const gchar **reason)
 {
   if (format_absolute)
     {
@@ -848,21 +831,28 @@ http_format_url(HttpURL *url, GString *encode_buf, gboolean format_absolute, gbo
         g_string_sprintfa(encode_buf, ":%d", url->port);
     }
 
-  if (!(permit_unicode_url ? http_string_append_url_encode_unicode : http_string_append_url_encode)
-          (encode_buf, HTTP_URL_FILE_ESCAPE_CHARS, url->file->str, url->file->len, reason))
+  if (!canonicalized)
     {
-      return FALSE;
+      g_string_append(encode_buf, url->original_local->str);
     }
-      
-  if (url->query->len)
+  else
     {
-      g_string_append_c(encode_buf, '?');
-      g_string_append(encode_buf, url->query->str);
-    }
-  if (url->fragment->len)
-    {
-      g_string_append_c(encode_buf, '#');
-      g_string_append(encode_buf, url->fragment->str);
+      if (!(permit_unicode_url ? http_string_append_url_encode_unicode : http_string_append_url_encode)
+              (encode_buf, HTTP_URL_FILE_ESCAPE_CHARS, url->file->str, url->file->len, reason))
+        {
+          return FALSE;
+        }
+          
+      if (url->query->len)
+        {
+          g_string_append_c(encode_buf, '?');
+          g_string_append(encode_buf, url->query->str);
+        }
+      if (url->fragment->len)
+        {
+          g_string_append_c(encode_buf, '#');
+          g_string_append(encode_buf, url->fragment->str);
+        }
     }
   return TRUE;
 }
@@ -876,6 +866,7 @@ http_format_url(HttpURL *url, GString *encode_buf, gboolean format_absolute, gbo
 void
 http_init_url(HttpURL *url)
 {
+  url->original_local = g_string_sized_new(64);
   url->scheme = g_string_sized_new(4);
   url->user = g_string_sized_new(0);
   url->passwd = g_string_sized_new(0);
@@ -894,6 +885,7 @@ http_init_url(HttpURL *url)
 void
 http_destroy_url(HttpURL *url)
 {
+  g_string_free(url->original_local, TRUE);
   g_string_free(url->scheme, TRUE);
   g_string_free(url->user, TRUE);
   g_string_free(url->passwd, TRUE);
@@ -947,18 +939,13 @@ http_split_request(HttpProxy *self, gchar *line, gint length)
   g_string_truncate(self->request_method, 0);
   self->request_flags = -1;
   self->request_version[0] = 0;
-
   g_string_truncate(self->request_url, 0);
-
   src = line;
   left = length;
   dst = self->request_method->str;
   avail = self->request_method->allocated_len;
-
   COPY_SPACE;
-  
   self->request_method->len = strlen(self->request_method->str);
-
   if (!self->request_method->len || (*src != ' ' && avail == 0))
     {
       /*LOG
@@ -966,16 +953,13 @@ http_split_request(HttpProxy *self, gchar *line, gint length)
         invalid.
        */
       z_proxy_log(self, HTTP_VIOLATION, 1, "Request method empty, or too long; line='%.*s'", left, src);
-      z_proxy_leave(self);
       /* request method empty, or request buffer overflow */
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
 
   SKIP_SPACES;
-
   avail = self->max_url_length;
   g_string_truncate(self->request_url, 0);
-
   while (left > 0 && avail > 0 && *src != ' ' && *src) 
     { 
       g_string_append_c(self->request_url, *src++);
@@ -990,17 +974,13 @@ http_split_request(HttpProxy *self, gchar *line, gint length)
         This message indicates that the URL sent by the client is invalid.
        */
       z_proxy_log(self, HTTP_VIOLATION, 1, "URL missing, or too long; line='%.*s'", left, src);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
 
   SKIP_SPACES;
-
   dst = self->request_version;
   avail = sizeof(self->request_version) - 1;
-  
   COPY_SPACE;
-
   if (*src != ' ' && avail == 0)
     {
       /* protocol version too long */
@@ -1009,16 +989,14 @@ http_split_request(HttpProxy *self, gchar *line, gint length)
         is invalid.
        */
       z_proxy_log(self, HTTP_VIOLATION, 1, "Protocol version missing, or too long; line='%.*s'", left, src);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
   
   /*LOG
     This message reports the processed request details.
    */
   z_proxy_log(self, HTTP_REQUEST, 6, "Request details; command='%s', url='%s', version='%s'", self->request_method->str, self->request_url->str, self->request_version);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /**
@@ -1040,15 +1018,11 @@ http_split_response(HttpProxy *self, gchar *line, gint line_length)
   self->response_version[0] = 0;
   self->response[0] = 0;
   g_string_truncate(self->response_msg, 0);
-  
   src = line;
   left = line_length;
-  
   dst = self->response_version;
   avail = sizeof(self->response_version) - 1;
-
   COPY_SPACE;
-  
   if (memcmp(self->response_version, "HTTP", 4) != 0)
     {
       /* no status line */
@@ -1056,8 +1030,7 @@ http_split_response(HttpProxy *self, gchar *line, gint line_length)
         This message indicates that the server sent an invalid response status line.
        */
       z_proxy_log(self, HTTP_RESPONSE, 6, "Invalid HTTP status line; line='%s'", dst);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
 
   if (!self->response_version[0] || (*src != ' ' && avail == 0))
@@ -1068,17 +1041,13 @@ http_split_response(HttpProxy *self, gchar *line, gint line_length)
         is invalid.
        */
       z_proxy_log(self, HTTP_VIOLATION, 1, "Response version empty or too long; line='%.*s'", line_length, line);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
   
   SKIP_SPACES;
-  
   dst = self->response;
   avail = sizeof(self->response) - 1;
-
   COPY_SPACE;
-  
   if (!self->response[0] || (*src != ' ' && left && avail == 0))
     {
       /* response code empty or too long */
@@ -1087,16 +1056,12 @@ http_split_response(HttpProxy *self, gchar *line, gint line_length)
         invalid.
        */
       z_proxy_log(self, HTTP_VIOLATION, 1, "Response code empty or too long; line='%.*s'", line_length, line);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
   
   self->response_code = atoi(self->response);
-    
   SKIP_SPACES;
-
   avail = 256;
-
   while (left > 0 && avail > 0) 
     { 
       g_string_append_c(self->response_msg, *src);
@@ -1110,8 +1075,7 @@ http_split_response(HttpProxy *self, gchar *line, gint line_length)
     This message reports the processed response details.
    */
   z_proxy_log(self, HTTP_RESPONSE, 7, "Response details; version='%s', response_code='%s', response_msg='%s'", self->response_version, self->response, self->response_msg->str);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 #undef SKIP_SPACES
@@ -1158,9 +1122,7 @@ http_parse_version(HttpProxy *self, gint side, gchar *version_str)
 	 */
         z_proxy_log(self, HTTP_RESPONSE, 3, "Unknown protocol version; version='%s'", version_str);
       self->proto_version[side] = 0x0100;
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
