@@ -1,6 +1,7 @@
 ############################################################################
 ##
-## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+## 2010 BalaBit IT Ltd, Budapest, Hungary
 ##
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -41,6 +42,11 @@ SSL_VERIFY_OPTIONAL           = 1
 SSL_VERIFY_REQUIRED_UNTRUSTED = 2
 SSL_VERIFY_REQUIRED_TRUSTED   = 3
 
+PSSL_VERIFY_NONE               = 0
+PSSL_VERIFY_OPTIONAL           = 1
+PSSL_VERIFY_REQUIRED_UNTRUSTED = 2
+PSSL_VERIFY_REQUIRED_TRUSTED   = 3
+
 
 class AbstractPsslProxy(Proxy):
 	"""Encapsulates an SSL enabled plug proxy.
@@ -164,7 +170,10 @@ class AbstractPsslProxy(Proxy):
 					 certificate chain depth during server
 					 verification.
 
-	  timeout                     -- [INTEGER:-1:RW:R] I/O timeout in
+	  timeout                     -- [INTEGER:600000:RW:R] I/O timeout in
+					 millisecundum.
+
+	  handshake_timeout           -- [INTEGER:30000:RW:R] I/O timeout in
 					 millisecundum.
 
           shutdown_soft               -- [BOOLEAN:FALSE:RW:R] Set TRUE if
