@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -70,8 +71,9 @@ typedef struct _PsslProxy
   GString *server_ca_dir;
   GString *server_crl_dir;
   gboolean need_ssl[EP_MAX];
-  int verify_type[EP_MAX];
-  int verify_depth[EP_MAX];
+  gint handshake_timeout;
+  gint verify_type[EP_MAX];
+  gint verify_depth[EP_MAX];
 } PsslProxy;
 
 extern ZClass PsslProxy__class;
@@ -79,9 +81,13 @@ extern ZClass PsslProxy__class;
 #define ENABLE_PKT_STATS 1
 
 static gboolean
-pssl_packet_stat_event(ZProxy *s, ZSessionVars *vars, 
+pssl_packet_stat_event(ZPlugSession *session, 
                        guint64 client_bytes, guint64 client_pkts, 
-                       guint64 server_bytes, guint64 server_pkts);
+                       guint64 server_bytes, guint64 server_pkts,
+                       gpointer user_data);
+
+static void pssl_finish(ZPlugSession *session, gpointer user_data);
+
 
 static void
 pssl_config_set_defaults(PsslProxy *self)
@@ -89,13 +95,14 @@ pssl_config_set_defaults(PsslProxy *self)
   gint i;
   
   z_proxy_enter(self);
-
-  self->session_data.timeout = -1;
+  self->handshake_timeout = 30000;
+  self->session_data.timeout = 600000;
   self->session_data.copy_to_server = TRUE;
   self->session_data.copy_to_client = TRUE;
   self->session_data.buffer_size = 1500;
   self->session_data.shutdown_soft = TRUE;
   self->session_data.packet_stats = pssl_packet_stat_event;
+  self->session_data.finish = pssl_finish;
           
   self->poll = z_poll_new();
   
@@ -116,10 +123,8 @@ pssl_config_set_defaults(PsslProxy *self)
   self->verify_depth[EP_SERVER] = 1;
   self->verify_depth[EP_CLIENT] = 1;
   for (i = 0; i < EP_MAX; i++)
-    {
-      self->need_ssl[i] = TRUE;
-    }
-  z_proxy_leave(self);
+    self->need_ssl[i] = TRUE;
+  z_proxy_return(self);
 }
 
 static void
@@ -131,6 +136,11 @@ pssl_register_vars(PsslProxy *self)
                   Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
                   &self->session_data.timeout);
 
+  z_proxy_var_new(&self->super,
+                  "handshake_timeout",
+                  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
+                  &self->handshake_timeout);
+
   z_proxy_var_new(&self->super,
                   "copy_to_client",
                   Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_INT,
@@ -258,20 +268,21 @@ pssl_register_vars(PsslProxy *self)
                   "server_cert",
 		  Z_VAR_GET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_ALIAS,
 		  "server_cert_file");
+  z_proxy_return(self);
 }
 
 static gboolean
-pssl_packet_stat_event(ZProxy *s, ZSessionVars *vars, 
+pssl_packet_stat_event(ZPlugSession *session G_GNUC_UNUSED, 
                        guint64 client_bytes, guint64 client_pkts, 
-                       guint64 server_bytes, guint64 server_pkts)
+                       guint64 server_bytes, guint64 server_pkts,
+                       gpointer user_data)
 {
-  PsslProxy *self = (PsslProxy *) s;
+  PsslProxy *self = (PsslProxy *)user_data;
   ZPolicyObj *res;
   gboolean called;
   guint resc;
 
   z_policy_lock(self->super.thread);
-  z_proxy_vars_set_active_session(s, vars);
   res = z_policy_call(self->super.handler, "packetStats",
                       z_policy_var_build("iiii",
                                          (guint32) client_bytes,
@@ -280,7 +291,6 @@ pssl_packet_stat_event(ZProxy *s, ZSessionVars *vars,
                                          (guint32) server_pkts),
                       &called,
                       self->super.session_id);
-  z_proxy_vars_set_active_session(s, NULL);
   if (called)
     {
       resc = Z_REJECT;
@@ -297,14 +307,24 @@ pssl_packet_stat_event(ZProxy *s, ZSessionVars *vars,
         }
     }
   else
-    resc = Z_ACCEPT;
+    {
+      resc = Z_ACCEPT;
+    }
   z_policy_var_unref(res);
   z_policy_unlock(self->super.thread);
   return resc == Z_ACCEPT;
 }
     
+static void
+pssl_finish(ZPlugSession *session G_GNUC_UNUSED, gpointer user_data)
+{
+  PsslProxy *self = (PsslProxy *)user_data;
+  
+  z_poll_quit(self->poll);
+}
+
 guint
-pssl_init_client_ssl(PsslProxy *self)
+pssl_perform_handshake(PsslProxy  *self, gint side)
 {
   ZSSLSession *ssl;
   X509 *peercert;
@@ -312,101 +332,40 @@ pssl_init_client_ssl(PsslProxy *self)
   ZStream *tmpstream;
   
   z_proxy_enter(self);
-  
   ssl = z_ssl_session_new(self->super.session_id, 
-                          Z_SSL_MODE_SERVER,
+                          side == EP_CLIENT ? Z_SSL_MODE_SERVER : Z_SSL_MODE_CLIENT,
                           self->client_keyfile->str,
                           self->client_certfile->str,
                           self->client_ca_dir->str,
                           self->client_crl_dir->str,
                           self->verify_depth[EP_CLIENT],
                           self->verify_type[EP_CLIENT]);
+
   if (!ssl)
     {
-      z_proxy_log(self, PSSL_ERROR, 1, "Error initializing SSL session on the client side;");
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_log(self, PSSL_ERROR, 1, "Error initializing SSL session; side='%s'", EP_STR(side));
+      z_proxy_return(self, FALSE);
     }
-
   SSL_set_options(ssl->ssl, SSL_MODE_ENABLE_PARTIAL_WRITE);
 
-  tmpstream = self->super.endpoints[EP_CLIENT];
-  self->super.endpoints[EP_CLIENT] = z_stream_ssl_new(tmpstream, ssl);
+  z_stream_set_timeout(self->super.endpoints[side], self->handshake_timeout);
+  tmpstream = self->super.endpoints[side];
+  self->super.endpoints[side] = z_stream_ssl_new(tmpstream, ssl);
   z_stream_unref(tmpstream);
 
-  z_stream_set_timeout(self->super.endpoints[EP_CLIENT], self->session_data.timeout);
+  if (side == EP_CLIENT)
+    ret = SSL_accept(ssl->ssl);
+  else
+    ret = SSL_connect(ssl->ssl);
 
-  ret = SSL_accept(ssl->ssl);
-  if (ret <= 0)
-    {
-      char buf[1024];
-
-      z_ssl_session_unref(ssl);      
-      z_proxy_log(self, PSSL_ERROR, 1, "SSL handshake failed on the client side; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
-      z_proxy_leave(self);
-      return FALSE;
-    }
-  
-  peercert = SSL_get_peer_certificate(ssl->ssl);
-  if (peercert)
-    {
-      gchar tmp[1024];
-      
-      X509_NAME_oneline(X509_get_subject_name(peercert), tmp, sizeof(tmp) - 1);
-      X509_free(peercert);
-      z_proxy_log(self, PSSL_DEBUG, 4, "Identified peer on the client side; peer='%s'", tmp);
-    }
-
-  z_ssl_session_unref(ssl);  
-  z_proxy_log(self, PSSL_DEBUG, 6, "Client side SSL handshake successful;");
-  z_proxy_leave(self);
-  return TRUE;
-}
-
-guint
-pssl_init_server_ssl(PsslProxy *self)
-{
-  ZSSLSession *ssl;
-  X509 *peercert;
-  int ret;
-  ZStream *tmpstream;
-  
-  z_proxy_enter(self);
-  
-  ssl = z_ssl_session_new(self->super.session_id, 
-                          Z_SSL_MODE_CLIENT,
-                          self->server_keyfile->str,
-                          self->server_certfile->str,
-                          self->server_ca_dir->str,
-                          self->server_crl_dir->str,
-                          self->verify_depth[EP_SERVER],
-                          self->verify_type[EP_SERVER]);
-  if (!ssl)
-    {
-      z_proxy_log(self, PSSL_ERROR, 1, "Error initializing SSL session on the server side;");
-      z_proxy_leave(self);
-      return FALSE;
-    }
-	
-  SSL_set_options(ssl->ssl, SSL_MODE_ENABLE_PARTIAL_WRITE);
-
-  tmpstream = self->super.endpoints[EP_SERVER];
-  self->super.endpoints[EP_SERVER] = z_stream_ssl_new(tmpstream, ssl);
-  z_stream_unref(tmpstream);
-  
-  z_stream_set_timeout(self->super.endpoints[EP_SERVER], self->session_data.timeout);
-
-  ret = SSL_connect(ssl->ssl);
   if (ret <= 0)
     {
       char buf[1024];
       
       z_ssl_session_unref(ssl);
-      z_proxy_log(self, PSSL_ERROR, 1, "SSL handshake failed on the server side; error='%s'", z_ssl_get_error_str(buf, sizeof(buf)));
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_log(self, PSSL_ERROR, 1, "SSL handshake failed; side='%s', error='%s'", EP_STR(side), z_ssl_get_error_str(buf, sizeof(buf)));
+      z_proxy_return(self, FALSE);
     }
-  
   peercert = SSL_get_peer_certificate(ssl->ssl);
   if (peercert)
     {
@@ -414,66 +373,45 @@ pssl_init_server_ssl(PsslProxy *self)
       
       X509_NAME_oneline(X509_get_subject_name(peercert), tmp, sizeof(tmp));
       X509_free(peercert);
-      z_proxy_log(self, PSSL_DEBUG, 4, "Identified peer on the server side; peer='%s'", tmp);
+      z_proxy_log(self, PSSL_DEBUG, 4, "Identified peer; side='%s', peer='%s'", EP_STR(side), tmp);
     }
     
-  z_ssl_session_unref(ssl);
-  z_proxy_log(self, PSSL_DEBUG, 6, "Server side SSL handshake successful;");
-  z_proxy_leave(self);
-  return TRUE;
+  z_ssl_session_unref(ssl);  
+  z_stream_set_timeout(self->super.endpoints[side], -2);
+  z_proxy_log(self, PSSL_DEBUG, 6, "SSL handshake successful; side='%s'", EP_STR(side));
+  z_proxy_return(self, TRUE);
+
 }
 
 static gboolean
 pssl_init_streams(PsslProxy *self)
 {
-  ZStream *tmpstream1;
-  ZStream *tmpstream2;
+  ZStream *tmpstream;
+  gint ep;
 
   z_proxy_enter(self);
-
   if (!self->super.endpoints[EP_CLIENT] || 
       !self->super.endpoints[EP_SERVER] ||
       !self->poll)
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
+    z_proxy_return(self, FALSE);
 
-  tmpstream1 = self->super.endpoints[EP_CLIENT];
-  z_stream_ref(tmpstream1);
-  if (self->need_ssl[EP_CLIENT])
+  for (ep = EP_CLIENT; ep < EP_MAX; ep++)
     {
-      z_proxy_log(self, PSSL_DEBUG, 6, "Client needs ssl.");
-      if (!pssl_init_client_ssl(self))
+      tmpstream = self->super.endpoints[ep];
+      z_stream_ref(tmpstream);
+      if (self->need_ssl[ep])
         {
-          z_stream_unref(tmpstream1);
-          z_proxy_leave(self);
-          return FALSE;
+          z_proxy_log(self, PSSL_DEBUG, 6, "Performing SSL handshake; side='%s'", EP_STR(ep));
+          if (!pssl_perform_handshake(self, ep))
+            {
+              z_stream_unref(tmpstream);
+              z_proxy_return(self, FALSE);
+            }
         }
+      z_stream_set_nonblock(tmpstream, TRUE);
+      z_stream_unref(tmpstream);
     }
-
-  tmpstream2 = self->super.endpoints[EP_SERVER];
-  z_stream_ref(tmpstream2);
-  if (self->need_ssl[EP_SERVER])
-    {
-      z_proxy_log(self, PSSL_DEBUG, 6, "Server needs ssl.");
-      if (!pssl_init_server_ssl(self))
-        {
-          z_stream_unref(tmpstream1);
-          z_stream_unref(tmpstream2);
-          z_proxy_leave(self);
-          return FALSE;
-        }
-    }
-
-  z_stream_set_nonblock(tmpstream1, TRUE);
-  z_stream_unref(tmpstream1);
-
-  z_stream_set_nonblock(tmpstream2, TRUE);
-  z_stream_unref(tmpstream2);
-
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 static gboolean
@@ -494,49 +432,15 @@ pssl_request_stack_event(PsslProxy *self, ZStackedProxy **stacked)
   if (res)
     {
       if (res != z_policy_none)
-        *stacked = z_proxy_stack_object(&self->super, res);
+        rc = z_proxy_stack_object(&self->super, res, stacked);
     }
-  else
-    if (called)
+  else if (called)
+    {
       rc = FALSE;
+    }
   z_policy_var_unref(res);
   z_policy_unlock(self->super.thread);
-  z_proxy_leave(self);
-  return rc;
-}
-
-static gboolean
-pssl_start_main_session(PsslProxy *self)
-{
-  ZStackedProxy *stacked;
-  ZPlugSession *session;
-  
-  if (!pssl_request_stack_event(self, &stacked))
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
-  session = z_plug_session_new(&self->super, &self->session_data, self->super.endpoints[EP_CLIENT], self->super.endpoints[EP_SERVER], stacked);
-  if (!session)
-    {
-      z_stacked_proxy_destroy(stacked);
-      z_proxy_leave(self);
-      return FALSE;
-    }
-
-  z_stream_unref(self->super.endpoints[EP_CLIENT]);
-  z_stream_unref(self->super.endpoints[EP_SERVER]);
-  self->super.endpoints[EP_CLIENT] = self->super.endpoints[EP_SERVER] = NULL;
-
-  if (!z_plug_session_start(session, self->poll))
-    {
-      z_plug_session_free(session);
-      z_proxy_leave(self);
-      return FALSE;
-    }
-  z_proxy_leave(self);
-  return TRUE;
-
+  z_proxy_return(self, rc);
 }
 
 static gboolean
@@ -546,51 +450,53 @@ pssl_config(ZProxy *s)
   gboolean success;
 
   z_proxy_enter(self);
-
   pssl_config_set_defaults(self);
   pssl_register_vars(self);
   success = Z_SUPER(s, ZProxy)->config(s);
-
-  z_proxy_leave(self);
-  return success;
+  z_proxy_return(self, success);
 }
 
 static void
 pssl_main(ZProxy *s)
 {
   PsslProxy *self = Z_CAST(s, PsslProxy);
+  ZStackedProxy *stacked;
+  ZPlugSession *session = NULL;
 
   z_proxy_enter(self);
-
   /* this sets the server side endpoint if successful */
-  if (!z_proxy_connect_server(&self->super, NULL, 0))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+  if (!z_proxy_connect_server(&self->super, NULL, 0) ||
+      !pssl_init_streams(self))
+    goto error;
   
-  if (!pssl_init_streams(self))
+  if (!pssl_request_stack_event(self, &stacked))
+    goto error;
+    
+  session = z_plug_session_new(&self->session_data, self->super.endpoints[EP_CLIENT], self->super.endpoints[EP_SERVER], stacked, (gpointer)&self->super);
+  if (!session)
     {
-      z_proxy_leave(self);
-      return;
+      z_stacked_proxy_destroy(stacked);
+      goto error;
     }
-  
-  if (!pssl_start_main_session(self))
+
+  if (!z_plug_session_start(session, self->poll))
     {
-      z_proxy_leave(self);
-      return;
+      goto error;
     }
     
-  while (z_plug_sessions_running(&self->session_data) &&
-	 z_poll_is_running(self->poll))
+  while (z_poll_is_running(self->poll) && z_poll_iter_timeout(self->poll, -1))
     {
-      /* NOTE: timeouts are handled by ZPlugSession */
-      z_poll_iter_timeout(self->poll, -1);
-      z_plug_sessions_purge(&self->session_data);
-
+      if (!z_proxy_loop_iteration(s))
+        {
+          break;
+        }
+    }
+ error:
+  if (session)
+    {
+      z_plug_session_cancel(session);
+      z_plug_session_unref(session);
     }
-  z_plug_sessions_purge(&self->session_data);   
-
   z_proxy_leave(self);
 }
 
@@ -601,9 +507,7 @@ pssl_proxy_new(ZProxyParams *params)
 
   z_enter();
   self = Z_CAST(z_proxy_new(Z_CLASS(PsslProxy), params), PsslProxy);
-  z_proxy_start(&self->super);
-  z_proxy_leave(self);
-  return &self->super;
+  z_return(&self->super);
 }
 
 static void
@@ -611,16 +515,15 @@ pssl_proxy_free(ZObject *s)
 {
   PsslProxy *self = Z_CAST(s, PsslProxy);
 
-  z_proxy_enter(self);
-
+  z_enter();
   if (self->poll)
     {
       z_poll_unref(self->poll);
       self->poll = NULL;
     }
-
   ERR_remove_state(0);
-  z_proxy_leave(self);
+  z_proxy_free_method(s);
+  z_return();
 }
 
 ZProxyFuncs pssl_proxy_funcs =
@@ -629,11 +532,8 @@ ZProxyFuncs pssl_proxy_funcs =
     Z_FUNCS_COUNT(ZProxy),
     pssl_proxy_free,
   },
-  pssl_config,
-  NULL,
-  pssl_main,
-  NULL,
-  NULL,
+  .config = pssl_config,
+  .main = pssl_main,
   NULL
 };
 
