@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -410,6 +411,8 @@
 /* macro to convert endpoint number to string */
 #define WHICH_EP(x)             (((x) == EP_CLIENT) ? "client" : "server")
 
+#define TELNET_AUDIT_FORMAT_VERSION        "0.0"
+
 struct _TelnetProxy;
 
 typedef guint (*TelnetOptionFunction)(struct _TelnetProxy *, guint);
@@ -439,7 +442,7 @@ typedef struct _ZIOBufferDyn
 
 #define TELNET_ERROR        "telnet.error"
 #define TELNET_POLICY       "telnet.policy"
-#define TELNET_VIOLATIONS   "telnet.violations"
+#define TELNET_VIOLATION    "telnet.violation"
 
 typedef struct _TelnetProxy
 {
@@ -461,7 +464,6 @@ typedef struct _TelnetProxy
     /* Private variables */
     gint                    state[EP_MAX];
     gint                    ep;
-    gint                    pri[EP_MAX];
 
     /* input buffers */
     ZIOBuffer               read_buffers[EP_MAX];
@@ -477,15 +479,21 @@ typedef struct _TelnetProxy
 
     /* WILL_OPTION, WONT_OPTION, DO_OPTION or DONT_OPTION
      * shows what was requested in TELNET_GOT_OPNEG state
+     *
+     * NOTE: the code assumes that these are octet sized, please be sure to
+     * audit the code if the type of these values is ever changed.
      */
     guchar                  command[EP_MAX];
     guchar                  opneg_option[EP_MAX];
 
     /* suboption check function lookup table */
     TelnetOptionFunction    telnet_options[256];
+    
 
     ZPoll                   *poll;
 } TelnetProxy;
 
+gboolean telnet_collect_meta(TelnetProxy *self, const gchar *name, const gchar *value, gboolean commit) G_GNUC_WARN_UNUSED_RESULT;
+
 extern ZClass TelnetProxy__class;
 #endif
