@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -69,7 +70,6 @@ telnet_opt_terminal_type(TelnetProxy *self, guint ep)
   guint         res;
 
   z_proxy_enter(self);
-
   ptr = sbuf->ofs;
   if (sbuf->buf[ptr] == TELNET_SB_TERMINAL_TYPE_IS)
     {
@@ -77,8 +77,7 @@ telnet_opt_terminal_type(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][ep] & SENT_WILL))
         {
           z_proxy_log(self, TELNET_POLICY, 3, "TERMINAL TYPE IS option not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       /* check if valid */
@@ -87,8 +86,7 @@ telnet_opt_terminal_type(TelnetProxy *self, guint ep)
             {
               /* FIXME: the value should be logged */
               z_proxy_log(self, TELNET_VIOLATION, 3, "Invalid TERMINAL TYPE value, it contains invalid characters;");
-              z_proxy_leave(self);
-              return TELNET_CHECK_ABORT;
+              z_proxy_return(self, TELNET_CHECK_ABORT);
             }
 
       /* copy to buf */
@@ -98,12 +96,10 @@ telnet_opt_terminal_type(TelnetProxy *self, guint ep)
       if (i >= sizeof(buf))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "Invalid TERMINAL TYPE value, it is too long;");
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       buf[i] = 0;
-
       z_proxy_log(self, TELNET_DEBUG, 6, "TERMINAL TYPE option; value='%s'", buf);
       g_string_assign(self->policy_name, TELNET_POLICY_TERMTYPE_NAME);
       g_string_assign(self->policy_value, buf);
@@ -121,24 +117,19 @@ telnet_opt_terminal_type(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][OTHER_EP(ep)] & GOT_DO))
         {
           z_proxy_log(self, TELNET_POLICY, 3, "TERMINAL TYPE SEND option not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       g_string_assign(self->policy_name, TELNET_POLICY_TERMTYPE_NAME);
       g_string_assign(self->policy_value, "");
-
       res = telnet_policy_suboption(self, 1, TELNET_POLICY_TERMTYPE_NAME, "");
     }
   else /* suboption code is  INVALID */
     {
       z_proxy_log(self, TELNET_VIOLATION, 3, "TERMINAL TYPE option, invalid subcommand or invalid suboption length;");
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
+      z_proxy_return(self, TELNET_CHECK_ABORT);
     }
-
-  z_proxy_leave(self);
-  return res;
+  z_proxy_return(self, res);
 }
 
 /**
@@ -160,7 +151,6 @@ telnet_opt_terminal_speed(TelnetProxy *self, guint ep)
   guint         res;
 
   z_proxy_enter(self);
-
   ptr = sbuf->ofs;
   if (sbuf->buf[ptr] == TELNET_SB_TERMINAL_SPEED_IS)
     {
@@ -168,8 +158,7 @@ telnet_opt_terminal_speed(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][ep] & SENT_WILL))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "TERMINAL SPEED IS option not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       for (ptr++; ptr < sbuf->end; ptr++)
@@ -177,8 +166,7 @@ telnet_opt_terminal_speed(TelnetProxy *self, guint ep)
           if (!isdigit(sbuf->buf[ptr]) && sbuf->buf[ptr] != ',')
             {
               z_proxy_log(self, TELNET_VIOLATION, 3, "TERMINAL SPEED IS option, invalid speed string;");
-              z_proxy_leave(self);
-              return TELNET_CHECK_ABORT;
+              z_proxy_return(self, TELNET_CHECK_ABORT);
             }
         }
 
@@ -188,16 +176,13 @@ telnet_opt_terminal_speed(TelnetProxy *self, guint ep)
       if (i >= sizeof(buf))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "TERMINAL SPEED IS option, value too long");
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       buf[i] = 0;
       z_proxy_log(self, TELNET_DEBUG, 6, "TERMINAL SPEED IS option; value='%s'", buf);
-
       g_string_assign(self->policy_name, TELNET_POLICY_TERMSPEED_NAME);
       g_string_assign(self->policy_value, buf);
-
       res = telnet_policy_suboption(self, 0, TELNET_POLICY_TERMSPEED_NAME, buf);
       if (res == TELNET_CHECK_OK)
         {
@@ -212,10 +197,8 @@ telnet_opt_terminal_speed(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][OTHER_EP(ep)] & GOT_DO))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "TERMINAL SPEED SEND option not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
-
       g_string_assign(self->policy_name, TELNET_POLICY_TERMSPEED_NAME);
       g_string_assign(self->policy_value, "");
       res = telnet_policy_suboption(self, 1, TELNET_POLICY_TERMSPEED_NAME, "");
@@ -223,12 +206,9 @@ telnet_opt_terminal_speed(TelnetProxy *self, guint ep)
   else /* suboption code is INVALID */
     {
       z_proxy_log(self, TELNET_VIOLATION, 3, "TERMINAL SPEED option, invalid subcommand;");
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
+      z_proxy_return(self, TELNET_CHECK_ABORT);
     }
-
-  z_proxy_leave(self);
-  return res;
+  z_proxy_return(self, res);
 }
 
 /**
@@ -250,7 +230,6 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
   guint         res;
 
   z_proxy_enter(self);
-
   ptr = sbuf->ofs;
   if (sbuf->buf[ptr] == TELNET_SB_X_DISPLAY_LOCATION_IS)
     {
@@ -258,8 +237,7 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][ep] & SENT_WILL))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "X DISPLAY LOCATION IS option not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       for (ptr++; ptr < sbuf->end; ptr++)
@@ -267,8 +245,7 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
           if (!isalnum(sbuf->buf[ptr]) && sbuf->buf[ptr] != '.' && sbuf->buf[ptr] != ':')
             {
               z_proxy_log(self, TELNET_VIOLATION, 3, "X DISPLAY LOCATION IS option, invalid speed string;");
-              z_proxy_leave(self);
-              return TELNET_CHECK_ABORT;
+              z_proxy_return(self, TELNET_CHECK_ABORT);
             }
         }
 
@@ -278,8 +255,7 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
       if (i >= sizeof(buf))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "X DISPLAY LOCATION IS option, value too long;");
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       buf[i] = 0;
@@ -287,7 +263,6 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
 
       g_string_assign(self->policy_name, TELNET_POLICY_XDISPLAY_NAME);
       g_string_assign(self->policy_value, buf);
-
       res = telnet_policy_suboption(self, 0, TELNET_POLICY_XDISPLAY_NAME, buf);
       if (res == TELNET_CHECK_OK)
         {
@@ -302,8 +277,7 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][OTHER_EP(ep)] & GOT_DO))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "X DISPLAY LOCATION SEND option is not allowed from this side;");
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       g_string_assign(self->policy_name, TELNET_POLICY_XDISPLAY_NAME);
@@ -313,12 +287,9 @@ telnet_opt_x_display(TelnetProxy *self, guint ep)
   else /* suboption code is INVALID */
     {
       z_proxy_log(self, TELNET_VIOLATION, 3, "X DISPLAY LOCATION option, invalid subcommand or invalid suboption length;");
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
+      z_proxy_return(self, TELNET_CHECK_ABORT);
     }
-
-  z_proxy_leave(self);
-  return res;
+  z_proxy_return(self, res);
 }
 
 /**
@@ -343,23 +314,19 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
   gboolean      valid = FALSE; /* TRUE if there was anything accepted by policy */
 
   z_proxy_enter(self);
-
   ptr = sbuf->ofs++;
   command = sbuf->buf[ptr++];
-
   /* initialize cbuf */
   cbuf.ofs = 0;
   cbuf.end = 1;
   cbuf.buf[0] = command;
-
   if (command == TELNET_SB_ENVIRONMENT_IS || command == TELNET_SB_ENVIRONMENT_INFO)
     {
       /* check if this side sent WILL */
       if (!(self->options[self->opneg_option[ep]][ep] & SENT_WILL))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "NEW ENVIRON IS or INFO option not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       if (ptr == sbuf->end)
@@ -367,11 +334,9 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
           /* if this was an empty IS or INFO reply */
           g_string_assign(self->policy_name, "");
           g_string_assign(self->policy_value, "");
-
           res = telnet_policy_suboption(self, command, "", "");
-
           if (res == TELNET_CHECK_OK) 
-              valid = TRUE;
+            valid = TRUE;
         }
       else while (ptr < sbuf->end)
         {
@@ -386,18 +351,18 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
                   if (sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_VAR || 
                       sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_VALUE || 
                       sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_USERVAR)
-                      break;  /* got VAR, VALUE, or USERVAR, here ends variable name */
+                    break;  /* got VAR, VALUE, or USERVAR, here ends variable name */
 
                   if (sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_ESC || 
                       sbuf->buf[ptr] == TELNET_IAC)
-                      ptr++;  /* got ESC or IAC, these are followed by the real byte */
+                    ptr++;  /* got ESC or IAC, these are followed by the real byte */
 
                   if (i < sizeof(name))
-                      name[i] = sbuf->buf[ptr];
+                    name[i] = sbuf->buf[ptr];
                 }
               /* terminate name */
               if (i < sizeof(name))
-                  name[i] = '\0';
+                name[i] = '\0';
               else
                 {
                   z_proxy_log(self, TELNET_VIOLATION, 3, "NEW-ENVIRON option, variable name too long;");
@@ -412,17 +377,20 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
                       if (sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_VAR || 
                           sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_VALUE || 
                           sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_USERVAR)
-                          break;  /* got VAR, VALUE or USERVAR, here ends value */
+                        break;  /* got VAR, VALUE or USERVAR, here ends value */
 
                       if (sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_ESC || 
                           sbuf->buf[ptr] == TELNET_IAC) 
-                          ptr++;  /* got ESC or IAC, these are followed by the real byte */
+                        ptr++;  /* got ESC or IAC, these are followed by the real byte */
 
-                      if (ptr < sbuf->end) value[i] = sbuf->buf[ptr];
+                      if (ptr < sbuf->end)
+                        value[i] = sbuf->buf[ptr];
                     }
                   /* terminate value */
                   if (i < sizeof(value)) 
+                    {
                       value[i] = '\0';
+                    }
                   else
                     {
                       z_proxy_log(self, TELNET_VIOLATION, 3, "NEW-ENVIRON option, variable value too long;");
@@ -435,7 +403,6 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
 
               g_string_assign(self->policy_name, name);
               g_string_assign(self->policy_value, value);
-
               res = telnet_policy_suboption(self, command, name, value);
               if (res == TELNET_CHECK_OK)
                 {
@@ -443,15 +410,19 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
                   /* we may forward variable, so copy it */
                   cbuf.buf[cbuf.end++] = type;
                   if (cbuf.end < sizeof(cbuf.buf))
+                    {
                       for (i = 0; cbuf.end < sizeof(cbuf.buf) && i < self->policy_name->len; i++, cbuf.end++)
-                          cbuf.buf[cbuf.end] = self->policy_name->str[i];
+                        cbuf.buf[cbuf.end] = self->policy_name->str[i];
+                    }
                   
                   if (cbuf.end < sizeof(cbuf.buf))
-                      cbuf.buf[cbuf.end++] = 1;
+                    cbuf.buf[cbuf.end++] = 1;
                   
                   if (cbuf.end < sizeof(cbuf.buf))
+                    {
                       for (i = 0; cbuf.end < sizeof(cbuf.buf) && i < self->policy_value->len; i++, cbuf.end++)
-                          cbuf.buf[cbuf.end] = self->policy_value->str[i];
+                        cbuf.buf[cbuf.end] = self->policy_value->str[i];
+                    }
                   
                   if (cbuf.end >= sizeof(cbuf.buf))
                     {
@@ -478,8 +449,7 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
       if (!(self->options[self->opneg_option[ep]][OTHER_EP(ep)] & GOT_DO))
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "NEW-ENVIRON SEND otpion not allowed from this side; side='%s'", WHICH_EP(ep));
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
 
       if (ptr == sbuf->end)
@@ -487,9 +457,9 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
           /* if this was an empty SEND request */
           g_string_assign(self->policy_name, "");
           g_string_assign(self->policy_value, "");
-
           res = telnet_policy_suboption(self, command, "", "");
-          if (res == TELNET_CHECK_OK) valid = TRUE;
+          if (res == TELNET_CHECK_OK)
+            valid = TRUE;
         }
       else while (ptr < sbuf->end)
         {
@@ -503,18 +473,20 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
                 {
                   if (sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_VAR ||
                       sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_USERVAR)
-                      break;  /* got VAR or USERVAR, here ends variable name */
+                    break;  /* got VAR or USERVAR, here ends variable name */
 
                   if (sbuf->buf[ptr] == TELNET_OPTARG_ENVIRONMENT_ESC ||
                       sbuf->buf[ptr] == TELNET_IAC)
-                      ptr++;  /* got ESC or IAC, these are followed by the real byte */
+                    ptr++;  /* got ESC or IAC, these are followed by the real byte */
 
                   if (i < sizeof(name))
-                      name[i] = sbuf->buf[ptr];
+                    name[i] = sbuf->buf[ptr];
                 }
               /* terminate name */
               if (i < sizeof(name)) 
+                {
                   name[i] = '\0';
+                }
               else
                 {
                   z_proxy_log(self, TELNET_VIOLATION, 3, "NEW-ENVIRON SEND option, variable name too long");
@@ -531,10 +503,12 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
                   valid = TRUE;
                   /* we may forward option, so copy it */
                   for (; sbuf->ofs < ptr; sbuf->ofs++, cbuf.end++)
-                      cbuf.buf[cbuf.end] = sbuf->buf[sbuf->ofs];
+                    cbuf.buf[cbuf.end] = sbuf->buf[sbuf->ofs];
                 }
               else 
+                {
                   sbuf->ofs = ptr;
+                }
             break;
 
             default: /* not VAR or USERVAR, invalid */
@@ -545,24 +519,21 @@ telnet_opt_new_env(TelnetProxy *self, guint ep)
         } /* while */
     }
   else /* suboption code is INVALID */
+    {
       z_proxy_log(self, TELNET_VIOLATION, 3, "NEW-ENVIRON option, invalid subcommand;");
+    }
 
   /* if there wasn't any accepted variable, return DENY */
   if (!valid) 
-    {
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
-    }
+    z_proxy_return(self, TELNET_CHECK_ABORT);
 
   /* else copy accepted bytes to suboption buffer */
   for (i = 0; i < cbuf.end; i++)
-      sbuf->buf[i] = cbuf.buf[i];
+    sbuf->buf[i] = cbuf.buf[i];
 
   sbuf->ofs = 0;
   sbuf->end = cbuf.end;
-
-  z_proxy_leave(self);
-  return TELNET_CHECK_OK;
+  z_proxy_return(self, TELNET_CHECK_OK);
 }
 
 /**
@@ -584,15 +555,14 @@ telnet_opt_naws(TelnetProxy *self, guint ep)
   guint         res;
   guchar        value[SB_BUF_SIZE];
   guint16       width, height;
+  gchar         width_cols[256], height_rows[256];
 
   z_proxy_enter(self);
-
   /* check if this side sent WILL */
   if (!(self->options[self->opneg_option[ep]][ep] & SENT_WILL))
     {
       z_proxy_log(self, TELNET_DEBUG, 5, "NAWS option not allowed from this side; side='%s'", WHICH_EP(ep));
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
+      z_proxy_return(self, TELNET_CHECK_ABORT);
     }
 
   if (sbuf->end - sbuf->ofs != 4)
@@ -601,29 +571,31 @@ telnet_opt_naws(TelnetProxy *self, guint ep)
         {
           buf[i] = sbuf->buf[ptr];
           if (sbuf->buf[ptr] == TELNET_IAC) 
-              ptr++;
+            ptr++;
         }
       if (i != 4)
         {
           z_proxy_log(self, TELNET_VIOLATION, 3, "NAWS option, invalid length");
-          z_proxy_leave(self);
-          return TELNET_CHECK_ABORT;
+          z_proxy_return(self, TELNET_CHECK_ABORT);
         }
     }
   else 
     {
       for (ptr = sbuf->ofs, i = 0; i < 4; ptr++, i++)
-          buf[i] = sbuf->buf[ptr];
+        buf[i] = sbuf->buf[ptr];
     }
 
   width = (buf[0] << 8) + buf[1];
   height = (buf[2] << 8) + buf[3];
-
   g_string_assign(self->policy_name, TELNET_POLICY_NAWS_NAME);
   g_string_sprintf(self->policy_value, "%hd,%hd", width, height);
   snprintf(value, sizeof(value), "%hd,%hd", width, height);
-
   res = telnet_policy_suboption(self, 0, TELNET_POLICY_NAWS_NAME, value);
-  z_proxy_leave(self);
-  return res;
+  if (res == Z_ACCEPT)
+    {
+      g_snprintf(width_cols, sizeof(width_cols), "%hd", width);
+      g_snprintf(height_rows, sizeof(height_rows), "%hd", height);
+      
+    }
+  z_proxy_return(self, res);
 }
