@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -96,28 +97,23 @@ telnet_policy_option(TelnetProxy *self)
   gboolean      type_found;
 
   z_proxy_enter(self);
-
   z_proxy_log(self, TELNET_DEBUG, 8, "Policy option negotiation check; option='%d'", self->opneg_option[self->ep]);
   g_snprintf(lookup_str, sizeof(lookup_str), "%d", self->opneg_option[self->ep]);
   keys[0] = lookup_str;
   tmp = z_dim_hash_table_search(self->telnet_policy, 1, keys);
-
   if (!tmp)
     {
-      z_proxy_leave(self);
       z_proxy_log(self, TELNET_POLICY, 2, "Option not found in policy; option='%s'", lookup_str);
-      return TELNET_CHECK_DROP;
+      z_proxy_return(self, TELNET_CHECK_DROP);
     }
 
   z_policy_lock(self->super.thread);
   type_found = telnet_hash_get_type(tmp, &command_do);
   z_policy_unlock(self->super.thread);
-
   if (!type_found )
     {
       z_proxy_log(self, TELNET_POLICY, 2, "Policy type invalid; option='%s'", lookup_str);
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
+      z_proxy_return(self, TELNET_CHECK_ABORT);
     }
 
   switch (command_do)
@@ -125,13 +121,13 @@ telnet_policy_option(TelnetProxy *self)
     case TELNET_OPTION_DROP:
       z_proxy_log(self, TELNET_POLICY, 3, "Policy denied option; option='%s'", lookup_str);
       res = TELNET_CHECK_DROP;
-    break;
-    
+      break;
+
     case TELNET_OPTION_ACCEPT:
       z_proxy_log(self, TELNET_POLICY, 6, "Policy accepted option; option='%s'", lookup_str);
       res = TELNET_CHECK_OK;
-    break;
-    
+      break;
+
     case TELNET_OPTION_POLICY:
       z_policy_lock(self->super.thread);
       if (!z_policy_var_parse(tmp, "(iO)", &command_do, &command_where))
@@ -159,44 +155,42 @@ telnet_policy_option(TelnetProxy *self)
                 case Z_ACCEPT:
                   z_proxy_log(self, TELNET_POLICY, 6, "Policy function accepted option; option='%s'", lookup_str);
                   res = TELNET_CHECK_OK;
-                break;
-                
+                  break;
+
                 case Z_UNSPEC:
                 case Z_DROP:
                   z_proxy_log(self, TELNET_POLICY, 3, "Policy function drop option; option='%s'", lookup_str);
                   res = TELNET_CHECK_DROP;
-                break;
-                
+                  break;
+
                 case TELNET_OPTION_REJECT:
                   z_proxy_log(self, TELNET_POLICY, 3, "Policy function reject option; option='%s'", lookup_str);
                   res = TELNET_CHECK_REJECT;
-                break;
-                
+                  break;
+
                 case Z_ABORT:
                 default:
                   z_proxy_log(self, TELNET_POLICY, 1, "Policy function aborted session; option='%s'", lookup_str);
                   res = TELNET_CHECK_ABORT;
-                break;
+                  break;
                 }
             }
         }
       z_policy_unlock(self->super.thread);
-    break;
-    
+      break;
+
     case TELNET_OPTION_REJECT:
       z_proxy_log(self, TELNET_POLICY, 3, "Policy rejected option; option='%s'", lookup_str);
       res = TELNET_CHECK_REJECT;
-    break;
-    
+      break;
+
     case TELNET_OPTION_ABORT:
     default:
       z_proxy_log(self, TELNET_POLICY, 3, "Policy aborted session; option='%s'", lookup_str);
       res = TELNET_CHECK_ABORT;
-    break;
+      break;
     }
-
-  z_proxy_leave(self);
-  return res;
+  z_proxy_return(self, res);
 }
 
 /**
@@ -224,31 +218,25 @@ telnet_policy_suboption(TelnetProxy *self, guchar command, gchar *name, gchar *v
   gboolean      type_found;
 
   z_proxy_enter(self);
-
   z_proxy_log(self, TELNET_DEBUG, 8, "Policy suboption negotiation check;");
   g_snprintf(lookup_str[0], sizeof(lookup_str[0]), "%d", self->opneg_option[self->ep]);
   g_snprintf(lookup_str[1], sizeof(lookup_str[1]), "%d", command);
   keys[0] = lookup_str[0];
   keys[1] = lookup_str[1];
-
   tmp = z_dim_hash_table_search(self->telnet_policy, 2, keys);
-
   if (!tmp)
     {
-      z_proxy_leave(self);
       z_proxy_log(self, TELNET_POLICY, 1, "Option not found in policy hash, dropping; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
-      return TELNET_CHECK_DROP;
+      z_proxy_return(self, TELNET_CHECK_DROP);
     }
 
   z_policy_lock(self->super.thread);
   type_found = telnet_hash_get_type(tmp, &command_do);
   z_policy_unlock(self->super.thread);
-      
   if (!type_found)
     {
       z_proxy_log(self, TELNET_POLICY, 2, "Policy type invalid!");
-      z_proxy_leave(self);
-      return TELNET_CHECK_ABORT;
+      z_proxy_return(self, TELNET_CHECK_ABORT);
     }
 
   switch (command_do)
@@ -256,13 +244,13 @@ telnet_policy_suboption(TelnetProxy *self, guchar command, gchar *name, gchar *v
     case TELNET_OPTION_DROP:
       z_proxy_log(self, TELNET_POLICY, 6, "Policy denied suboption; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
       res = TELNET_CHECK_DROP;
-    break;
-    
+      break;
+
     case TELNET_OPTION_ACCEPT:
       z_proxy_log(self, TELNET_POLICY, 6, "Policy accepted suboption; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
       res = TELNET_CHECK_OK;
-    break;
-    
+      break;
+
     case TELNET_OPTION_POLICY:
       z_policy_lock(self->super.thread);
       if (!z_policy_var_parse(tmp, "(iO)", &command_do, &command_where))
@@ -281,11 +269,11 @@ telnet_policy_suboption(TelnetProxy *self, guchar command, gchar *name, gchar *v
             case TELNET_OPTION_ENVIRONMENT:
             case TELNET_OPTION_NAWS:
               pol_res = z_policy_call_object(command_where, z_policy_var_build("(iss)", &self->opneg_option[self->ep], name, value), self->super.session_id);
-            break;
+              break;
 
             default:
               pol_res = z_policy_call_object(command_where, z_policy_var_build("(i)", &self->opneg_option[self->ep]), self->super.session_id);
-            break;
+              break;
             }
 
           if (pol_res == NULL)
@@ -305,34 +293,32 @@ telnet_policy_suboption(TelnetProxy *self, guchar command, gchar *name, gchar *v
                 case Z_ACCEPT:
                   z_proxy_log(self, TELNET_POLICY, 6, "Policy function accepted suboption; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
                   res = TELNET_CHECK_OK;
-                break;
+                  break;
 
                 case Z_UNSPEC:
                 case Z_REJECT:
                 case Z_DROP:
                   z_proxy_log(self, TELNET_POLICY, 3, "Policy function denied suboption; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
                   res = TELNET_CHECK_DROP;
-                break;
+                  break;
 
                 case Z_ABORT:
                 default:
                   z_proxy_log(self, TELNET_POLICY, 3, "Policy function aborted suboption; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
                   res = TELNET_CHECK_ABORT;
-                break;
+                  break;
                 }
             }
         }
       z_policy_unlock(self->super.thread);
-    break;
-    
+      break;
+
     case TELNET_OPTION_ABORT:
     default:
       z_proxy_log(self, TELNET_POLICY, 3, "Policy aborted session; command=`%s', option=`%s'", lookup_str[1], lookup_str[0]);
       res = TELNET_CHECK_ABORT;
-    break;
+      break;
     }
-
-  z_proxy_leave(self);
-  return res;
+  z_proxy_return(self, res);
 }
 
