@@ -1,6 +1,7 @@
 /***************************************************************************
  *
- * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+ * Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+ * 2010 BalaBit IT Ltd, Budapest, Hungary
  *
  * This program is free software; you can redistribute it and/or modify it
  * under the terms of the GNU General Public License version 2 as published
@@ -37,6 +38,7 @@
 #include <zorp/registry.h>
 #include <zorp/log.h>
 #include <zorp/policy.h>
+#include <zorp/proxy.h>
 
 #include <ctype.h>
 
@@ -82,7 +84,7 @@ whois_config_set_defaults(WhoisProxy *self)
   self->response_header = g_string_sized_new(0);
   self->response_footer = g_string_sized_new(0);
   self->timeout = 30000;
-  z_proxy_leave(self);
+  z_proxy_return(self);
 }
 
 /*+
@@ -100,7 +102,7 @@ whois_register_vars(WhoisProxy *self)
   z_proxy_var_new(&self->super, "request", Z_VAR_GET | Z_VAR_SET | Z_VAR_TYPE_STRING, self->request);
   z_proxy_var_new(&self->super, "response_header", Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_STRING, self->response_header);
   z_proxy_var_new(&self->super, "response_footer", Z_VAR_GET | Z_VAR_SET | Z_VAR_SET_CONFIG | Z_VAR_TYPE_STRING, self->response_footer);
-  z_proxy_leave(self);
+  z_proxy_return(self);
 }
 
 /*+
@@ -124,15 +126,13 @@ static gboolean
 whois_init_client_stream(WhoisProxy *self)
 {
   ZStream *tmpstream;
-  z_proxy_enter(self);
   
+  z_proxy_enter(self);
   self->super.endpoints[EP_CLIENT]->timeout = self->timeout;
   tmpstream = self->super.endpoints[EP_CLIENT];
   self->super.endpoints[EP_CLIENT] = z_stream_line_new(tmpstream, self->max_line_length, ZRL_EOL_CRLF);
   z_stream_unref(tmpstream);
-  
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /*+
@@ -145,20 +145,15 @@ static gboolean
 whois_init_server_stream(WhoisProxy *self)
 {
   ZStream *tmpstream;
+  
   z_proxy_enter(self);
-  
   if (!self->super.endpoints[EP_SERVER])
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
-  
+    z_proxy_return(self, FALSE);
   self->super.endpoints[EP_SERVER]->timeout = self->timeout;
   tmpstream = self->super.endpoints[EP_SERVER];
   self->super.endpoints[EP_SERVER] = z_stream_line_new(tmpstream, self->max_line_length, ZRL_EOL_CRLF);
   z_stream_unref(tmpstream);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /*+ 
@@ -181,23 +176,19 @@ whois_fetch_request(WhoisProxy *self)
   if (res != G_IO_STATUS_NORMAL)
     {
       z_proxy_log(self, WHOIS_ERROR, 1, "Empty request received or I/O error;");
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
     
   z_proxy_log(self, WHOIS_REQUEST, 7, "Incoming request; line='%.*s'", (gint) line_length, line);
-  
   if (line_length > (guint) self->max_request_length)
     {
       z_proxy_log(self, WHOIS_REQUEST, 6, "Request too long; length='%zd', max_request_length='%d'", line_length, self->max_request_length);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
     }
   
   g_string_truncate(self->request, 0);
   g_string_append_len(self->request, line, line_length);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 /*+
@@ -214,18 +205,14 @@ whois_send_request(WhoisProxy *self)
   
   z_proxy_enter(self);
   g_snprintf(request, sizeof(request), "%s\r\n", self->request->str);
-    
   if (z_stream_write(self->super.endpoints[EP_SERVER],
                      request,
                      strlen(request),
                      &bytes_written,
                      NULL) != G_IO_STATUS_NORMAL)
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
-  z_proxy_leave(self);
-  return TRUE;
+    z_proxy_return(self, FALSE);
+
+  z_proxy_return(self, TRUE);
 }
 
 /*+
@@ -245,10 +232,7 @@ whois_copy_response(WhoisProxy *self)
                      self->response_header->len,
                      &bytes_written,
                      NULL) != G_IO_STATUS_NORMAL)
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
+    z_proxy_return(self, FALSE);
 
   while (1)
     {
@@ -257,6 +241,9 @@ whois_copy_response(WhoisProxy *self)
       gint res;
       gchar *response;
       
+      if (!z_proxy_loop_iteration(&self->super))
+        break;
+      
       res = z_stream_line_get(self->super.endpoints[EP_SERVER], &line, &line_len, NULL);
       if (res != G_IO_STATUS_NORMAL)
         /* EOF or read error */
@@ -265,17 +252,12 @@ whois_copy_response(WhoisProxy *self)
       response = alloca(line_len + 3);
       memcpy(response, line, line_len);
       strcpy(response + line_len, "\r\n");
-
       if (z_stream_write(self->super.endpoints[EP_CLIENT],
                          response,
                          line_len + 2,
                          &bytes_written,
                          NULL) != G_IO_STATUS_NORMAL)
-        {
-          z_proxy_leave(self);
-          return FALSE;
-        }
-
+        z_proxy_return(self, FALSE);
     }
   if (self->response_footer->len &&
       z_stream_write(self->super.endpoints[EP_CLIENT],
@@ -283,12 +265,9 @@ whois_copy_response(WhoisProxy *self)
                      self->response_footer->len,
                      &bytes_written,
                      NULL) != G_IO_STATUS_NORMAL)
-    {
-      z_proxy_leave(self);
-      return FALSE;
-    }
-  z_proxy_leave(self);
-  return TRUE;
+    z_proxy_return(self, FALSE);
+
+  z_proxy_return(self, TRUE);
 }
 
 static gboolean
@@ -305,7 +284,6 @@ whois_query_policy(WhoisProxy *self)
     {
     case Z_REJECT:
     case Z_ABORT:
-      
       z_stream_write(self->super.endpoints[EP_CLIENT],
                      errmsg,
                      strlen(errmsg),
@@ -315,16 +293,15 @@ whois_query_policy(WhoisProxy *self)
       
     case Z_DROP:
       z_policy_unlock(self->super.thread);
-      z_proxy_leave(self);
-      return FALSE;
+      z_proxy_return(self, FALSE);
+
     case Z_UNSPEC:
     case Z_ACCEPT:
     default:
       break;
     }
   z_policy_unlock(self->super.thread);
-  z_proxy_leave(self);
-  return TRUE;
+  z_proxy_return(self, TRUE);
 }
 
 static gboolean
@@ -336,26 +313,22 @@ whois_config(ZProxy *s)
   z_proxy_enter(self);
   whois_config_set_defaults(self);  
   whois_register_vars(self);
-  
   if (Z_SUPER(s, ZProxy)->config(s))
     {
       whois_config_init(self);
       success = TRUE;
     }
-  z_proxy_leave(self);
-  return success;
+  z_proxy_return(self, success);
 }
 
 static void
 whois_main(ZProxy *s)
 {
   WhoisProxy *self = Z_CAST(s, WhoisProxy);
+
   z_proxy_enter(self);  
   if (!whois_init_client_stream(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
 
   z_proxy_log(self, WHOIS_DEBUG, 6, "fetching request;");
   if (!whois_fetch_request(self))
@@ -368,49 +341,31 @@ whois_main(ZProxy *s)
                      strlen(errmsg),
                      &bytes_written,
                      NULL);
-      z_proxy_leave(self);
-      return;
+      z_proxy_return(self);
     }
   
   z_proxy_log(self, WHOIS_DEBUG, 6, "checking policy;");
-  
   if (!whois_query_policy(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
   
   z_proxy_log(self, WHOIS_DEBUG, 6, "connecting to server;");
-  
   /* this sets the server side endpoint if successful */
   if (!z_proxy_connect_server(&self->super, NULL, 0))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
 
   if (!whois_init_server_stream(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
     
   z_proxy_log(self, WHOIS_DEBUG, 6, "sending request;");
   if (!whois_send_request(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
     
   z_proxy_log(self, WHOIS_DEBUG, 6, "copying response;");
   if (!whois_copy_response(self))
-    {
-      z_proxy_leave(self);
-      return;
-    }
+    z_proxy_return(self);
 
   z_proxy_log(self, WHOIS_DEBUG, 6, "processing done;");
-  z_proxy_leave(self);
+  z_proxy_return(self);
 }
 
 
@@ -427,10 +382,7 @@ whois_proxy_new(ZProxyParams *params)
   
   z_enter();
   self = Z_CAST(z_proxy_new(Z_CLASS(WhoisProxy), params), WhoisProxy);
-
-  z_proxy_start(&self->super);
-  z_leave();
-  return (ZProxy *) self;
+  z_return((ZProxy *) self);
 }
 
 static void
@@ -438,9 +390,9 @@ whois_proxy_free(ZObject *s)
 {
   WhoisProxy *self G_GNUC_UNUSED = Z_CAST(s, WhoisProxy);
   
-  z_proxy_enter(self);
+  z_enter();
   z_proxy_free_method(s);
-  z_proxy_leave(self);
+  z_return();
 }
 
 ZProxyFuncs whois_proxy_funcs =
@@ -450,11 +402,8 @@ ZProxyFuncs whois_proxy_funcs =
     whois_proxy_free,
   },
   .config = whois_config,
-  .startup = NULL,
   .main = whois_main,
-  .shutdown = NULL,
-  .destroy = NULL,
-  .run = NULL
+  NULL
 };
 
 ZClass WhoisProxy__class = 
