@@ -1,6 +1,7 @@
 ############################################################################
 ##
-## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+## 2010 BalaBit IT Ltd, Budapest, Hungary
 ##
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -17,33 +18,40 @@
 ## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ##
 ##
-## $Id: Core.py,v 1.26 2004/07/07 15:28:51 bazsi Exp $
-##
-## Author  : Bazsi
-## Auditor :
-## Last audited version:
-## Notes:
-##
 ############################################################################
-
+"""
+<module maturity="stable">
+  <summary>Module defining interface to the Core.</summary>
+  <description>
+    <para>
+      This module imports all public Zorp interfaces and makes it easy to use
+      those from the user policy file by simply importing all symbols from
+      Zorp.Core.
+    </para>
+  </description>
+</module>
+"""
+  
 import new
 
 from Zorp import *
 import Zorp
 from Zone import InetZone
-from Service import Service
+from Service import Service, PFService
 from SockAddr import SockAddrInet, SockAddrInetRange, SockAddrUnix
 
 
 from Router import TransparentRouter, DirectedRouter, InbandRouter
-from Chainer import ConnectChainer, FailoverChainer, SideStackChainer
+from Chainer import ConnectChainer, MultiTargetChainer, StateBasedChainer, RoundRobinChainer, FailoverChainer, SideStackChainer
 from Domain import InetDomain
 from Listener import Listener, ZoneListener, CSZoneListener
+from Dispatch import Dispatcher, ZoneDispatcher, CSZoneDispatcher
 from NAT import NATPolicy, ForgeClientSourceNAT, StaticNAT, OneToOneNAT, OneToOneMultiNAT, RandomNAT, HashNAT, GeneralNAT
 from Proxy import proxyLog
-from Auth import InbandAuthentication, AuthCache, AuthPolicy
-from Matcher import AbstractMatcher, RegexpMatcher, RegexpFileMatcher
-from Resolver import DNSResolver
+from Auth import InbandAuthentication, AuthCache, AuthPolicy, AuthenticationPolicy
+from Stack import StackingProvider, RemoteStackingBackend
+from Matcher import MatcherPolicy, AbstractMatcher, RegexpMatcher, RegexpFileMatcher, CombineMatcher, DNSMatcher, WindowsUpdateMatcher, SmtpInvalidRecipientMatcher
+from Resolver import DNSResolver, HashResolver, ResolverPolicy
 
 # conntrack support
 try:
@@ -58,13 +66,3 @@ try:
 except:
 	pass
 
-config = new.module('config')
-
-addConfigContainer(config, 'blob')
-config.blob.default_tmpdir = '/var/lib/zorp/tmp/'
-config.blob.default_max_disk_usage = 1024*0x100000
-config.blob.default_max_mem_usage = 256*0x100000
-config.blob.default_lowat = 96*0x100000
-config.blob.default_hiwat = 128*0x100000
-config.blob.default_noswap_max = 16384
-
