@@ -1,6 +1,7 @@
 ############################################################################
 ##
-## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+## 2010 BalaBit IT Ltd, Budapest, Hungary
 ##
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -17,20 +18,35 @@
 ## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ##
 ##
-## $Id: Globals.py,v 1.6 2003/12/15 13:43:10 bazsi Exp $
-##
-## Author  : Bazsi
-## Auditor : 
-## Last audited version:
-## Notes:
-##
 ############################################################################
 
+"""
+<module maturity="stable">
+  <summary>Module defining interface to the Global.</summary>
+  <description>
+    <para>
+      Global variables used by Zorp policy layer.
+    </para>
+  </description>
+</module>
+"""
+import threading
+
 services = {}
-authentication_pv = {}
+authentication_providers = {}
+authentication_policies = {}
+authorization_policies = {}
+neyes_waiting_authorization_policies_lock = threading.Lock()
+neyes_waiting_authorization_policies = {}
+resolver_policies = {}
+auth_policies = {}
+stacking_providers = {}
 zones = {}
 nat_policies = {}
-auth_policies = {}
 dispatches = []
 auth_caches = {}
+matchers = {}
+notification_policy = None
+instance_name = ''
 
+deinit_callbacks = []
