@@ -1,6 +1,7 @@
 ############################################################################
 ##
-## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005 BalaBit IT Ltd, Budapest, Hungary
+## Copyright (c) 2000, 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009,
+## 2010 BalaBit IT Ltd, Budapest, Hungary
 ##
 ## This program is free software; you can redistribute it and/or modify
 ## it under the terms of the GNU General Public License as published by
@@ -17,11 +18,9 @@
 ## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 ##
 ##
-## $Id: __init__.py,v 1.7 2002/04/19 12:41:03 sasa Exp $
-##
-## Author  : Bazsi
-## Auditor : kisza
-## Last audited version: 1.2
-## Notes:
-##
 ############################################################################
+"""
+<module internal="yes">
+  <summary>Module defining interface to the constructors</summary>
+</module>
+"""
