@@ -6,6 +6,8 @@ from Zorp.Session import MasterSession
 from traceback import *
 from time import time
 
+config.options.kzorp_enabled = FALSE
+
 def test(str, res, expect):
 	if res != expect:
 		print str, 'failed,', res, 'should be: ', expect
@@ -28,6 +30,7 @@ def init(name):
 		t11 = InetZone("test11", "192.168.1.4/32", admin_parent="test9")
 		t12 = InetZone("test12", "192.168.1.5/32", inbound_services=['*'])
 		t13 = InetZone("test13", "192.168.1.6/32", outbound_services=['*'])
+		t14 = InetZone("test14", "192.168.0.184", outbound_services=['*'])
 		
 		test('192.168.0.1', root_zone.findZone(SockAddrInet('192.168.0.1', 10)), t3)
 		test('192.168.0.33', root_zone.findZone(SockAddrInet('192.168.0.33', 10)), t2)
@@ -35,6 +38,7 @@ def init(name):
 		test('192.168.0.97', root_zone.findZone(SockAddrInet('192.168.0.97', 10)), t5)
 		test('192.168.0.129', root_zone.findZone(SockAddrInet('192.168.0.129', 10)), t1)
 		test('192.168.1.129', root_zone.findZone(SockAddrInet('192.168.1.129', 10)), t7)
+		test('192.168.0.184', root_zone.findZone(SockAddrInet('192.168.0.184', 10)), t14)
 
 		inet = InetZone("internet", "0.0.0.0/0", inbound_services=["s2"], outbound_services=["s1"])
 		test('1.1.1.1', root_zone.findZone(SockAddrInet('1.1.1.1', 10)), inet)
