@@ -101,6 +101,37 @@ z_szig_logging(ZSzigContext *ctx, const char *subcmd, const char *param, char *r
     }
 }
 
+int
+z_szig_deadlockcheck(ZSzigContext *ctx, const char *subcmd, char *result, size_t result_len)
+{
+  char buf[SZIG_MAX_CMD_LENGTH];
+  char res_buf[128];
+
+  snprintf(buf, sizeof(buf), "DEADLOCKCHECK %s\n", subcmd);
+  if (!z_szig_write_request(ctx, buf))
+    return 0;
+  if (!z_szig_read_response(ctx, res_buf, sizeof(res_buf)))
+    return 0;
+
+  if (strncmp(res_buf, "FAIL ", 5) == 0)
+    {
+      return 0;
+    }
+  else if (strncmp(res_buf, "OK ", 3) == 0)
+    {
+      if (result)
+        {
+          strncpy(result, res_buf + 3, result_len);
+          res_buf[result_len - 1] = 0;
+        }
+      return 1;
+    }
+  else
+    {
+      return 0;
+    }
+}
+
 int
 z_szig_reload(ZSzigContext *ctx, const char *subcmd, char *result, size_t result_len)
 {
@@ -137,6 +168,33 @@ z_szig_reload(ZSzigContext *ctx, const char *subcmd, char *result, size_t result
 }
 
 
+
+int
+z_szig_coredump(ZSzigContext *ctx)
+{
+  char buf[SZIG_MAX_CMD_LENGTH];
+  char res_buf[128];
+
+  snprintf(buf, sizeof(buf), "COREDUMP\n");
+  if (!z_szig_write_request(ctx, buf))
+    return 0;
+  if (!z_szig_read_response(ctx, res_buf, sizeof(res_buf)))
+    return 0;
+
+  if (strncmp(res_buf, "FAIL ", 5) == 0)
+    {
+      return 0;
+    }
+  else if (strncmp(res_buf, "OK ", 3) == 0)
+    {
+      return 1;
+    }
+  else
+    {
+      return 0;
+    }
+}
+
 ZSzigContext *
 z_szig_context_new(const char *instance_name)
 {
@@ -148,7 +206,7 @@ z_szig_context_new(const char *instance_name)
   if (fd == -1)
     return NULL;  
   unaddr.sun_family = AF_UNIX;
-  snprintf(unaddr.sun_path, sizeof(unaddr.sun_path), "/var/run/zorp/zorpctl.%s", instance_name);
+  snprintf(unaddr.sun_path, sizeof(unaddr.sun_path), ZORP_PIDFILEDIR "zorpctl.%s", instance_name);
   if (connect(fd, (struct sockaddr *) &unaddr, sizeof(unaddr)) < 0)
     {
       close(fd);
@@ -169,3 +227,4 @@ z_szig_context_destroy(ZSzigContext *ctx)
       free(ctx);
     }
 }
+
